// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package algo;

import java.util.Hashtable;
import java.util.Vector;

import model.DatumZeit;
import model.Haltestelle;
import model.HaltestellenPaar;
import model.IFahrt;
import model.IFahrtStatic;
import model.Route;
import model.RoutenAbschnitt;
import data.Data;
import dev.Logger;
import dev.algoProfiler.AlgoProfilerEmpty;
import dev.algoProfiler.IAlgoProfiler;

public class Algorithmus {

	// Konstanten für calcRoute

	// OPTIMIERUNG_X müssen in der Reihenfolge aufsteigend sein nach Genauigkeit
	public static final short OPTIMIERUNG_KEINE = 0;
	public static final short OPTIMIERUNG_NORMAL = 1;
	public static final short OPTIMIERUNG_LANG = 2;
	public static final short OPTIMIERUNG_TESTCASE_VERSION = 3;
	public static final short OPTIMIERUNG_ENDLOS = 4;
	
	private static final String[] optimierungsString = {"keine", "normal", "lang", "testcase - definiert", "endlos"};

	private static IAlgoProfiler algoProfiler = new AlgoProfilerEmpty();
	
	// Schaltet alle Debug Meldungen im Algorithmus Teil ein oder aus.
	public static boolean showAlgoDebugMsgs = false;
	public static boolean showEinsZweiOptiRoute = false;
	public static boolean showBuildRoute = false;
	public static boolean showFullTable = false;

	public static final String getOptimierungsString(short optimierungsTyp) {
		if (optimierungsTyp < 0 || optimierungsTyp >= optimierungsString.length) {
			return "KEIN GÜLTERIGER OPTIMIERUNGSTYPE";
		}
		return optimierungsString[optimierungsTyp];
	}

	/**
	 * Berechnet eine Route die früher losfährt und ankommt als die übergebene.
	 */
	public static Route calcFormerRoute(Route vorherigeRoute) {
		// Wenn die vorherige Route keine Verbindung gefunden hat, dann auch nicht diese.
		if (vorherigeRoute.getErrorCode() != Route.STATUS_KEIN_FEHLER) {
			return vorherigeRoute;
		}

		// CACHE
		// schauen, ob wir schon was im Cache haben
		Route cache = RoutenCache.findCachedEntry(vorherigeRoute.getHaltestellenPaar(), vorherigeRoute.getAnkunftszeit() - 1, new DataWrapperReverse());
		if (cache != Route.NULL) {
			return cache;
		}

		// ERSTE BERECHNUNG
		Algorithmus.algoProfiler.profilingStarten();
		RoutenResultHolder eins = calcDijkstraRoute(vorherigeRoute.getHaltestelleNach(), vorherigeRoute.getHaltestelleVon(), vorherigeRoute.getAnkunftszeit() - 1, new DataWrapperReverse(), null, vorherigeRoute.getOptimierungsTyp());
		if (showEinsZweiOptiRoute || showAlgoDebugMsgs)
			System.out.println("### EINS ### \n" + eins.toString());
		Algorithmus.algoProfiler.nachErsteDijkstraBerechnung();

		// AUSNAHME: REINE LAUFWEGE
		// falls die letzte Route ein Laufweg war und die nachfolgende auch werden solange neue Routen
		// berechnet bis es nicht mehr ein reiner Laufweg ist
		// FIXME: Nicht performant und Codedublizierung (mit calcLaterRoute)
		if (vorherigeRoute.isReinerLaufweg()) {
			Route nextRoute = eins.route;
			int timeoffset = 1;
			int maxTimeOffset = 600;
			while (nextRoute.isReinerLaufweg()) {
				eins = calcDijkstraRoute(vorherigeRoute.getHaltestelleNach(), vorherigeRoute.getHaltestelleVon(), vorherigeRoute.getAnkunftszeit() - ++timeoffset, new DataWrapperReverse(), null, vorherigeRoute.getOptimierungsTyp());
				nextRoute = eins.route;
				if (timeoffset == maxTimeOffset) {
					Logger.error(Algorithmus.class, "timeOffset > " + maxTimeOffset);
					break;
				}
			}
		}

		// Sollte die erste berechnete Route ein reiner Laufweg sein kann die reverse Route und die
		// Optimierung weggelassen werden Evtl. gehen dabei die Optimierung verloren, dass bei gleichschnellem
		// Laufweg wie Fahrstrecke nur der Laufweg angezeigt wird
		if (eins.route.isReinerLaufweg()) {
			return eins.route;
		}

		// ZWEITE BERECHNUNG
		RoutenResultHolder zwei = calcDijkstraRoute(vorherigeRoute.getHaltestelleVon(), vorherigeRoute.getHaltestelleNach(), eins.route.getAbfahrtszeit(), new DataWrapperNormal(), eins.zeitBisHaltestelle, vorherigeRoute.getOptimierungsTyp());
		if (showEinsZweiOptiRoute || showAlgoDebugMsgs)
			System.out.println("### ZWEI ### \n" + zwei.toString());
		Algorithmus.algoProfiler.nachZweiteDijkstraBerechnung();

		// zu lange Wartezeit? ja - dann neue rechnen
		// nur eine einfache Rekursion, keine mehrfache Rekursion
		if (isLangeWarten(zwei.route) && (!isLangeWarten(vorherigeRoute))) {
			if (showEinsZweiOptiRoute || showAlgoDebugMsgs) {
				System.out.println("### Lange warten ### \n");
			}
			return calcFormerRoute(zwei.route);
		}

		// OPTMIERUNG
		Route ret = RoutenOptimierer.optimiereRoute(eins, zwei, vorherigeRoute.getOptimierungsTyp());
		if (showEinsZweiOptiRoute || showAlgoDebugMsgs) {
			System.out.println("### RETURN BESTE ROUTE ### \n" + ret.toString());
		}
		Algorithmus.algoProfiler.nachRoutenOptimierung();
		Algorithmus.algoProfiler.profilingFertig();

		// NACHBEARBEITUNG
		// setze die ursprüngliche Anforderungszeit + Datawrapper
		ret.setZeitDerAnforderung(vorherigeRoute.getAnkunftszeit() - 1);
		ret.setDataWrapper(new DataWrapperReverse());

		// ab in den Cache
		RoutenCache.addRoute(ret);

		return ret;
	}

	/**
	 * Berechnet eine spätere Route, als die gegebene
	 */
	public static Route calcLaterRoute(Route vorherigeRoute) {
		// Wenn die vorherige Route keine Verbindung gefunden hat, dann auch nicht diese.
		if (vorherigeRoute.getErrorCode() != Route.STATUS_KEIN_FEHLER) {
			return vorherigeRoute;
		}

		// BERECHNE ROUTE
		Route nextRoute = Algorithmus.calcRoute(vorherigeRoute.getHaltestelleVon(), vorherigeRoute.getHaltestelleNach(), vorherigeRoute.getAbfahrtszeit() + 1, vorherigeRoute.getOptimierungsTyp(), vorherigeRoute);

		// AUSNAHME: REINE LAUFWEGE
		// falls die letzte Route ein Laufweg war und die nachfolgende auch werden solange neue Routen
		// berechnet bis es nicht mehr ein reiner Laufweg ist
		// FIXME: Nicht performant und Codedublizierung (mit calcFormerRoute)
		if (vorherigeRoute.isReinerLaufweg()) {
			if (showEinsZweiOptiRoute || showAlgoDebugMsgs) {
				System.out.println("### reiner Laufweg ### \n");
			}
			int timeoffset = 1;
			int maxTimeOffset = 600;
			while (nextRoute.isReinerLaufweg()) {
				nextRoute = Algorithmus.calcRoute(vorherigeRoute.getHaltestelleVon(), vorherigeRoute.getHaltestelleNach(), vorherigeRoute.getAbfahrtszeit() + (++timeoffset), vorherigeRoute.getOptimierungsTyp());
				if (timeoffset == maxTimeOffset) {
					Logger.error(Algorithmus.class, "timeOffset > " + maxTimeOffset);
					break;
				}
			}
		}

		return nextRoute;
	}

	/**
	 * Berechnet eine Route, mit gewählter Optimierung.
	 * 
	 * @param von
	 * @param nach
	 * @param datumZeit
	 *            Zeit der Anforderung
	 * @param optimierungsTyp
	 *            Bitte die entsprechenden Konstanten verwenden:<br>
	 *            Algorithmus.OPTIMIERUNG_KEINE <br>
	 *            Algorithmus.OPTIMIERUNG_NORMAL <br>
	 *            Algorithmus.OPTIMIERUNG_LANG <br>
	 *            Algorithmus.OPTIMIERUNG_ENDLOS
	 * @return
	 */
	public static Route calcRoute(Haltestelle von, Haltestelle nach, int datumZeit, short optimierungsTyp) {
		return calcRoute(von, nach, datumZeit, optimierungsTyp, null);
	}

	private static Route calcRoute(Haltestelle von, Haltestelle nach, int datumZeit, short optimierungsTyp, Route vorherigeRoute) {

		// Zwei gleiche Haltestellen?
		if (von.equals(nach)) {
			System.out.print("Von und nach Haltestelle sind gleich.");
			return new Route(Route.STATUS_GLEICHER_START_UND_ZIELORT);
		}

		// CACHE
		// schauen, ob wir schon was im Cache haben
		Route cachedRoute = RoutenCache.findCachedEntry(new HaltestellenPaar(von, nach), datumZeit, new DataWrapperNormal());
		if (cachedRoute != Route.NULL) {
			// gibt die gecachete Route nur zurück, falls die Optimierung abgeschlossen wurde oder bereits
			// eine Route im Cache liegt mit mindestens der gewünschten Optimierungstiefe
			if ((cachedRoute.getOptimierungsStatus() == Route.OPTIMIERUNG_ABGESCHLOSSEN) || (cachedRoute.getOptimierungsTyp() >= optimierungsTyp)) {
				return cachedRoute;
			}
		}

		// ERSTE BERECHNUNG
		Algorithmus.algoProfiler.profilingStarten();
		RoutenResultHolder eins = calcDijkstraRoute(von, nach, datumZeit, new DataWrapperNormal(), null, optimierungsTyp);
		if (showEinsZweiOptiRoute || showAlgoDebugMsgs) {
			System.out.println("### EINS ### \n" + eins.route.toString());
		}
		Algorithmus.algoProfiler.nachErsteDijkstraBerechnung();

		if (!eins.route.isValid()) {
			Logger.error(Algorithmus.class, "keine Verbindung gefunden! " + von + " -> " + nach);
			return eins.route;
		}

		// ZWEITE BERECHNUNG
		RoutenResultHolder zwei = calcDijkstraRoute(nach, von, eins.route.getAnkunftszeit(), new DataWrapperReverse(), eins.zeitBisHaltestelle, optimierungsTyp);
		if (showEinsZweiOptiRoute || showAlgoDebugMsgs) {
			System.out.println("### ZWEI ### \n" + zwei.route.toString());
		}

		Algorithmus.algoProfiler.nachZweiteDijkstraBerechnung();

		// zu lange Wartezeit? ja - dann neue rechnen
		// nur eine einfache Rekursion, keine mehrfache Rekursion
		if (isLangeWarten(zwei.route) && ((vorherigeRoute == null) || (!isLangeWarten(vorherigeRoute)))) {
			if (showEinsZweiOptiRoute || showAlgoDebugMsgs) {
				System.out.println("### Lange warten ### \n");
			}
			return calcLaterRoute(zwei.route);
		}

		// OPTIMIERUNG
		Route ret;
		if (optimierungsTyp != Algorithmus.OPTIMIERUNG_KEINE) {
			ret = RoutenOptimierer.optimiereRoute(eins, zwei, optimierungsTyp);
			if (showEinsZweiOptiRoute || showAlgoDebugMsgs) {
				System.out.println("### RETURN BESTE ROUTE ### \n" + ret.toString());
			}
			Algorithmus.algoProfiler.nachRoutenOptimierung();
		} else {
			ret = zwei.route;
		}
		Algorithmus.algoProfiler.profilingFertig();

		// NACHBEARBEITUNG
		// setze die ursprüngliche Anforderungszeit + Datawrapper
		ret.setZeitDerAnforderung(datumZeit);
		ret.setDataWrapper(new DataWrapperNormal());

		// Ab in den Cache, falls optimiert wurde.
		if (ret.getOptimierungsStatus() != Route.OPTIMIERUNG_NOCHT_NICHT_OPTIMIERT) {
			RoutenCache.addRoute(ret);
		}

		return ret;
	}

	/**
	 * Berechnet eine Route mit normaler Optimierung<br>
	 * Wrapper für calcRoute(Haltestelle von, Haltestelle nach, int datumZeit, short optimierung)
	 */
	public static Route calcRoute(Haltestelle von, Haltestelle nach, int datumZeit) {
		return Algorithmus.calcRoute(von, nach, datumZeit, Algorithmus.OPTIMIERUNG_NORMAL);
	}

	/**
	 * Berechnet eine Route für den StartScreen, die weder optimiert noch gecacht wird.<br>
	 * Wrapper für calcRoute(Haltestelle von, Haltestelle nach, int datumZeit, short optimierung)
	 */
	public static Route calcStartScreenRoute(Haltestelle von, Haltestelle nach) {
		return Algorithmus.calcRoute(von, nach, DatumZeit.Now(), Algorithmus.OPTIMIERUNG_KEINE);
	}

	/**
	 * Kapselt die Verwendung des DijkstraAlgorithmus
	 * 
	 * @param routeVon
	 * @param routeNach
	 * @param aktDatumZeit
	 * @param iDataWrapper
	 * @param optimierungsZeitTabelle
	 * @return
	 */
	protected static RoutenResultHolder calcDijkstraRoute(Haltestelle routeVon, Haltestelle routeNach, int aktDatumZeit, IDataWrapper iDataWrapper, int[] optimierungsZeitTabelle, short optimierungsTyp) {

		// Berechne Route mit Dijkstra
		DijkstraResultHolder result = DijkstraImpl.calcDijkstra(routeVon, routeNach, aktDatumZeit, iDataWrapper, optimierungsZeitTabelle);

		IFahrt[] fahrtBisHaltestelle = result.fahrtBisHaltestelle;
		Hashtable htMultipleEntries = result.htMultipleEntries;
		int[] zeitBisHaltestelle = result.zeitBisHaltestelle;
		
		if (showBuildRoute || showFullTable || showAlgoDebugMsgs) {
			System.out.println("################## NACH DIJKSTRA #####################");
		}
	
		// <DEBUG>
		if (showFullTable || showAlgoDebugMsgs) {
			System.out.println("\n\n      ############ komplette Tabelle #####################\n");
			for (int i = 0; i < result.fahrtBisHaltestelle.length; i++) {
				Haltestelle haltestelle = Data.getAlleHaltestellen()[i];
				int anzahlOptionen;
				Vector entries;
				if (fahrtBisHaltestelle[haltestelle.getId()] != DijkstraImpl.MULTIPLE_ENTRIES) {
					anzahlOptionen = 1;
					entries = null;
				} else {
					entries = (Vector) htMultipleEntries.get(haltestelle);
					anzahlOptionen = entries.size();
				}

				for (int iOptionen = 0; iOptionen < anzahlOptionen; iOptionen++) {
					IFahrt fahrtBisActHaltestelle;
					int zeitBisActHaltestelle;
					if (anzahlOptionen == 1) {
						fahrtBisActHaltestelle = fahrtBisHaltestelle[haltestelle.getId()];
						zeitBisActHaltestelle = zeitBisHaltestelle[haltestelle.getId()];
					} else {
						XBisHaltestelle xBisHaltestelle = (XBisHaltestelle) entries.elementAt(iOptionen);
						fahrtBisActHaltestelle = xBisHaltestelle.fahrtBisHaltestelle;
						zeitBisActHaltestelle = xBisHaltestelle.zeitBisHaltestelle;
					}
					if (fahrtBisActHaltestelle == IFahrt.NULL) {
						System.out.println(haltestelle + " - NOT ARRIVED -");
					} else {
						if (haltestelle == routeVon) {
							System.out.println(haltestelle + " - ABFAHRTSHALTESTELLE -");
						} else {
							System.out.println(haltestelle + " @ " + DatumZeit.toTimeString(zeitBisActHaltestelle) + " with " + fahrtBisActHaltestelle.getLinienInfo() + " " + (Data.getIdOfFahrt(fahrtBisActHaltestelle)) + " ("+(iOptionen+1)+"/"+anzahlOptionen+")");
						}
					}
				}
			}
		}
		// </DEBUG>

		// Erstelle Rückgabeobjekt Route
		Vector routenAbschnitte = new Vector(); // RoutenAbschnitte der Route

		// wenn es keine Fahrt zur Zielhaltestelle gibt wurde sie nicht erreicht
		if (result.fahrtBisHaltestelle[routeNach.getId()] == IFahrt.NULL) {
			return new RoutenResultHolder(new Route(Route.STATUS_KEINE_ROUTE_GEFUNDEN), null);
		}

		if (showBuildRoute || showAlgoDebugMsgs) {
			System.out.println("\n\n      ############ Zusammenbau #####################\n");		
		}
		
		// falls die Nach-Haltestelle mehrere Optionen hat die 'früheste' Ankunftszeit auswählen und zurückschreiben
		if (fahrtBisHaltestelle[routeNach.getId()] == DijkstraImpl.MULTIPLE_ENTRIES) {
			if (showBuildRoute || showAlgoDebugMsgs) {
				System.out.println("Zielhaltestelle hat mehrere Einträge - selektiere den Besten\n");		
			}

			Vector entries = (Vector) htMultipleEntries.get(routeNach);
			XBisHaltestelle bestXBisHaltestelle = (XBisHaltestelle) entries.elementAt(0);
			for (int i = 0; i < entries.size(); i++) {
				XBisHaltestelle xBisHaltestelle = (XBisHaltestelle) entries.elementAt(i);
				int zeitBisActHaltestelle = xBisHaltestelle.zeitBisHaltestelle;
				if (showBuildRoute || showAlgoDebugMsgs) {
					System.out.println(" ^^^ #"+i+": "+xBisHaltestelle.fahrtBisHaltestelle + " @ " + DatumZeit.toTimeString(xBisHaltestelle.zeitBisHaltestelle));		
				}
				if (iDataWrapper.isSmallerThan(zeitBisActHaltestelle, bestXBisHaltestelle.zeitBisHaltestelle)) {
					bestXBisHaltestelle = xBisHaltestelle;
				}
			}
			fahrtBisHaltestelle[routeNach.getId()] = bestXBisHaltestelle.fahrtBisHaltestelle;
			zeitBisHaltestelle[routeNach.getId()] = bestXBisHaltestelle.zeitBisHaltestelle;
		}
		
		// Durchläuft previous vom Ziel rückwärts zum Ausgangspunkt
		// bei Alternativen wird die erste Möglichkeit verwendet
		Haltestelle routenPosition = routeNach;
		
		while (routenPosition != routeVon) {
			int anzahlOptionen;
			Vector entries;
			if (fahrtBisHaltestelle[routenPosition.getId()] != DijkstraImpl.MULTIPLE_ENTRIES) {
				anzahlOptionen = 1;
				entries = null;
			} else {
				entries = (Vector) htMultipleEntries.get(routenPosition);
				if (entries == null || entries.size() < 2) {
					Logger.error(Algorithmus.class, "Vector von multipleEntries: zu wenig Elemente");
					return null;
				}
				anzahlOptionen = entries.size();
			}

			boolean gueltigeOptionGefunden = false;
			for (int iOptionen = 0; iOptionen < anzahlOptionen; iOptionen++) {
				IFahrt fahrtBisActHaltestelle;
				int zeitBisActHaltestelle;
				if (anzahlOptionen == 1) {
					fahrtBisActHaltestelle = fahrtBisHaltestelle[routenPosition.getId()];
					zeitBisActHaltestelle = zeitBisHaltestelle[routenPosition.getId()];
				} else {
					XBisHaltestelle xBisHaltestelle = (XBisHaltestelle) entries.elementAt(iOptionen);
					fahrtBisActHaltestelle = xBisHaltestelle.fahrtBisHaltestelle;
					zeitBisActHaltestelle = xBisHaltestelle.zeitBisHaltestelle;
				}
				if (showBuildRoute || showAlgoDebugMsgs) {
					System.out.println(routenPosition + " @ " + DatumZeit.toTimeString(zeitBisActHaltestelle) + " with " + fahrtBisActHaltestelle.getLinienInfo() + " " + (Data.getIdOfFahrt(fahrtBisActHaltestelle)) + " ("+(iOptionen+1)+"/"+anzahlOptionen+")");
				}

				Haltestelle nach = routenPosition;
				Haltestelle von;
				if (iDataWrapper instanceof DataWrapperNormal) {
					von = fahrtBisActHaltestelle.getNextHaltestelleReverse(nach);
				} else {
					von = fahrtBisActHaltestelle.getNextHaltestelle(nach);
				}

				// Ermittle die Zeit wann wir an der Nach-Haltestelle ankommen.
				int ankunftAnNachHaltestelle = zeitBisActHaltestelle;
				
				// Entscheidung, ob aktuelle Option gültig ist
				// nicht beim ersten Durchlauf
				if (routenPosition != routeNach) {
					RoutenAbschnitt nachfolgenderRoutenabschnitt = (RoutenAbschnitt) routenAbschnitte.lastElement();
					int nachfolgendeAbfahrtsZeit = nachfolgenderRoutenabschnitt.getAbfahrtsZeit();
					int umsteigeZeit = IFahrtStatic.getUmsteigezeit(fahrtBisActHaltestelle.getLinie(), nachfolgenderRoutenabschnitt.getFahrt().getLinie(), nachfolgenderRoutenabschnitt.getVon());
					if (iDataWrapper.isGreaterThan(iDataWrapper.addTime(zeitBisActHaltestelle, umsteigeZeit), nachfolgendeAbfahrtsZeit)) {
						continue;
					}
				}

				// BERECHNE ABFAHRTSZEIT
				// Berechne die Abfahrtszeit, indem die Fahrzeit zwischen den Haltestellen von der
				// ankunftAnNachHaltestelle abgezogen wird.
				int abfahrtsZeit = DatumZeit.NULL;
				if (iDataWrapper instanceof DataWrapperNormal) {
					abfahrtsZeit = ankunftAnNachHaltestelle - fahrtBisActHaltestelle.getTimeToNextHaltestelle(von);
				} else if (iDataWrapper instanceof DataWrapperReverse) {
					abfahrtsZeit = ankunftAnNachHaltestelle + fahrtBisActHaltestelle.getTimeToNextHaltestelle(nach);
				}
				
				routenAbschnitte.addElement(new RoutenAbschnitt(von, nach, abfahrtsZeit, ankunftAnNachHaltestelle, fahrtBisActHaltestelle));
				routenPosition = von;

				gueltigeOptionGefunden = true;
				// falls eine gültige Option gefunden wurde müssen keine weiteren untersucht werden
				break;
			}
			if (!gueltigeOptionGefunden) {
				Logger.error(Algorithmus.class, "keine gültige Option gefunden");
				return null;
			}
		}

		// Wandelt den Vector in ein Array um
		// und dreht je nach DataWrapper dabei die Reihenfolge um
		Route route = new Route(iDataWrapper.buildRoutenAbschnittArray(routenAbschnitte), aktDatumZeit, iDataWrapper);
		route.setOptimierungsTyp(optimierungsTyp);

		return new RoutenResultHolder(route, result.zeitBisHaltestelle);
	}

	/**
	 * Gibt true zurück falls in der Route mehr als 120 Minuten an einer Haltestelle gewartet wird.
	 */
	private static boolean isLangeWarten(Route route) {
		RoutenAbschnitt[] abschnitte = route.getKompakteRoutenAbschnitte();
		for (int i = 1; i < abschnitte.length; i++) {
			RoutenAbschnitt abschnittVorher = abschnitte[i - 1];
			RoutenAbschnitt abschnittAktuell = abschnitte[i];
			if (abschnittAktuell.getAbfahrtsZeit() - abschnittVorher.getAnkunftsZeit() > 120) {
				return true;
			}
		}
		return false;
	}

	public static void setAlgoProfiler(IAlgoProfiler algoProfiler) {
		Algorithmus.algoProfiler = algoProfiler;
	}
}
