// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package algo;

import java.util.Hashtable;
import java.util.Vector;

import model.DatumZeit;
import model.Fahrt;
import model.Haltestelle;
import model.IFahrt;
import model.Linie;
import utils.Queue;
import control.Global;
import data.Data;
import dev.Logger;

/**
 * Implementierung des Dijkstra Algorithmuses (kürzeste Wege in Graphen).
 */
public abstract class DijkstraImpl {

	public static final IFahrt MULTIPLE_ENTRIES = new Fahrt(null, null, null, null, null, null, null);

	private final static int ACTION_OVERWRITE = 0;
	private final static int ACTION_ADD = 1;
	private final static int ACTION_ADD_AS_NEW_MULTIPLE = 2;
	private final static int ACTION_NOTHING = 3;

	/**
	 * Führt den Dijkstra Algorithmus durch.
	 * 
	 * @param von
	 *            Haltestelle
	 * @param nach
	 *            Haltestelle
	 * @param aktDatumZeit
	 *            Zeit der Anforderung
	 * @param iDataWrapper
	 * @param zeitBisHaltestelleVorinitialisierung
	 *            Kommt immer von einer Route mit dem GLEICHEN dataWrapper wie die Route die hier berechnet
	 *            werden soll.
	 * @param optimierungsZeitTabelle
	 *            Kommt immer von einer Route mit dem UMGEKEHRTEN dataWrapper wie die Route die hier berechnet
	 *            werden soll.
	 * @return Route
	 */
	protected static DijkstraResultHolder calcDijkstra(Haltestelle von, Haltestelle nach, int aktDatumZeit, IDataWrapper iDataWrapper, int[] optimierungsZeitTabelle) {

		// (Haltestelle) -> (Ankunftszeit)
		int[] zeitBisHaltestelle = new int[Data.getAlleHaltestellen().length];

		// (Haltestelle) -> (Fahrt, mit der man die Haltestelle erreicht)
		IFahrt[] fahrtBisHaltestelle = new IFahrt[Data.getAlleHaltestellen().length];

		// (Haltestelle) -> Vector(int, Fahrt)
		Hashtable htMultipleEntries = new Hashtable();

		Queue myQueue = new Queue(zeitBisHaltestelle, (iDataWrapper instanceof DataWrapperReverse));

		for (int i = 0; i < Data.getAlleHaltestellen().length; i++) {
			fahrtBisHaltestelle[i] = IFahrt.NULL;
			// ... Setzt Distanz (Ankunftszeit) von jeder Haltestelle auf "UNENDLICH"
			zeitBisHaltestelle[i] = iDataWrapper.unendlichSpaet();
		}

		// INITIALISIERUNG des Startknoten 'von':
		// Feld im Array muss besetzt werden, dass der Algorithmus weiß, dass er schon dort war.
		// ACHTUNG: diese Fahrt darf nicht NULL_FAHRT sein
		fahrtBisHaltestelle[von.getId()] = new Fahrt(Linie.NULL, null, null, null, null, null, null);
		// Speichere als Distanz (Ankunftszeit) zum Startknotens die Jetzt-Zeit
		zeitBisHaltestelle[von.getId()] = aktDatumZeit;
		// ... Füge den Startknoten 'von', als erstes Element der Queue hinzu
		myQueue.add(von);

		// Solange es noch eine Haltestelle auf der Queue gibt, bearbeite diese
		while (!myQueue.isEmpty()) {

			// Suche von allen Haltestellen auf der Queue diejenige heraus,
			// welche die kürzeste Distanz (früheste Ankunftszeit) hat.
			// (Und loescht somit den Eintrag aus der Queue.)
			Haltestelle actHaltestelle = myQueue.getNext();
			if (Algorithmus.showAlgoDebugMsgs)
				System.out.println("NEXT ELEMENT FROM QUEUE: " + actHaltestelle.getFullName());

			/*
			 * ABBRUCHKRITERIUM: Ermittelt die Zeit an der aktuellen Haltestelle und wann die Ziel-Haltestelle
			 * erreicht wurde. (Falls das Ziel noch nicht erreicht wurde ist die Zeit dort Integer.MAX_VALUE
			 * bzw. Integer.MIN_VALUE bei Reverse. Gesetzt durch DataWrapper.unedlichSpaet()) Wenn das
			 * erreichen der aktuellen Haltestelle zeitlich hinter dem erreich der Ziel Haltestelle liegt kann
			 * man abbrechen, weil die Route nicht mehr besser werden kann. Brachte beim Benchmark ca. 1/5
			 * Ersparnis
			 */
			int timeAtCurrentHaltestelle = zeitBisHaltestelle[actHaltestelle.getId()];
			int timeAtDestinationHaltestelle = zeitBisHaltestelle[nach.getId()];
			if (iDataWrapper.isGreaterThan(timeAtCurrentHaltestelle, timeAtDestinationHaltestelle)) {
				continue;
			}

			IFahrt[] alleFartenVonActHaltestelle = actHaltestelle.getAlleFahrten();
			for (int iFahrt = 0; iFahrt < alleFartenVonActHaltestelle.length; iFahrt++) {
				// Lädt die aktuelle Fahrt
				IFahrt actFahrt = alleFartenVonActHaltestelle[iFahrt];

				if (Algorithmus.showAlgoDebugMsgs)
					System.out.println("\n - DO Fahrt " + actFahrt);

				// Wenn letzte Haltestelle auf Fahrt, dann verwende Fahrt nicht, da
				// es keine abzuarbeitenden Knoten auf dieser Fahrt gibt
				if (!actFahrt.isLaufweg()) {
					Fahrt fahrt = (Fahrt) actFahrt;
					Haltestelle letzteHaltestelle;
					if (iDataWrapper instanceof DataWrapperNormal) {
						letzteHaltestelle = fahrt.getLastHaltestelle();
					} else {
						letzteHaltestelle = fahrt.getFirstHaltestelle();
					}
					if (letzteHaltestelle == actHaltestelle) {
						if (Algorithmus.showAlgoDebugMsgs)
							System.out.println(" -- Break Letzte Haltestelle auf Fahrt");
						continue;
					}
				}

				// Naechste Haltestelle der Fahrt
				Haltestelle nextHaltestelle = iDataWrapper.getNextHaltestelle(actFahrt, actHaltestelle);
				if (Algorithmus.showAlgoDebugMsgs)
					System.out.println(" -- NextHaltestelle = " + nextHaltestelle.getFullName() + " (ID = " + nextHaltestelle.getId() + ")");

				int anzahlOptionen;
				Vector entries;
				if (fahrtBisHaltestelle[actHaltestelle.getId()] != MULTIPLE_ENTRIES) {
					anzahlOptionen = 1;
					entries = null;
				} else {
					entries = (Vector) htMultipleEntries.get(actHaltestelle);
					if (entries == null || entries.size() < 2) {
						Logger.error(DijkstraImpl.class, "Vector von multipleEntries: zu wenig Elemente");
						return null;
					}
					anzahlOptionen = entries.size();
				}

				for (int iOptionen = 0; iOptionen < anzahlOptionen; iOptionen++) {
					IFahrt fahrtBisActHaltestelle;
					int zeitBisActHaltestelle;
					if (anzahlOptionen == 1) {
						fahrtBisActHaltestelle = fahrtBisHaltestelle[actHaltestelle.getId()];
						zeitBisActHaltestelle = zeitBisHaltestelle[actHaltestelle.getId()];
					} else {
						XBisHaltestelle xBisHaltestelle = (XBisHaltestelle) entries.elementAt(iOptionen);
						if (xBisHaltestelle.isBearbeitet()) {
							if (Algorithmus.showAlgoDebugMsgs)
								System.out.println("^^^ SKIP - schon bearbeitet " + xBisHaltestelle.fahrtBisHaltestelle.getLinie() + " (" + (iOptionen + 1) + "/" + anzahlOptionen + ")");
							continue;
						}
						fahrtBisActHaltestelle = xBisHaltestelle.fahrtBisHaltestelle;
						zeitBisActHaltestelle = xBisHaltestelle.zeitBisHaltestelle;
					}

					// Berechne die Ankunftszeit an der nächsten Haltestelle auf der Fahrt, wenn die Bahn zur
					// gegebenen Zeit an der aktuellen Haltestelle ankommt.
					// Gegeben: Fahrt, aktuelleHaltestelle, Zeit bis aktuelleHaltestelle
					// Gesucht: Ankunftszeit + Wartezeit + Fahrzeit
					// dist = Ankunftszeit an der nächsten Haltestelle
					int zeitNextHaltestelle = iDataWrapper.getTimeAtNextHaltestelle(actFahrt, actHaltestelle, zeitBisActHaltestelle, fahrtBisActHaltestelle.getLinie());

					if (Algorithmus.showAlgoDebugMsgs)
						System.out.println(" -- TimeAtNextHaltestelle = " + DatumZeit.toDateTimeString(zeitNextHaltestelle) + " (diff=" + (zeitNextHaltestelle - aktDatumZeit) + "min)");

					// Wenn eine Distanz von -1 zurück gegeben wird: Durchlauf abbrechen, weil es mit dieser
					// Fahrt
					// keine Fahrtmöglichkeit mehr gibt.
					if (zeitNextHaltestelle == -1) {
						if (Algorithmus.showAlgoDebugMsgs)
							System.out.println(" -- Break dist == -1");
						continue;
					}

					// <OPTIMIERUNG>
					// Optmimierung mit DistanzTabelle aus alten Berechnungen
					// Die alte Berechnung muss genau die umgekehrte zu der aktuellen Berechnungs-Richtung
					// sein!!
					// Umsteigen wird ignoriert, da die Umsteigezeit dist nur noch weiter erhöhen würde.
					// Wenn für die Haltestelle keine Zeit in der Tabelle gefunden wurde dann verwerfen wir
					// die Fahrt.
					if (optimierungsZeitTabelle != null) {
						int zeitZuNextHaltestelleAusUmgekehrterBerechnung = optimierungsZeitTabelle[nextHaltestelle.getId()];
						if (zeitZuNextHaltestelleAusUmgekehrterBerechnung == iDataWrapper.unendlichSpaet() || iDataWrapper.isSmallerThan(zeitZuNextHaltestelleAusUmgekehrterBerechnung, zeitNextHaltestelle)) {
							if (Algorithmus.showAlgoDebugMsgs)
								System.out.println(" -- Break optimierungsZeitTabelle");
							continue;
						}
					}
					// </OPTIMIERUNG>

					/*  Bestimmung der ACTION */
					int action = ACTION_NOTHING;

					int nextHaltestelleId = nextHaltestelle.getId();

					int alteZeit = zeitBisHaltestelle[nextHaltestelleId];
					IFahrt alteFahrt = fahrtBisHaltestelle[nextHaltestelleId];

					boolean isMultiple;
					if (fahrtBisHaltestelle[nextHaltestelleId] == MULTIPLE_ENTRIES) {
						isMultiple = true;
					} else {
						isMultiple = false;
					}

					if (fahrtBisHaltestelle[nextHaltestelle.getId()] == IFahrt.NULL) {
						action = ACTION_OVERWRITE;
						// muss nicht geprüft werden, ob es schon auf der Queue liegt
						myQueue.add(nextHaltestelle);
						if (Algorithmus.showAlgoDebugMsgs)
							System.out.println(" --- First time seen, adding. ");

					} else if (iDataWrapper.isSmallerThan(zeitNextHaltestelle, alteZeit)) {

						if (isMultiple) {
							// mehrere Elemente sind bereits eingetragen

							if (iDataWrapper.isEqualOrSmallerThan(alteZeit, iDataWrapper.addTime(zeitNextHaltestelle, Global.UMSTEIGEZEIT_VERKEHRSMITTELWECHSEL))) {
								// ^^^ innerhalb des Zeitfensters

								Vector oldVector = (Vector) htMultipleEntries.get(nextHaltestelle);
								Vector newVector = new Vector();
								Linie linie = actFahrt.getLinie();

								// früherer Ankunftszeitpunkt
								int fruehesteZeit = zeitNextHaltestelle;

								// überprüft welche Fahrten innerhalb des Zeitfensters liegen und eine
								// unterschiedliche Linie besitzen
								for (int i = 0; i < oldVector.size(); i++) {
									XBisHaltestelle aktuellerEintrag = (XBisHaltestelle) oldVector.elementAt(i);
									if (iDataWrapper.isEqualOrSmallerThan(aktuellerEintrag.zeitBisHaltestelle, iDataWrapper.addTime(fruehesteZeit, Global.UMSTEIGEZEIT_VERKEHRSMITTELWECHSEL))) {
										if (aktuellerEintrag.fahrtBisHaltestelle.getLinie() != linie) {
											newVector.addElement(aktuellerEintrag);
										}
									}
								}

								if (newVector.size() == 0) {
									htMultipleEntries.remove(nextHaltestelle);
									action = ACTION_OVERWRITE;
									if (Algorithmus.showAlgoDebugMsgs)
										System.out.println(" --- Multiple time seen and better. -- multiEntry -- remove all (special) (<- single)");
								} else {
									htMultipleEntries.remove(nextHaltestelle);
									htMultipleEntries.put(nextHaltestelle, newVector);
									action = ACTION_ADD;
									if (Algorithmus.showAlgoDebugMsgs)
										System.out.println(" --- Multiple time seen and better. -- multiEntry -- ?remove some?");
								}
							} else {
								// keine alte Fahrt befindet sich innerhalb des Zeitfensters --> remove all
								htMultipleEntries.remove(nextHaltestelle);
								action = ACTION_OVERWRITE;
								if (Algorithmus.showAlgoDebugMsgs)
									System.out.println(" --- Multiple time seen and better. -- multiEntry -- remove all (<- single)");
							}

						} else {
							// ein einzelnes Element ist bereits eingetragen

							if (iDataWrapper.isEqualOrSmallerThan(alteZeit, iDataWrapper.addTime(zeitNextHaltestelle, Global.UMSTEIGEZEIT_VERKEHRSMITTELWECHSEL))) {
								// ^^^ innerhalb des Zeitfensters

								Linie alteLinie = alteFahrt.getLinie();
								if (alteLinie == actFahrt.getLinie()) {
									action = ACTION_OVERWRITE;
									if (Algorithmus.showAlgoDebugMsgs)
										System.out.println(" --- Multiple time seen and better. -- singleEntry -- overwrite (special)");
								} else {
									action = ACTION_ADD_AS_NEW_MULTIPLE;
									if (Algorithmus.showAlgoDebugMsgs)
										System.out.println(" --- Multiple time seen and better. -- singleEntry (-> mutliple)");
								}
							} else {
								action = ACTION_OVERWRITE;
								if (Algorithmus.showAlgoDebugMsgs)
									System.out.println(" --- Multiple time seen and better. -- singleEntry -- overwrite");
							}
						}

						zeitBisHaltestelle[nextHaltestelleId] = zeitNextHaltestelle;
						// liegt schon auf der Queue

					} else {
						// gleicher oder späterer Ankunftszeitpunkt
						if (iDataWrapper.isEqualOrSmallerThan(zeitNextHaltestelle, iDataWrapper.addTime(zeitBisHaltestelle[nextHaltestelle.getId()], Global.UMSTEIGEZEIT_VERKEHRSMITTELWECHSEL))) {
							if (isMultiple) {
								// mehrere Elemente sind bereits eingetragen

								action = ACTION_ADD;
								// zeitBisHaltestelle passt, da dort die früheste Zeit gespeichert ist
								Vector v = (Vector) htMultipleEntries.get(nextHaltestelle);
								Linie linie = actFahrt.getLinie();
								for (int i = 0; i < v.size(); i++) {
									XBisHaltestelle aktuellerEintrag = (XBisHaltestelle) v.elementAt(i);
									if (aktuellerEintrag.fahrtBisHaltestelle.getLinie() == linie) {
										action = ACTION_NOTHING;
										break;
									}
								}
								if (Algorithmus.showAlgoDebugMsgs) {
									if (action == ACTION_ADD) {
										System.out.println(" --- Multiple time seen NOT better. -- multiEntry -- adding");
									} else {
										System.out.println(" --- Multiple time seen NOT better. -- multiEntry -- (special)");
									}
								}
							} else {
								// ein einzelnes Element ist bereits eingetragen

								Linie alteLinie = alteFahrt.getLinie();
								if (alteLinie == actFahrt.getLinie()) {
									action = ACTION_NOTHING;
									if (Algorithmus.showAlgoDebugMsgs)
										System.out.println(" --- Multiple time seen NOT better. -- singleEntry -- (special)");
								} else {
									action = ACTION_ADD_AS_NEW_MULTIPLE;
									if (Algorithmus.showAlgoDebugMsgs)
										System.out.println(" --- Multiple time seen NOT better. -- singleEntry -- adding (-> multiple)");
								}
							}
						} else {
							action = ACTION_NOTHING;
							if (Algorithmus.showAlgoDebugMsgs)
								System.out.println(" --- Multiple time seen NOT better.");
						}

						// falls ein Element hinzugefügt wurde und die Haltestelle nicht (mehr) auf der Queue
						// liegt, wird sie hinzugefügt
						if (action != ACTION_NOTHING) {
							if (!myQueue.contains(nextHaltestelle)) {
								myQueue.add(nextHaltestelle);
							}
						}
					}

					/*  Behandlung der ACTION */
					switch (action) {
					case ACTION_NOTHING:
						break;
					case ACTION_ADD_AS_NEW_MULTIPLE:
						XBisHaltestelle alterEintrag1 = new XBisHaltestelle(alteZeit, alteFahrt);
						XBisHaltestelle neuerEintrag1 = new XBisHaltestelle(zeitNextHaltestelle, actFahrt);
						// TODO ^^^hier kann optimiert werden, da evtl als nicht bearbeitet markiert sind,
						// welche aber schon bearbeitet wurden
						Vector v1 = new Vector();
						v1.addElement(alterEintrag1);
						v1.addElement(neuerEintrag1);
						fahrtBisHaltestelle[nextHaltestelleId] = MULTIPLE_ENTRIES;
						htMultipleEntries.put(nextHaltestelle, v1);
						break;
					case ACTION_ADD:
						XBisHaltestelle neuerEintrag2 = new XBisHaltestelle(zeitNextHaltestelle, actFahrt);
						Vector v2 = (Vector) htMultipleEntries.get(nextHaltestelle);
						v2.addElement(neuerEintrag2);
						break;
					case ACTION_OVERWRITE:
						zeitBisHaltestelle[nextHaltestelleId] = zeitNextHaltestelle;
						fahrtBisHaltestelle[nextHaltestelleId] = actFahrt;
						break;
					default:
						Logger.error(DijkstraImpl.class, "have a nice day...");
					}
				}
			} // Ende der Schleife über alle Fahrten der aktuellen Haltestelle
			// markiere die bereits abgearbeiteten Haltestellen (nur bei Mehrfacheinträgen)
			if (fahrtBisHaltestelle[actHaltestelle.getId()] == MULTIPLE_ENTRIES) {
				Vector entries = (Vector) htMultipleEntries.get(actHaltestelle);
				int anzahlOptionen = entries.size();
				for (int iOptionen = 0; iOptionen < anzahlOptionen; iOptionen++) {
					XBisHaltestelle xBisHaltestelle = (XBisHaltestelle) entries.elementAt(iOptionen);
					xBisHaltestelle.setBearbeitet();
				}
			}
		}

		// Erstelle Ergebnis Objekt
		DijkstraResultHolder result = new DijkstraResultHolder(fahrtBisHaltestelle, zeitBisHaltestelle, htMultipleEntries);
		return result;
	}
}

class XBisHaltestelle {
	public XBisHaltestelle(int zeitBisHaltestelle, IFahrt fahrtBisHaltestelle) {
		this.zeitBisHaltestelle = zeitBisHaltestelle;
		this.fahrtBisHaltestelle = fahrtBisHaltestelle;
	}

	protected final int zeitBisHaltestelle;
	protected final IFahrt fahrtBisHaltestelle;
	private boolean bearbeitet = false;

	protected boolean isBearbeitet() {
		return bearbeitet;
	}

	protected void setBearbeitet() {
		bearbeitet = true;
	}
}
