// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package algo;

import model.HaltestellenPaar;
import model.Route;
import control.Global;
import dev.Logger;

/**
 * Kapselt die Cachefunktionalität. Der Cache ist aufgeteilt in hin und rück Cache und hat die Größe
 * CACHE_SIZE überschrieben wird immer das nächstfolgende Element mit writePointer
 */

public class RoutenCache {

	private final static int CACHE_SIZE = Global.ROUTE_CACHE_SIZE;

	private static int writePointerForward = 0;
	private static int writePointerBackward = 0;

	private static Route[] cacheEntriesForward;
	private static Route[] cacheEntriesBackward;

	/**
	 * Fügt die übergebene Route zum Cache hinzu, falls dieser aktivert ist und die Route kein reiner Laufweg
	 * ist die Unterscheidung für reverse oder normal wird per route.getDatawrapper ermittelt
	 */
	public static void addRoute(Route route) {
		// Diverse Kriterien bei denen der Route nicht in den Cache kommt.
		if (!Global.USE_ROUTE_CACHE) {
			return;
		}
		if (route.isReinerLaufweg()) {
			return;
		}
		if (cacheEntriesForward == null) {
			init();
		}
		if (route.getDataWrapper() == null) {
			Logger.error(RoutenCache.class, "route muss einen Datawrapper besitzen");
			return;
		}

		// Falls eine solche Route schon im Cache vorhanden ist wird die Position bestimmt und diese
		// überschrieben. Wenn nicht ist cachePosition == -1.
		int cachePosition = getCachePosition(route.getHaltestellenPaar(), route.getZeitDerAnforderung(),
			route.getDataWrapper());

		// Schreibt die Route in den Cache.
		if (route.getDataWrapper() instanceof DataWrapperNormal) {
			if (cachePosition == -1) {
				writePointerForward = (writePointerForward +1) % CACHE_SIZE;
				cachePosition = writePointerForward;
			}
			cacheEntriesForward[cachePosition] = route;

		} else {
			if (cachePosition == -1) {
				writePointerBackward = (writePointerBackward + 1) % CACHE_SIZE;  
				cachePosition = writePointerBackward;
			}
			cacheEntriesBackward[cachePosition] = route;
		}
	}
	
	/**
	 * Sucht nach einem passenden Cache-Entry falls kein Entry gefunden wird kommt Global.NULL_ROUTE zurück
	 */
	public static Route findCachedEntry(HaltestellenPaar haltestellenPaar, int zeitDerAnforderung,
		IDataWrapper datawrapper) {
		if (cacheEntriesForward == null) {
			init();
			return Route.NULL;
		}

		int cachePosition = getCachePosition(haltestellenPaar, zeitDerAnforderung, datawrapper);

		if (cachePosition == -1) {
			return Route.NULL;
		}

		Route[] cacheEntries;
		if (datawrapper instanceof DataWrapperNormal) {
			cacheEntries = cacheEntriesForward;
		} else {
			cacheEntries = cacheEntriesBackward;
		}
		return cacheEntries[cachePosition];
	}

	private static int getCachePosition(HaltestellenPaar haltestellenPaar, int zeitDerAnforderung,
		IDataWrapper datawrapper) {
		Route[] cacheEntries;
		if (datawrapper instanceof DataWrapperNormal) {
			cacheEntries = cacheEntriesForward;
			for (int i = 0; i < CACHE_SIZE; i++) {
				Route cacheCurrent = cacheEntries[i];
				HaltestellenPaar chacheHaltestellenPaar = cacheCurrent.getHaltestellenPaar();
				if (chacheHaltestellenPaar.equals(haltestellenPaar)) {
					if (cacheCurrent.getZeitDerAnforderung() <= zeitDerAnforderung
						&& zeitDerAnforderung <= cacheCurrent.getAbfahrtszeit()) {
						return i;
					}
				}
			}
		} else {
			cacheEntries = cacheEntriesBackward;
			for (int i = 0; i < CACHE_SIZE; i++) {
				Route cacheCurrent = cacheEntries[i];
				HaltestellenPaar chacheHaltestellenPaar = cacheCurrent.getHaltestellenPaar();
				if (chacheHaltestellenPaar.equals(haltestellenPaar)) {
					if (cacheCurrent.getZeitDerAnforderung() >= zeitDerAnforderung
						&& zeitDerAnforderung >= cacheCurrent.getAnkunftszeit()) {
						return i;
					}
				}
			}
		}
		return -1;
	}

	private static synchronized void init() {
		cacheEntriesForward = new Route[CACHE_SIZE];
		cacheEntriesBackward = new Route[CACHE_SIZE];
		for (int i = 0; i < CACHE_SIZE; i++) {
			cacheEntriesForward[i] = Route.NULL;
			cacheEntriesBackward[i] = Route.NULL;
		}
	}
	
	public static void removeAllEntries() {
		init();
	}
}
