// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package algo;

import java.util.Vector;

import model.DatumZeit;
import model.Haltestelle;
import model.IFahrt;
import model.IFahrtStatic;
import model.Laufweg;
import model.Linie;
import model.Route;
import model.RoutenAbschnitt;
import utils.StringUtils;
import data.Data;
import dev.Logger;

public class RoutenOptimierer {

	private static final long MAX_OPTIMIZATION_TIME = 400; // ms
	private static final long MAX_OPTIMIZATION_LONG_TIME = 3000; // ms
	private static final long MAX_OPTIMIZATION_CECK_INTERVAL = 500; // cycles
	private static final int MAX_OPTIMIZATION_TESTCASE = 100000; // cycles
	private static final boolean DEBUG = false;

	/**
	 * Optimiert die zwei berechneten Routen und ermittelt daraus eine möglichst beste Route. Die Optimierung
	 * wir abgebrochen wenn sie zu lange dauert, ansonsten gibt die das optimale Ergebnis zurück.
	 * 
	 * @param eins
	 * @param zwei
	 * @param optimierungsTyp
	 *            Auswahl aus Algorithmus.OPTIMIERUNG_XXXX
	 *            Wird verwendet um eine Route bei der die Optimierung im
	 *            Übersichtsscreen abgebrochen wurde, vor dem Anzeigen auf dem Details Screen nochmal komplett
	 *            optimiert zu berechnen.
	 * @return
	 */
	public static Route optimiereRoute(RoutenResultHolder eins, RoutenResultHolder zwei, short optimierungsTyp) {
		/*
		 * Wenn die Route fehlerhaft ist können wir nichts optimieren Wenn dieser Fehler auftritt gibt es
		 * entweder einen Fehler im Datensatz oder im Algorithmus. Das gleiche wird unten nochmal geprüft
		 */
		if (!eins.route.isValid()) {
			Logger.error(RoutenOptimierer.class, "FATAL! Route eins ist fehlerhaft."
				+ eins.route.toString());
			return eins.route;
		}

		if (zwei.route.isValid() && eins.route.getDataWrapper() == zwei.route.getDataWrapper()) {
			Logger.error(RoutenOptimierer.class, "FATAL! Übergebene DataWrapper sind gleich!");
			return eins.route;
		}
		
		if (optimierungsTyp == Algorithmus.OPTIMIERUNG_KEINE) {
			Logger.error(Route.class, "kein Optimierungstyp ausgewählt");
			return eins.route;
		}
			
		// Referenzen auf die Werte die wir oft benötigen
		Haltestelle von = eins.route.getHaltestelleVon();
		Haltestelle nach = eins.route.getHaltestelleNach();

		/**
		 * Eine kleine Skizze:<br>
		 * |------------------->| hin<br>
		 * <-|--------------------| zurueck Fehlerfall<br>
		 * |<-------------------| zurueck Zulässiger<br>
		 * | <------------------| zurueck Optimierungsfall<br>
		 * <br>
		 * + Die hin Routen besitzen die Zeiten die wir weiterverwenden, weil der Optimierer nacher vorwärts
		 * die Route berechnet. + FEHLERFALL: Grunsätzlich soll hier der Fehler abgefangen werden wenn die
		 * zweite berechnete Route über die Grenzen (Abfahrts bzw. Ankunftszeit) "hinausschießt". Dieser
		 * Fehler entsteht dadurch das es parallel verlaufende Routen gibt, die zwischen machen Haltestellen
		 * unterschiedliche Fahrzeiten haben. Dadurch verliert der Dijkstra Zeit wenn er eine bestimmte Linie
		 * verwendet und schafft es so evtl. nicht mehr auf eine andere Bahn umzusteigen Beispiel: VVS Netz:
		 * Stuttgart Ostendplatz -> Stuttgart Wangen Marktplatz am 04.11.2008 um 09:50<br>
		 * + OPTIMIERUNGSFALL: Zeiten sind eigentlich schon in Ordnung, die dritte Route wird aber noch
		 * berechnet um die Haltestellen in der Menge der Haltestellen des Optimierer zu verwenden.
		 */
		RoutenResultHolder hin, zurueck;
		RoutenResultHolder drei = null;
		RoutenResultHolder[] routenFuerHaltestellen, routenFuerOptimierung;
		int zeitDerAnforderung;

		// FALL: A
		// Das ist der Fall mit Algorithmus.calcRoute() und Algorithmus.calcLaterRoute()
		if (eins.route.getDataWrapper() instanceof DataWrapperNormal) {
			hin = eins;
			zurueck = zwei; // korrekte Zeiten
			zeitDerAnforderung = hin.route.getZeitDerAnforderung();

			// Fall: A1
			if (!zurueck.route.isValid() || zurueck.route.getAbfahrtszeit() < hin.route.getAbfahrtszeit()) {
				routenFuerOptimierung = new RoutenResultHolder[] { hin };

				if (zurueck.route.isValid()) {
					routenFuerHaltestellen = new RoutenResultHolder[] { hin, zurueck };
				} else {
					routenFuerHaltestellen = new RoutenResultHolder[] { hin };
				}

			// Fall: A2
			} else if (zurueck.route.getAbfahrtszeit() == hin.route.getAbfahrtszeit()) {
				if (zurueck.route.getAnkunftszeit() < hin.route.getAnkunftszeit()) {
					routenFuerOptimierung = new RoutenResultHolder[] { zurueck };
					Logger.info(RoutenOptimierer.class, "Fall A2: Sollte nicht so oft vorkommen.");
				} else {
					routenFuerOptimierung = new RoutenResultHolder[] { hin, zurueck };
				}
				routenFuerHaltestellen = new RoutenResultHolder[] { hin, zurueck };

			// Fall: A3
			// if(zurueck.route.getAbfahrtszeit() > hin.route.getAbfahrtszeit())
			} else {
				drei = Algorithmus.calcDijkstraRoute(von, nach, zurueck.route.getAbfahrtszeit(),
					new DataWrapperNormal(), zurueck.zeitBisHaltestelle, optimierungsTyp);

				// Fall: A31
				if (!drei.route.isValid() || drei.route.getAnkunftszeit() > zurueck.route.getAnkunftszeit()) {
					routenFuerOptimierung = new RoutenResultHolder[] { zurueck };

					if (drei.route.isValid()) {
						routenFuerHaltestellen = new RoutenResultHolder[] { hin, zurueck, drei };
					} else {
						routenFuerHaltestellen = new RoutenResultHolder[] { hin, zurueck };
					}

				// Fall: A32
				} else if (drei.route.getAnkunftszeit() == zurueck.route.getAnkunftszeit()) {

					// Es gibt Sonderfälle in denen die Route drei nochmal
					// eine kürzere Route findet, also später losfährt als die
					// Abfahrtszeit mit der sie berechnet wurde (zurueck Abfahrtszeit).
					if (zurueck.route.getAbfahrtszeit() < drei.route.getAbfahrtszeit()) {
						routenFuerOptimierung = new RoutenResultHolder[] { drei };
						Logger.info(RoutenOptimierer.class, "Fall A32: Es gibt Sonderfälle in denen die "
							+ "Route drei nochmal eine kürzere Route findet, also später losfährt als die "
							+ "Abfahrtszeit mit der sie berechnet wurde (zurueck Abfahrtszeit).");
					} else {
						routenFuerOptimierung = new RoutenResultHolder[] { drei, zurueck };
					}
					routenFuerHaltestellen = new RoutenResultHolder[] { drei, zurueck };

				// Fall: A33
				// drei.route.getAnkunftszeit() < zurueck.route.getAnkunftszeit()
				// Kann nicht vorkommen
				} else {
					// Keine Verwendung der Optimierungstabelle, obwohl sie in zurueck gültig wäre.
					routenFuerOptimierung = new RoutenResultHolder[] { drei };
					routenFuerHaltestellen = new RoutenResultHolder[] { drei, zurueck };
					Logger.info(RoutenOptimierer.class, "Fall A33: Keine Verwendung der Optimierungstabelle, "
						+ "obwohl sie in zurueck gültig wäre.");
				}
			}

		// FALL: B
		// Das ist der Fall mit Algorithmus.calcFormerRoute()
		} else if (eins.route.getDataWrapper() instanceof DataWrapperReverse) {
			hin = zwei; // korrekte Zeiten
			zurueck = eins;
			zeitDerAnforderung = zurueck.route.getAbfahrtszeit();

			// Fall: B1
			if (!hin.route.isValid() || hin.route.getAnkunftszeit() > zurueck.route.getAnkunftszeit()) {
				routenFuerOptimierung = new RoutenResultHolder[] { zurueck };

				if (hin.route.isValid()) {
					routenFuerHaltestellen = new RoutenResultHolder[] { hin, zurueck };
				} else {
					routenFuerHaltestellen = new RoutenResultHolder[] { zurueck };
				}

			// Fall: B2
			} else if (hin.route.getAnkunftszeit() == zurueck.route.getAnkunftszeit()) {
				if (hin.route.getAbfahrtszeit() > zurueck.route.getAbfahrtszeit()) {
					Logger.info(RoutenOptimierer.class, "Fall B2: Sollte nicht so oft vorkommen.");
					routenFuerOptimierung = new RoutenResultHolder[] { hin };
				} else {
					routenFuerOptimierung = new RoutenResultHolder[] { hin, zurueck };
				}
				routenFuerHaltestellen = new RoutenResultHolder[] { hin, zurueck };

			// Fall: B3
			// hin.route.getAnkunftszeit() < zurueck.route.getAnkunftszeit()
			} else {
				drei = Algorithmus.calcDijkstraRoute(nach, von, hin.route.getAnkunftszeit(),
					new DataWrapperReverse(), hin.zeitBisHaltestelle, optimierungsTyp);

				// Fall: B31
				if (!drei.route.isValid() || hin.route.getAbfahrtszeit() < drei.route.getAbfahrtszeit()) {
					routenFuerOptimierung = new RoutenResultHolder[] { hin };

					if (drei.route.isValid()) {
						routenFuerHaltestellen = new RoutenResultHolder[] { hin, zurueck, drei };
					} else {
						routenFuerHaltestellen = new RoutenResultHolder[] { hin, zurueck };
					}

				// Fall: B32
				} else if (drei.route.getAbfahrtszeit() == hin.route.getAbfahrtszeit()) {

					// Es gibt Sonderfälle in denen die Route drei nochmal
					// eine kürzere Route findet, also später losfährt als die
					// Ankunftszeit mit der sie berechnet wurde (hin Ankunftszeit).
					if (drei.route.getAnkunftszeit() < hin.route.getAnkunftszeit()) {
						routenFuerOptimierung = new RoutenResultHolder[] { drei };
						Logger.info(RoutenOptimierer.class, "Fall B32: Es gibt Sonderfälle in denen die "
							+ "Route drei nochmal eine kürzere Route findet, also später losfährt als die "
							+ "Ankunftszeit mit der sie berechnet wurde (hin Ankunftszeit).");
					} else {
						routenFuerOptimierung = new RoutenResultHolder[] { drei, hin };
					}
					routenFuerHaltestellen = new RoutenResultHolder[] { drei, hin };

				// Fall: B33
				// hin.route.getAbfahrtszeit() > drei.route.getAbfahrtszeit()
				// Kann nicht vorkommen
				} else {
					routenFuerOptimierung = new RoutenResultHolder[] { drei };
					routenFuerHaltestellen = new RoutenResultHolder[] { drei, hin };
					Logger.info(RoutenOptimierer.class, "Fall B33: hin.route.getAbfahrtszeit() > drei.route.getAbfahrtszeit() "
						+ "Kann nicht vorkommen!");
				}
			}

		} else {
			Logger.error(RoutenOptimierer.class, "Inkonsitente Routen übergeben.");
			return null;
		}

		// Extrahiere die Haltestellen
		Vector haltestellen = RoutenOptimierer.extrahiereHaltestellen(routenFuerHaltestellen);

		// Bestimme den Initialen-Score
		int lowestQuickScore = Integer.MAX_VALUE;
		for (int iRoute = 0; iRoute < routenFuerOptimierung.length; iRoute++) {
			RoutenResultHolder route = routenFuerOptimierung[iRoute];
			int quickScore = RoutenOptimiererElement.buildRoutenOptioniererElementsOfRoute(route.route)
				.getQuickScore();
			if (quickScore < lowestQuickScore) {
				lowestQuickScore = quickScore;
			}
		}

		// Bestimmte Optimale Abfahrts und Ankunftszeit
		int abfahrtszeitOptimal = routenFuerOptimierung[0].route.getAbfahrtszeit();
		int ankunftszeitOptimal = routenFuerOptimierung[0].route.getAnkunftszeit();
		for (int iRoute = 1; iRoute < routenFuerOptimierung.length; iRoute++) {
			int abfahrtszeit = routenFuerOptimierung[iRoute].route.getAbfahrtszeit();
			int ankunftszeit = routenFuerOptimierung[iRoute].route.getAnkunftszeit();
			if (abfahrtszeit != abfahrtszeitOptimal || ankunftszeit != ankunftszeitOptimal) {
				Logger.error(RoutenOptimierer.class, "An- und Abfahrtszeiten inkonsistent!\n\nEINS\n" + eins + "\n\nZWEI\n" + zwei
					+ "\n\nDREI\n" + drei + "\n\nCURRENT\n" + routenFuerOptimierung[iRoute].route);
				return eins.route;
			}
		}

		/*
		 * Bestimmt den Baum von guten Routen. Bewertet jede Route und ermittelt so die Beste Route.
		 */
		Route[] routen = RoutenOptimierer.bestimmeRouten(von, nach, abfahrtszeitOptimal, ankunftszeitOptimal,
			zeitDerAnforderung, haltestellen, lowestQuickScore, routenFuerOptimierung, optimierungsTyp);

		// Selektiere die Beste Route aus den zurückgegebenen
		Route besteRoute = null;
		int lowestScore = Integer.MAX_VALUE;
		for (int iRouten = 0; iRouten < routen.length; iRouten++) {
			int score = RoutenOptimierer.scoreRoute(routen[iRouten]);
			if (score < lowestScore) {
				lowestScore = score;
				besteRoute = routen[iRouten];
			}
		}

		// Routen von routenFuerOptimierung auch scoren und schauen ob die Beste Route dabei ist
		for (int iRoute = 0; iRoute < routenFuerOptimierung.length; iRoute++) {
			int score = RoutenOptimierer.scoreRoute(routenFuerOptimierung[iRoute].route);
			if (score < lowestScore) {
				lowestScore = score;
				besteRoute = routenFuerOptimierung[iRoute].route;
			}
		}

		if (SHOW_DEBUG_OUTPUT_SCORE_DEVEL) {
			System.out.println("lowestScore: " + lowestScore + "\n");
		}

		// Falls der RoutenOptimierer andere Fahrzeiten ermittelt, als Dijkstra.
		// Falls die Abfahrtszeit früher ist, ist es nicht gut.
		if (abfahrtszeitOptimal > besteRoute.getAbfahrtszeit()) {
			Logger.error(RoutenOptimierer.class, "ERROR: abfahrtszeitOptimal != besteRoute.getAbfahrtszeit()\n\nHIN\n" + hin + "\n\nZURUECK\n"
				+ zurueck + "\n\nBESTEROUTE\n" + besteRoute);
		}
		// Falls Ankunftszeit später als die optimale Ankunftszeit, dann haben wir ein Problem.
		if (ankunftszeitOptimal < besteRoute.getAnkunftszeit()) {
			Logger.error(RoutenOptimierer.class, "ERROR: ankunftszeitOptimal != besteRoute.getAnkunftszeit()\n\nHIN\n" + hin + "\n\nZURUECK\n"
				+ zurueck + "\n\nBESTEROUTE\n" + besteRoute);
		}

		/*
		 * Gebe die beste Route
		 */
		besteRoute.setOptimierungsTyp(optimierungsTyp);
		besteRoute.setOptimierungAbgeschlossen();
		return besteRoute;
	}

	/**
	 * Extrahiert alle Haltestellen aus den übergebenen Routen in einen Vector. Jede Haltestelle ist nur
	 * einmal im Vector enthalten, die Reihenfolge ist zufällig.
	 */
	private static Vector extrahiereHaltestellen(RoutenResultHolder[] routen) {
		Vector haltestellen = new Vector();
		Vector linien = new Vector();
		for (int iRouten = 0; iRouten < routen.length; iRouten++) {
			// alle verwendeten Linien ermitteln (außer Laufwege, da beim Zusammenfassen von Laufwegen Haltestellen verloren gehen)
			RoutenAbschnitt[] routenAbschnitte = routen[iRouten].route.getKompakteRoutenAbschnitte();
			for (int iRoutenAbschnitte = 0; iRoutenAbschnitte < routenAbschnitte.length; iRoutenAbschnitte++) {				
				IFahrt fahrt = routenAbschnitte[iRoutenAbschnitte].getFahrt();
				if (fahrt.isLaufweg()) continue;
				Linie linie = fahrt.getLinie();
				if (!linien.contains(linie)) {
					linien.addElement(linie);
				}
			}

			// alle Haltestellen aus den Laufwegen hinzufügen
			routenAbschnitte = routen[iRouten].route.getRoutenAbschnitte();
			for (int iRoutenAbschnitte = 0; iRoutenAbschnitte < routenAbschnitte.length; iRoutenAbschnitte++) {				
				IFahrt fahrt = routenAbschnitte[iRoutenAbschnitte].getFahrt();
				if (!fahrt.isLaufweg()) continue;
				Haltestelle[] haltestellenOfLaufweg = fahrt.getAlleHaltestellen();
				for (int i = 0; i < haltestellenOfLaufweg.length; i++) {
					Haltestelle haltestelle = haltestellenOfLaufweg[i];
					if (!haltestellen.contains(haltestelle)) {
						haltestellen.addElement(haltestelle);
					}
				}
			}
		}

		// alle Haltestellen aus den verwendeten Linien hinzufügen
		for (int i = 0; i < linien.size(); i++) {
			Linie linie = (Linie)linien.elementAt(i);
			if (linie != Linie.LAUFWEG_LINIE) {
				Haltestelle[] hs = Data.getHaltestellenOfLinie(linie);
				for (int iHaltestelle = 0; iHaltestelle < hs.length; iHaltestelle++) {
					Haltestelle haltestelle = hs[iHaltestelle];
					if (!haltestellen.contains(haltestelle)) {
						haltestellen.addElement(haltestelle);
					}
				}
			}
		}
		
		return haltestellen;
	}

	/**
	 * berechnet alle Routenalternativen zwischen den gegebenen Übergabeparameter falls beim erstellen des
	 * Baums der Score über lowestQuickScore geht wird die Suche an dieser Stelle abgebrochen
	 */
	private static Route[] bestimmeRouten(Haltestelle von, Haltestelle nach, int abfahrtszeit,
		int ankunftszeit, int zeitDerAnforderung, Vector haltestellen, int lowestQuickScore,
		RoutenResultHolder[] routenFuerOptimierung, int optimierungsTyp) {

		// Timer um die Bestimmung der Routen nach oben beschränken zu können.
		long startTime = System.currentTimeMillis();

		// speichert ob die Bestimmung weiterer Routen abgebrochen wurde
		boolean optimierungAbgebrochen = false;

		if (SHOW_DEBUG_OUTPUT_SCORE_DEVEL) {
			System.out.println("Initial-Quickscore: " + lowestQuickScore);
		}

		// Bestimmte welche zeitBisHaltestelle-Tabelle für die Optimierung verwendet werden soll.
		int[] optimierungsTabelle = null;
		for (int iRoute = 0; iRoute < routenFuerOptimierung.length; iRoute++) {
			if (routenFuerOptimierung[iRoute].route.getDataWrapper() instanceof DataWrapperReverse) {
				if (optimierungsTabelle != null) {
					Logger.error(RoutenOptimierer.class, "Mehre Reverse Routen in routenFuerOptimierung.");
				}
				optimierungsTabelle = routenFuerOptimierung[iRoute].zeitBisHaltestelle;
			}
		}

		// Queue: RoutenAbschnittOptimierung
		Vector queue = new Vector();
		Vector zielHaltestellenKnoten = new Vector();

		// Starthaltestelle vorinitialisieren
		queue.addElement(new RoutenOptimiererElement(null, von, null, abfahrtszeit, null));

		int counterCycle = 0;
		int maxQueueSize = 0;
		if (DEBUG) {
			System.out.println("\n\n######### DEBUG START ################################################################");
		}
		while (queue.size() > 0) {
			if (queue.size() > maxQueueSize) {
				maxQueueSize = queue.size();
			}

			// Abbruchkriterium Zeit
			// Wenn das bestimmen der Routen zu lange dauert brechen wir ab. Da es
			// sich um eine Tiefensuche handelt, sollte zummindest eine Route gefunden
			// worden sein. Außerdem haben wir ja noch die zwei davor berechneten Routen.
			counterCycle++;
			if (counterCycle % MAX_OPTIMIZATION_CECK_INTERVAL == 0) {
				if (optimierungsTyp != Algorithmus.OPTIMIERUNG_ENDLOS) {
					if (optimierungsTyp == Algorithmus.OPTIMIERUNG_TESTCASE_VERSION) {
						if ( counterCycle >= MAX_OPTIMIZATION_TESTCASE ) {
							optimierungAbgebrochen = true;
							break;
						}
					} else {						
						long maxTime;
						if (optimierungsTyp == Algorithmus.OPTIMIERUNG_NORMAL) {
							maxTime = MAX_OPTIMIZATION_TIME;
						} else {
							maxTime = MAX_OPTIMIZATION_LONG_TIME;
						}

						if ((System.currentTimeMillis() - startTime) > maxTime) {
							// System.out.println("Optimierung abgebrochen (counter = " + counter + ")");
							optimierungAbgebrochen = true;
							break;
						}
					}
				}
			}

			// <DEBUG>
			if (DEBUG) {
				System.out
					.println("######################################################################################");
				System.out.println("couter = " + counterCycle);
				for (int j = 0; j < queue.size(); j++) {
					Vector a = new Vector();
					RoutenOptimiererElement current3 = (RoutenOptimiererElement) queue.elementAt(j);
					do {
						a.addElement(current3);
						current3 = current3.getVorgaenger();
					} while (current3 != null);

					String out = StringUtils.fillInt(j, '0', 2) + "|";
					for (int i = a.size() - 1; i >= 0; i--) {
						RoutenOptimiererElement e = (RoutenOptimiererElement) a.elementAt(i);
						if (e.getVon() != null) {
							out += "" + DatumZeit.toTimeString(e.getAbfahrtszeitVonHaltestelle()) + " ";
							out += StringUtils.fillStringBack(e.getVon().getFullShortName(), ' ', 9)
								.substring(0, 9);
							out += " -> ";
						}
					}
					System.out.println(out);
				}
			}
			// </DEBUG>

			// Tiefensuche
			RoutenOptimiererElement current = (RoutenOptimiererElement) queue.lastElement();
			queue.removeElement(current);

			Haltestelle currentHaltestelle = current.getNach();
			for (int iFahrten = 0; iFahrten < currentHaltestelle.getAlleFahrten().length; iFahrten++) {
				IFahrt currentFahrt = currentHaltestelle.getAlleFahrten()[iFahrten];

				// Wenn die nächste Haltestelle in haltestellen enthalten ist,
				// dann nächste Fahrt. Wenn nicht wurde die Haltestelle bei keiner
				// der vorherigen Dijkstra-Durchläufe verwendet und wird so auch
				// nicht hier für die optimale Route gebraucht.
				Haltestelle nextHaltestelle = currentFahrt.getNextHaltestelle(currentHaltestelle);
				if (!haltestellen.contains(nextHaltestelle)) {
					continue;
				}

				// Es macht keinen Sinn wieder zurück zu einer Station zu
				// fahren, weil man sonst dort hätte warten können.
				// TODO: Ist dieses Kriterium noch nötig oder reichen die anderen?
				RoutenOptimiererElement current2 = current;
				boolean skip = false;
				do {
					if (current2.getNach().equals(nextHaltestelle)) {
						skip = true;
						break;
					}
					current2 = current2.getVorgaenger();
				} while (current2 != null);
				if (skip) {
					continue;
				}

				// Wenn die nächste Haltstelle in haltestellen enthalten ist, dann wird die Fahrt verwendet.
				Linie lastLinie;
				if (current.getVerwendeteFahrtVonLetzterHaltestelle() == null) {
					lastLinie = null;
				} else {
					lastLinie = current.getVerwendeteFahrtVonLetzterHaltestelle().getLinie();
				}
				int timeNextHaltestelle = currentFahrt.getTimeAtNextHaltestelle(currentHaltestelle, current
					.getAnkunftszeitNachHaltestelle(), lastLinie);

				// Fängt -1 Zeiten ab.
				// Das ist der Fall wenn es keine Möglichkeit gibt mit dieser Fahrt zur nächsten Haltestelle
				// zu fahren
				if (timeNextHaltestelle == -1) {
					continue;
				}

				// Verwendung der Optimierungstabelle
				if (optimierungsTabelle != null
					&& optimierungsTabelle[nextHaltestelle.getId()] < timeNextHaltestelle) {
					continue;
				}

				// Falls wir keine Optimierungstabelle haben, wenden wir hier ein sehr
				// schwaches Kriterium an, was aber auch wirkt.
				if (timeNextHaltestelle > ankunftszeit) {
					continue;
				}

				// Wenn alles ok -> neues RoutenOptimiererElement hinzufügen.
				RoutenOptimiererElement newRoutenOptimiererElement = new RoutenOptimiererElement(current
					.getNach(), nextHaltestelle, currentFahrt, timeNextHaltestelle, current);

				/*
				 * TODO: Kann besser angeordnet werden, score Abfrage erst im else Zweig von nextHaltestelle
				 * == nach und dann auch kicken wenn >=. evtl. kann abgebrochen werden weil der Score schon zu
				 * schlecht ist
				 */
				int newRoutenScore = newRoutenOptimiererElement.getQuickScore();
				if (newRoutenScore > lowestQuickScore) {
					continue;
				}
				if (nextHaltestelle == nach) {
					// Nur hinzufügen wenn <= lowestQuickScrore ??
					zielHaltestellenKnoten.addElement(newRoutenOptimiererElement);

					if (newRoutenScore < lowestQuickScore) {
						lowestQuickScore = newRoutenScore;
						if (SHOW_DEBUG_OUTPUT_SCORE_DEVEL) {
							System.out.println("verbesserter Quickscore: " + lowestQuickScore);
						}
					}
				} else {
					queue.addElement(newRoutenOptimiererElement);
				}
			}
		}

		// Bestimmt Array Größe, d.h schaut welche Routen sich qualifizieren dadurch das sie unter dem
		// lowestQuickScrore liegen.
		Vector resultsVector = new Vector();
		for (int iZieleHaltestellenKnoten = 0; iZieleHaltestellenKnoten < zielHaltestellenKnoten.size(); iZieleHaltestellenKnoten++) {
			RoutenOptimiererElement element = ((RoutenOptimiererElement) zielHaltestellenKnoten
				.elementAt(iZieleHaltestellenKnoten));
			if (element.getQuickScore() == lowestQuickScore) {
				resultsVector.addElement(zielHaltestellenKnoten.elementAt(iZieleHaltestellenKnoten));
			} else if (element.getQuickScore() < lowestQuickScore) {
				Logger.error(RoutenOptimierer.class, "FATAL: e.getQuickScore() < lowestQuickScore");
			}
		}

		// Routen erstellen
		Route[] results = new Route[resultsVector.size()];
		int resultArrayPosition = 0;
		for (int i = 0; i < resultsVector.size(); i++) {
			RoutenOptimiererElement element = ((RoutenOptimiererElement) resultsVector.elementAt(i));
			results[resultArrayPosition] = element.getRoute(zeitDerAnforderung);
			// Schreibt in die Route das die Optimierung abgebrochen wurde.
			if (optimierungAbgebrochen) {
				results[resultArrayPosition].setOptimierungAbgebrochen();
			}
			resultArrayPosition++;
		}

		return results;
	}

	/**
	 * 
	 */
	private static final boolean SHOW_DEBUG_OUTPUT = false;
	private static final boolean SHOW_DEBUG_OUTPUT_SCORE_DEVEL = false;
	private static final int P_LANGER_LAUFWEG_MIN_VALUE = 5;
	private static final int P_LANGER_LAUFWEG_LINEAR_FACTOR = 10;
	private static final int P_LAUFWEG_LINEAR_FACTOR = 1;
	private static final int SCORE_UMSTEIGEN = 200;
	private static final int MAX_PENALTY_ZU_KURZ_UMSTEIGEN = 16; 

	/**
	 * Berechnet den QuickScore der aufgrund des letzten Eintrags (element) hinzukommt berücksichtigt werden
	 * nur - Umstiege - langer Laufweg
	 */
	protected static int calcAdditionalQuickScore(RoutenOptimiererElement element) {
		// erstes Element
		if (element.getVorgaenger() == null) {
			return 0;
		}
		// ab dem zweiten kann ein Score ermittelt werden
		int score = 0;

		IFahrt aktuelleFahrt = element.getVerwendeteFahrtVonLetzterHaltestelle();

		// Umsteigescore
		score += calcUmsteigePenalty(element);
		// langer Laufweg
		if (aktuelleFahrt.isLaufweg()) {
			Laufweg laufweg = (Laufweg) aktuelleFahrt;
			score += calcLangerLaufwegPenalty(laufweg);
		}
		return score;
	}

	/**
	 * Überprüft ob am element umgestiegen werden muss gibt den entsprechenden Penalty zurück
	 */
	private static int calcUmsteigePenalty(RoutenOptimiererElement element) {
		RoutenOptimiererElement vorgaenger = element.getVorgaenger();

		// die ersten beiden Elemente (startElement und erste Fahrt) können
		// keinen Umstieg enthalten
		if (vorgaenger == null || (vorgaenger.getVorgaenger() == null)) {
			return 0;
		}

		Linie vorgaengerLinie = vorgaenger.getVerwendeteFahrtVonLetzterHaltestelle().getLinie();
		IFahrt aktuelleFahrt = element.getVerwendeteFahrtVonLetzterHaltestelle();
		Linie aktuelleLinie = aktuelleFahrt.getLinie();

		if (!aktuelleFahrt.isLaufweg() && (vorgaengerLinie != aktuelleLinie)) {
			return SCORE_UMSTEIGEN;
		} else {
			return 0;
		}
	}

	/**
	 * gibt den entsprechenden Penalty zurück, falls der übergebene Laufweg lang ist
	 * 
	 * @param laufweg
	 * @return
	 */
	private static int calcLangerLaufwegPenalty(Laufweg laufweg) {
		// Lange Laufwege: Ab 5min werden Strafpunkte vergeben
		if (laufweg.getLaufzeit() - P_LANGER_LAUFWEG_MIN_VALUE > 0) {
			return squarePenalty(laufweg.getLaufzeit() - P_LANGER_LAUFWEG_MIN_VALUE,
				P_LANGER_LAUFWEG_LINEAR_FACTOR);
		} else {
			return 0;
		}
	}

	/**
	 * Berechnet den exakten Penalty neben den Kriterien von QuickScore - lange Laufwege - Umstiege werden -
	 * Laufwege - genug Zeit zum Umsteigen bewertet
	 */
	protected static int scoreRoute(Route route) {

		// Ermittle Laufwege
		Vector laufwege = new Vector();
		for (int iRoutenAbschnitte = 0; iRoutenAbschnitte < route.getRoutenAbschnitte().length; iRoutenAbschnitte++) {
			RoutenAbschnitt routenAbschnitt = route.getRoutenAbschnitte()[iRoutenAbschnitte];
			if (routenAbschnitt.isLaufweg()) {
				laufwege.addElement(routenAbschnitt);
			}
		}

		// Laufwege
		int pLangeLaufwege = 0;
		int pLaufwege = 0;
		for (int iLaufwege = 0; iLaufwege < laufwege.size(); iLaufwege++) {
			Laufweg laufweg = (Laufweg) ((RoutenAbschnitt) laufwege.elementAt(iLaufwege)).getFahrt();

			// Lange Laufwege: Ab 5min werden Strafpunkte vergeben
			if (laufweg.getLaufzeit() - P_LANGER_LAUFWEG_MIN_VALUE > 0) {
				pLangeLaufwege += squarePenalty(laufweg.getLaufzeit() - P_LANGER_LAUFWEG_MIN_VALUE,
					P_LANGER_LAUFWEG_LINEAR_FACTOR);
			}

			// Laufweg
			pLaufwege += linearPenalty(laufweg.getLaufzeit(), P_LAUFWEG_LINEAR_FACTOR);
		}

		// Anzahl umsteigen
		int pUmsteigen = route.getAnzUmsteigen() * SCORE_UMSTEIGEN;

		// Genug Zeit zum Umsteigen
		// zu kurze Umsteigezeiten werden bestraft
		int pKlassicherUmstieg_ZuKurzeZeitDifferenz = 0;
		int pUmstiegAufLaufweg_ZuKurzeZeitDifferenz = 0;
		RoutenAbschnitt currentRoutenAbschnitt = null;
		RoutenAbschnitt lastRoutenAbschnitt = route.getKompakteRoutenAbschnitte()[0];
		for (int iRoutenAbschnitte = 1; iRoutenAbschnitte < route.getKompakteRoutenAbschnitte().length; iRoutenAbschnitte++) {
			currentRoutenAbschnitt = route.getKompakteRoutenAbschnitte()[iRoutenAbschnitte];
			
			int umsteigezeit = IFahrtStatic.getUmsteigezeit(lastRoutenAbschnitt.getFahrt().getLinie(),
					currentRoutenAbschnitt.getFahrt().getLinie(), currentRoutenAbschnitt.getVon());
			
			int zeitDifferenz = currentRoutenAbschnitt.getAbfahrtsZeit()
				- lastRoutenAbschnitt.getAnkunftsZeit();

			int freizeit = zeitDifferenz - umsteigezeit;
			
			// Klassischer Umstieg
			if (!lastRoutenAbschnitt.getFahrt().isLaufweg() && !currentRoutenAbschnitt.getFahrt().isLaufweg()) {
				pKlassicherUmstieg_ZuKurzeZeitDifferenz += MAX_PENALTY_ZU_KURZ_UMSTEIGEN / (freizeit + 1);
				// Von Laufweg auf Nicht-Laufweg
			} else if (lastRoutenAbschnitt.getFahrt().isLaufweg() && !currentRoutenAbschnitt.getFahrt().isLaufweg()) {
				pUmstiegAufLaufweg_ZuKurzeZeitDifferenz += MAX_PENALTY_ZU_KURZ_UMSTEIGEN / (freizeit + 1);
			}
			// Von Nicht-Laufweg auf Laufweg wird ignoriert, genauso wie wenn beide Laufwege sind.

			// Schleppointer
			lastRoutenAbschnitt = currentRoutenAbschnitt;
		}

		// Score berechnen
		int penalty = pLangeLaufwege;
		penalty += pLaufwege;
		penalty += pUmsteigen;
		penalty += pKlassicherUmstieg_ZuKurzeZeitDifferenz;
		penalty += pUmstiegAufLaufweg_ZuKurzeZeitDifferenz;

		// Debug Ausgabe
		if (SHOW_DEBUG_OUTPUT) {
			System.out
				.println("\n\n##########################################################################");
			String einruecken = "    ";
			System.out.println(route.toString());
			System.out.println("pLangeLaufwege = " + pLangeLaufwege);
			System.out.println(einruecken + "pLaufwege = " + pLaufwege);
			System.out.println(einruecken + "pUmsteigen = " + pUmsteigen);
			System.out.println(einruecken + "pKlassicherUmstieg_ZuKurzeZeitDifferenz = "
				+ pKlassicherUmstieg_ZuKurzeZeitDifferenz);
			System.out.println(einruecken + "pUmstiegAufLaufweg_ZuKurzeZeitDifferenz = "
				+ pUmstiegAufLaufweg_ZuKurzeZeitDifferenz);
			System.out.println(einruecken + "PENALTY TOTAL = " + penalty);
			System.out
				.println("##########################################################################\n\n");
		}

		return penalty;
	}

	private static int linearPenalty(int baseValue, int factor) {
		return baseValue * factor;
	}

	private static int squarePenalty(int baseValue, int factor) {
		return factor * (baseValue * baseValue);
	}

}
