// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package algo;

import java.util.Vector;

import model.DatumZeit;
import model.Haltestelle;
import model.IFahrt;
import model.Route;
import model.RoutenAbschnitt;
import dev.Logger;

public class RoutenOptimiererElement {

	private final int SCORE_UNDEFINED = -1;

	private final IFahrt verwendeteFahrtVonLetzterHaltestelle;
	private final int ankunftszeitNachHaltestelle;
	private final int abfahrtszeitVonHaltestelle;
	private final RoutenOptimiererElement vorgaenger;

	private final Haltestelle von;
	private final Haltestelle nach;

	private int quickScore;
	private int score;

	public RoutenOptimiererElement(Haltestelle von, Haltestelle nach,
		IFahrt verwendeteFahrtVonLetzterHaltestelle, int ankunftszeitNachHaltestelle,
		RoutenOptimiererElement vorgaenger) {
		// super(von, nach);
		this.verwendeteFahrtVonLetzterHaltestelle = verwendeteFahrtVonLetzterHaltestelle;
		this.ankunftszeitNachHaltestelle = ankunftszeitNachHaltestelle;
		this.vorgaenger = vorgaenger;

		this.von = von;
		this.nach = nach;

		if (vorgaenger == null) {
			this.abfahrtszeitVonHaltestelle = ankunftszeitNachHaltestelle;
			// falls es das erste Element ist wird der QuickScore 0 gesetzt
			quickScore = 0;
		} else {
			this.abfahrtszeitVonHaltestelle = ankunftszeitNachHaltestelle
				- verwendeteFahrtVonLetzterHaltestelle.getTimeToNextHaltestelle(von);
			if(this.abfahrtszeitVonHaltestelle < 0) {
				Logger.error(RoutenOptimiererElement.class, "FATAL: this.abfahrtszeitVonHaltestelle < 0 !!!");
			}
			// falls es das späteres Element ist wird der QuickScore aufaddiert
			quickScore = vorgaenger.getQuickScore() + RoutenOptimierer.calcAdditionalQuickScore(this);
		}

		score = SCORE_UNDEFINED;
	}

	public int getQuickScore() {
		return quickScore;
	}

	public int getScore() {
		if (score == SCORE_UNDEFINED) {
			score = RoutenOptimierer.scoreRoute(getRoute(DatumZeit.NULL));
		}
		return score;
	}

	/**
	 * Gibt die Haltestelle zurueck, von der diese Abfrage gestartet ist (Anfangsort)
	 * 
	 * @return von - Haltestelle, von der diese Abfrage gestartet ist (Anfangsort)
	 */
	public Haltestelle getVon() {
		return this.von;
	}

	/**
	 * Gibt die Haltestelle zurueck, an der diese Abfrage geendet ist (Zielort)
	 * 
	 * @return nach - Haltestelle, an der diese Abfrage geendet ist (Zielort)
	 */
	public Haltestelle getNach() {
		return this.nach;
	}

	public IFahrt getVerwendeteFahrtVonLetzterHaltestelle() {
		return verwendeteFahrtVonLetzterHaltestelle;
	}

	public int getWartezeitVonLetzerHaltestelle() {
		if (vorgaenger == null) {
			return 0;
		}
		return abfahrtszeitVonHaltestelle - vorgaenger.getAnkunftszeitNachHaltestelle();
	}

	public int getAnkunftszeitNachHaltestelle() {
		return ankunftszeitNachHaltestelle;
	}

	public int getAbfahrtszeitVonHaltestelle() {
		return abfahrtszeitVonHaltestelle;
	}

	public RoutenOptimiererElement getVorgaenger() {
		return vorgaenger;
	}

	public boolean hasVorgaenger() {
		return vorgaenger != null;
	}

	/**
	 * Gibt ein komplettes Routenobjekt zurück das aus der Verkettung der RoutenOptimiererElemente erstellt
	 * wurde
	 */
	public Route getRoute(int zeitDerAnforderung) {
		// Stelle die Route zusammen, in dem man sie vom Zielknoten her rückwärts durchgeht.
		Vector routenAbschnitteReverse = new Vector();
		RoutenOptimiererElement current = this;
		while (current.getVorgaenger() != null) {
			RoutenAbschnitt newRoutenAbschnitt = new RoutenAbschnitt(current.getVon(), current.getNach(),
				current.getAbfahrtszeitVonHaltestelle(), current.getAnkunftszeitNachHaltestelle(), current
					.getVerwendeteFahrtVonLetzterHaltestelle());
			routenAbschnitteReverse.addElement(newRoutenAbschnitt);
			current = current.getVorgaenger();
		}

		// RoutenAbschnitte umdrehen und in Array umwandeln
		RoutenAbschnitt[] routenAbschnitte = new RoutenAbschnitt[routenAbschnitteReverse.size()];
		for (int iRoutenAbschnitt = routenAbschnitteReverse.size() - 1; iRoutenAbschnitt >= 0; iRoutenAbschnitt--) {
			routenAbschnitte[iRoutenAbschnitt] = (RoutenAbschnitt) routenAbschnitteReverse
				.elementAt(routenAbschnitteReverse.size() - iRoutenAbschnitt - 1);
		}

		// Route erstellen
		Route resultRoute = new Route(routenAbschnitte, zeitDerAnforderung, null);
		return resultRoute;
	}

	/**
	 * Wandelt ein Routenobjekt in eine Kette von RoutenOptimiererElementen um und gibt das letzte Element
	 * zurück
	 */
	public static RoutenOptimiererElement buildRoutenOptioniererElementsOfRoute(Route route) {
		if (route == null) {
			return null;
		}
		if (route.getErrorCode() != Route.STATUS_KEIN_FEHLER) {
			return null;
		}

		RoutenOptimiererElement currentElement = new RoutenOptimiererElement(null, route.getHaltestelleVon(),
			null, route.getAbfahrtszeit(), null);
		RoutenAbschnitt[] routenAbschnitte = route.getRoutenAbschnitte();
		for (int i = 0; i < routenAbschnitte.length; i++) {
			RoutenAbschnitt routenAbschnitt = routenAbschnitte[i];
			currentElement = new RoutenOptimiererElement(routenAbschnitt.getVon(), routenAbschnitt.getNach(),
				routenAbschnitt.getFahrt(), routenAbschnitt.getAnkunftsZeit(), currentElement);
		}
		return currentElement;
	}

	public String toString() {
		String vonString;
		String nachString;
		if (this.von == null) {
			vonString = "NULL";
		} else {
			vonString = this.von.getFullName();
		}
		if (this.nach == null) {
			nachString = "NULL";
		} else {
			nachString = this.nach.getFullName();
		}
		String linienname = "";
		if (this.verwendeteFahrtVonLetzterHaltestelle != null) {
			linienname = this.verwendeteFahrtVonLetzterHaltestelle.getLinie().getName();
		}
		return vonString + "@" + this.getAbfahrtszeitVonHaltestelle() + " -> " + nachString + "@"
			+ this.getAnkunftszeitNachHaltestelle() + " mit " + linienname + " QS: " + quickScore;
	}

	/**
	 * Stellt den String nicht nur für das aktuelle RoutenOptimiererElement zusammen sondern auch für alle
	 * Vorgänger
	 */
	public String showRec() {
		Vector v = new Vector();
		StringBuffer sb = new StringBuffer(this.toString());
		RoutenOptimiererElement currentElement = this;
		while (currentElement.hasVorgaenger()) {
			currentElement = currentElement.getVorgaenger();
			v.addElement(currentElement);
		}
		for (int i = 0; i < v.size(); i++) {
			currentElement = (RoutenOptimiererElement) v.elementAt(i);
			sb.append("\n" + currentElement.toString());
		}
		return sb.toString();
	}
}
