// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package algo.pricing;

import java.util.Vector;

import model.DatumZeit;
import model.Haltestelle;
import model.Linie;
import model.Route;
import model.RoutenAbschnitt;
import model.Verkehrsmittel;
import utils.DynamicIntMap;
import utils.StringUtils;
import dev.Logger;

/**
 * Bietet Funktionen zur Berechnung des Preises von Fahrten im VVS Netz Netz an.
 * <p>
 * Aufstellung welche Haltestelle in welcher Zone ist: http://vvs.de/download/84.pdf
 * <p>
 * Allgemeine Preisinfos + Regeln für Preisfindung: http://vvs.de/fahrk_tarifzonen.php
 * <p>
 * Preis VVS: http://vvs.de/fahrk_fahrkarten.php "Der Fahrpreis richtet sich nach der Anzahl der Tarifzonen,
 * die befahren werden. Start- und Zielzone zählen mit. Haltestellen auf Tarifzonengrenzen zählen zu der Zone,
 * in der die Fahrt durchgeführt wird."
 */
public class ZonePricing implements IPricing {
	
	/* TODO:
	 * Ein bekanntes Problem bei den Preise ist, dass zB im Bereich Korntal/Ditzingen die Zone 36 nicht
	 * hinzugefügt wird, weil alle Zonen auf dem Weg der S6, die Zone 36 enthalten mindestens noch eine andere
	 * Zone enthalten, so dass die Zone 36 einfach fehlt, und der Preis eine Zone zu billig ist.
	 * Das Problem wird durch hinzufügen fehlender Ringe gelöst, aber nur wenn nicht schon eine andere 3x
	 * Zone davor durchfahren wurde. Ich habe schon diverse Ansätze probiert, aber noch keine Lösung gefunden,
	 * die nicht andere Preise auch angreifen würde.
	 * Die Anz. der Routen bei denen das Problem auftritt ist aber sehr gering.
	 */

	private final Route route;
	private final DynamicIntMap zonenMap = new DynamicIntMap(10);
	private int status = STATUS_NICHT_INITIALISIERT;

	public ZonePricing(Route route) {
		// Status initialisieren
		// Ab hier nur noch überschreiben, wenn der OK Status nicht mehr passend ist!
		this.status = STATUS_OK;
		
		// Prüft die Route
		if (route == null || !route.isValid()) {
			Logger.error(this, "Mit einer fehlerhaften Route sollten keine Preise berechnet werden.");
			this.route = null;
			this.status = STATUS_FEHLERHAFT;
			return;
		}
		this.route = route;
		RoutenAbschnitt[] routenAbschnitte = this.route.getRoutenAbschnitte();
		
		if (routenAbschnitte == null || routenAbschnitte.length == 0) {
			this.status = STATUS_FEHLERHAFT;
			return;
		}
		
		// Sonderfall: Reiner Laufweg
		if (this.route.isReinerLaufweg()) {
			this.status = STATUS_REINER_LAUFWEG;
			return;
		}

		// Sonderfall: Kurzstrecke?
		if (this.isKurzstrecke(routenAbschnitte)) {
			this.status = STATUS_KURZSTRECKE;
			return;
		}
		
		// ZONENBERECHNUNG: Bestimme alle Haltestellen, bei denen die Zonenwahl mehrdeutig ist. Alle anderen
		// werden direkt zu den Zonen hinzugefügt. 
		Vector mehrdeutigeHaltestellen = new Vector();
		
		// Von-Haltestelle hinzufügen
		int vonHaltestelleVerkehrsmittel = routenAbschnitte[0].getFahrt().getLinie().getVerkehrsmittel();
		boolean letzterRoutenAbschnittWarLaufweg = (vonHaltestelleVerkehrsmittel != Verkehrsmittel.LAUFWEG);
		if(letzterRoutenAbschnittWarLaufweg
			&& zonenInHaltestelleSindMehrdeutig_SonstZoneHinzufuegen(routenAbschnitte[0].getVon())) {
				mehrdeutigeHaltestellen.addElement(routenAbschnitte[0].getVon());
		}
		
		// Alle Nach-Haltestellen hinzufügen
		for (int i = 0; i < routenAbschnitte.length; i++) {
			if(routenAbschnitte[i].getFahrt().getLinie().getVerkehrsmittel() == Verkehrsmittel.LAUFWEG) {
				// Dieser RoutenAbschnitt ist EIN Laufweg
				// 1. Sonderfall: LetzterRoutenabschnitt ist Laufweg, dann Nach-Haltestelle ignorieren.
				// 2. Wenn der letzte RoutenAbschnitt auch schon ein Laufweg war, dann braucht diese
				// Haltestelle nicht genommen werden, da von und zu ihr nur gelaufen wird.
				if (i + 1 == routenAbschnitte.length || letzterRoutenAbschnittWarLaufweg) {
					continue;
				}
				letzterRoutenAbschnittWarLaufweg = true;
			} else {
				// Dieser RoutenAbschnitt ist KEIN Laufweg
				letzterRoutenAbschnittWarLaufweg = false;
			}
			
			// Verarbeite diese Haltestelle
			if(zonenInHaltestelleSindMehrdeutig_SonstZoneHinzufuegen(routenAbschnitte[i].getNach())) {
				mehrdeutigeHaltestellen.addElement(routenAbschnitte[i].getNach());
			}
		}
		
		// Bearbeite mehrdeutige Haltestellen
		// Wir versuchen nochmal bei allen Haltestellen bei denen es vorher nicht ging diese einfach
		// hinzuzufügen, da wir erst jetzt alle Zonen aus der vorherigen Runde vollständig drin haben.
		Vector weiterhinMehrdeutigeHaltestellen = new Vector();
		for (int iHaltestelle = 0; iHaltestelle < mehrdeutigeHaltestellen.size(); iHaltestelle++) {
			Haltestelle haltestelle = (Haltestelle) mehrdeutigeHaltestellen.elementAt(iHaltestelle);
			if (zonenInHaltestelleSindMehrdeutig_SonstZoneHinzufuegen(haltestelle)) {
				weiterhinMehrdeutigeHaltestellen.addElement(haltestelle);
			}
		}
		
		// Bearbeite weiterhin mehrdeutige Haltestellen (jetzt müssen wir wirklich auswählen)
		rekursivMeistverwendeteZoneHinzufuegen(weiterhinMehrdeutigeHaltestellen);
		
		// Nacharbeit: Übersprunge Ringe hinzufügen
		this.addMissingRings();
		
		// Nacharbeit: Übersprunge Ringe hinzufügen
		//this.addMissingRingsExtended();
	}
	
	/**
	 * Fügt die Zone hinzu, prüft aber ob sie schon enthalten ist.
	 * @return true, wenn die zone hinzugefügt wurde.
	 */
	private boolean addZoneSafe(int zone) {
		// Prüfen ob zone schon enthalen
		if (zonenMap.lookupPosition(zone) == DynamicIntMap.KEY_NOT_FOUND) {
			zonenMap.add(zone, zone);
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Bestimmt die meistverwendete Zone in der Menge der übergebenen Haltestellen und fügt diese zu den Zonen
	 * hinzu. Alle Haltestellen die nun nicht durch diese Zone abgedeckt werden, kommen in die nächste Runde,
	 * bei der die nächste meistverwendete Zone bestimmt wird. Das Ganze läuft rekursiv, bis es keine
	 * Haltestellen mehr gibt. Falls mehrere Zonen gleich viele "Punkte" haben, wird die Erste verwendet (aus
	 * sicht der Reihenfolge der Haltestellen).
	 */
	private void rekursivMeistverwendeteZoneHinzufuegen(Vector haltestellen) {
		// Nichts zu tun, wir sind fertig.
		if(haltestellen.size() == 0) {
			return;
		}
		
		int mostPopularZone = -1;
		int cntMostPopularZone = 0;
		
		DynamicIntMap hitliste = new DynamicIntMap(haltestellen.size() * 2);
		for (int iHaltestelle = 0; iHaltestelle < haltestellen.size(); iHaltestelle++) {
			Haltestelle haltestelle = (Haltestelle) haltestellen.elementAt(iHaltestelle);
			short[] zonenHaltestelle = getZonenDerHaltestelle(haltestelle);
			
			for (int iZonen = 0; iZonen < zonenHaltestelle.length; iZonen++) {
				int zone = zonenHaltestelle[iZonen];
				int key = hitliste.lookupPosition(zone);
				int cnt = 0;
				if(key != DynamicIntMap.KEY_NOT_FOUND) {
					// Falls die zone schon in der hitliste ist, erhöhe deren Punkte
					cnt = hitliste.get(key);
				}
				// Wichtig!
				cnt = cnt + 1;
				hitliste.replace(zone, cnt);
				
				// Halte die meistverwendete Zone immer aktuell
				if(cnt > cntMostPopularZone) {
					mostPopularZone = zone;
					cntMostPopularZone = cnt;
				}
			}
		}
		
		// Es muss eine Zone geben, sonst ist ein Fehler aufgetreten
		if(mostPopularZone == -1) {
			throw new RuntimeException("Fehler, keine mostPopularZone konnte bestimmt werden!");
		}
		
		// Meistverwendete Zone hinzufügen
		addZoneSafe(mostPopularZone);
		
		// Alle Haltestellen entfernen die durch diese Zone abgedeckt werden, bzw. alle merken und
		// weiterverarbeiten, die nicht abgedeckt werden.
		Vector weiterhinMehrdeutigeHaltestellen = new Vector();
		for (int iHaltestelle = 0; iHaltestelle < haltestellen.size(); iHaltestelle++) {
			Haltestelle haltestelle = (Haltestelle) haltestellen.elementAt(iHaltestelle);
			boolean nichtAbgedeckt = true;
			short[] zonenHaltestelle = getZonenDerHaltestelle(haltestelle);
			for (int iZonen = 0; iZonen < zonenHaltestelle.length; iZonen++) {
				if(zonenHaltestelle[iZonen] == mostPopularZone) {
					nichtAbgedeckt = false;
					break;
				}
			}
			
			// Wenn die neue Zone die aktuelle Haltestelle nicht abdeckt muss sie in die nächste Runde.
			if(nichtAbgedeckt) {
				weiterhinMehrdeutigeHaltestellen.addElement(haltestelle);
			}
		}
		
		// Rekursiver Aufruf
		rekursivMeistverwendeteZoneHinzufuegen(weiterhinMehrdeutigeHaltestellen);
	}
	
	/**
	 * Wenn die Zone einer Haltestelle eindeutig hinzugefügt werden kann, dann füge sie hinzu und gebe false
	 * zurück. Wenn keine Zone eindeutig gewählt werden kann gebe true zurück.
	 * 
	 * @param boolean false wenn die Haltestelle abgearbeitet wurde und keine weitere Arbeit mehr nötig ist.
	 */
	private boolean zonenInHaltestelleSindMehrdeutig_SonstZoneHinzufuegen(Haltestelle haltestelle) {
		short[] zonenHaltestelle = getZonenDerHaltestelle(haltestelle);

		// Nicht VVS (keine VVS Zone)
		if (zonenHaltestelle == null || zonenHaltestelle.length == 0) {
			this.status = STATUS_OK_ZZGL_DB;
			return false;

		// Einfache Auswahl
		} else if (zonenHaltestelle.length == 1) {
			addZoneSafe(zonenHaltestelle[0]);
			return false;

		// Erweiterte Auswahl nötig
		} else {
			// Falls eine Zone schon enthalten ist, nehmen wir diese (fügen also keine weitere Zone hinzu).
			for (int iZonen = 0; iZonen < zonenHaltestelle.length; iZonen++) {
				if (zonenMap.lookupPosition(zonenHaltestelle[iZonen]) != DynamicIntMap.KEY_NOT_FOUND) {
					return false;
				}
			}
			
			// Wenn nicht muss eine komplexere Auswahl erfolgen
			return true;
		}
	}

	/**
	 * Gibt true zurück wenn es sich bei der Route um eine Kurzstrecke handelt.
	 */
	private boolean isKurzstrecke(RoutenAbschnitt[] routenAbschnitte) {
		/*
		 * Kurzstrecke: "Für eine Fahrt bis zur 3. Haltestelle nach dem Einstieg ohne Umsteigen und
		 * Fahrtunterbrechung. Ausgenommene Relationen s. Tarifinfo an der Haltestelle." "KurzstreckenTicket
		 * nicht gültig in S-Bahnen, Nahverkehrszügen (DB, WEG) sowie Nachtbussen." Also nur Bus und U-Bahn
		 * (und Laufwege)!
		 */
		boolean isKurzstrecke = true;
		int erlaubteHaltestellen = 3;
		Linie lastLinie = routenAbschnitte[0].getFahrt().getLinie();
		int lastVerkehrsmittel = lastLinie.getVerkehrsmittel();
		for (int i = 0; i < routenAbschnitte.length; i++) {
			Linie currentLinie = routenAbschnitte[i].getFahrt().getLinie();
			int currentVerkehrsmittel = currentLinie.getVerkehrsmittel();

			// Nur erlaubte Verkehrsmittel
			if (currentVerkehrsmittel != Verkehrsmittel.BUS && currentVerkehrsmittel != Verkehrsmittel.UBAHN
				&& currentVerkehrsmittel != Verkehrsmittel.LAUFWEG) {
				isKurzstrecke = false;
				break;
			}

			// Laufwege sind nur am Anfang und am Ende erlaubt
			// (Laufwege in der Mitte der Route würden zur Fahrtunterbrechung
			// bzw. umsteigen führen, was nicht erlaubt ist.)
			if (currentVerkehrsmittel == Verkehrsmittel.LAUFWEG && i != 0 && i != routenAbschnitte.length - 1) {
				isKurzstrecke = false;
				break;
			}

			// Kein Linienwechsel, außer von oder auf Laufweg
			// Fahrtmittelwechsel reicht nicht aus, weil man auch nicht zwei
			// Busse verwenden darf!
			// Laufwege in der Mitte der Route werden schon oben abgefangen
			if (currentLinie.getId() != lastLinie.getId() && currentVerkehrsmittel != Verkehrsmittel.LAUFWEG
				&& lastVerkehrsmittel != Verkehrsmittel.LAUFWEG) {
				isKurzstrecke = false;
				break;
			}

			// Fall der aktuelle RoutenAbschnitt kein Laufweg ist, zählen wir
			// ihn als eine angefahrere Haltestelle
			if (currentVerkehrsmittel != Verkehrsmittel.LAUFWEG) {
				erlaubteHaltestellen--;
			}

			// Zu viele Haltestellen angefahren
			if (erlaubteHaltestellen < 0) {
				isKurzstrecke = false;
				break;
			}

			lastLinie = currentLinie;
			lastVerkehrsmittel = currentVerkehrsmittel;
		}
		return isKurzstrecke;
	}
	
	/**
	 * Überprüft die berechneten Tarif-Ringe und prüft ob welche fehlen, also übersprungen wurden.<br>
	 * Das kann passieren wenn zwei Bahnhöfe der Route jeweils auf der Grenze einer Zone liegen, so dass der
	 * Zonen-Berechnungs-Optimierer die letzte Zone auslässt. 
	 */
	private void addMissingRings() {
		// Markiert welche Ringe genutzt werden. (Array Position 0 wird nicht verwendet)
		// Der VVS hat 7 Ringe
		boolean[] ringGenutzt = new boolean[] { true, false, false, false, false, false, false, false };

		// Markiert die genutzten Ringe
		int[] zonen = this.getZonen();
		for (int i = 0; i < zonen.length; i++) {
			int zone = zonen[i];
			int ring = zone / 10;
			ringGenutzt[ring] = true;
		}

		// Findet "Ringsprünge" und fügt die entsprechenden Ringe hinzu
		int letzterBenutzerRing = -1;
		for (int i = 1; i < ringGenutzt.length; i++) {
			if (letzterBenutzerRing != -1 && (letzterBenutzerRing + 1) < i && ringGenutzt[i]) {
				// Fügt eine besondere Zone ein, die nur angibt welcher Ring hinzugefügt wurde.
				// Es kann vorkommen das mehrere Zonen übersprungen werden!
				// 1003 -> Ring 3
				for(int add = i - 1; add > letzterBenutzerRing; add--) {
					int addZone = 1000 + add;
					addZoneSafe(addZone);
				}
			}
			if (ringGenutzt[i]) {
				letzterBenutzerRing = i;
			}
		}
	}

	/**
	 * Wandelt die Zonen so wie sie in der Haltestelle gespeichert werden in ein short Array um.<br>
	 * short[0] falls ein Fehler aufgetreten ist, oder die Haltestelle nicht einer Zone zugeordnet werden
	 * kann.
	 */
	private short[] getZonenDerHaltestelle(Haltestelle h) {
		String z = h.getTarifZone();
		if (z == null || z.length() == 0) {
			return new short[0];
		}
		String[] zs = StringUtils.split(z, "/");
		short[] ret = new short[zs.length];
		for (int i = 0; i < zs.length; i++) {
			String cur = StringUtils.trimToNumber(zs[i]);
			try {
				ret[i] = Short.parseShort(cur);
			} catch (NumberFormatException e) {
				Logger.errorException(this, e, "Ss!x8");
				return new short[0];
			}
		}
		return ret;
	}

	/**
	 * Gibt die Preisinformationen für einen Erwachsenen zurück.
	 * 
	 * @return String ohne Währungssysmbol
	 */
	public String getPreisErwachsener() {
		// Reihenfolge beachten!
		// Fehlerhaft
		if (this.status <= 0) {
			return "-1";
		}
		
 		// Kurzstrecke
		if(this.status == STATUS_KURZSTRECKE) {
			return "1,00";
		}

		// Reiner Laufweg
		if(this.status == STATUS_REINER_LAUFWEG) {
			return "-1";
		}

		// Die Preise wechseln nicht mit dem Fahrplan sondern mit dem Jahr, danke VVS :-)
		if(DatumZeit.getYear(route.getAbfahrtszeit()) <= 2010) {
			// Bis Ende 2010
			switch (this.getAnzahlZonen()) {
			case 1:
				return "2,00";
			case 2:
				return "2,45";
			case 3:
				return "3,30";
			case 4:
				return "4,40";
			case 5:
				return "5,65";
			default:
				// 6 und mehr
				return "6,65";
			}
		
		} else {
			// Ab 2011 gelten die neuen, höheren Preise
			switch (this.getAnzahlZonen()) {
			case 1:
				return "2,00";
			case 2:
				return "2,50";
			case 3:
				return "3,40";
			case 4:
				return "4,50";
			case 5:
				return "5,75";
			default:
				// 6 und mehr
				return "6,75";
			}
		}
	}
	
	/**
	 * Gibt die die kompletten Preis-Informationen, so wie sie in der GUI angezeigt werden sollen.
	 */
	public String getInfoText() {
		// Fehlerhaft: Nichts ausgeben
		if (this.status == STATUS_FEHLERHAFT || getPreisErwachsener().equals("-1")) {
			return "Preis nicht verfügbar.";
		}

		// Sonderfall: Laufweg
		if (this.status == STATUS_REINER_LAUFWEG) {
			return "Reiner Laufweg";
		}

		// Sonderfall: Nur DBRegio
		if (this.status == STATUS_OK_ZZGL_DB && this.getAnzahlZonen() == 0) {
			return "DB Regio";
		}

		// Sonderfall: Kurzstrecke
		String ret = "";
		if (this.status == STATUS_KURZSTRECKE) {
			ret += "Kurzstrecke";
			return ret;
		}

		// Normalfall: Zonen ausgeben
		ret += "" + this.getAnzahlZonen() + " ";
		if (this.getAnzahlZonen() == 1) {
			ret += "Zone";
		} else {
			ret += "Zonen";
		}

		// DB Region Info hinzufügen, falls eine Haltestelle außerhalb des Tarifgebietes liegt
		if (this.status == STATUS_OK_ZZGL_DB) {
			ret += " zzgl. DB Regio";
		}
		
		return ret;
	}

	/**
	 * Gibt die Zonen zurück, welche von der Route verwendet werden.
	 * 
	 * @return int[]
	 */
	public int[] getZonen() {
		int[] ret = new int[zonenMap.getSize()];
		for (int i = 0; i < zonenMap.getSize(); i++) {
			ret[i] = zonenMap.get(zonenMap.lookupPosition(zonenMap.getKeySet()[i]));
		}
		return ret;
	}

	/**
	 * Gibt zurück durch wie viele Zonen diese Route fährt.
	 */
	public int getAnzahlZonen() {
		return this.zonenMap.getSize();
	}
	
	/**
	 * Gibt den Status dieser Preisberechnung zurück. Das ist eine der in IPricing genannten Konstanten.
	 */
	public int getStatus() {
		return this.status;
	}
	
	public String toString() {
		return "PREIS " + getPreisErwachsener() + "€ (Zonen: " + this.zonenMap + ", " + getInfoText() + ", Status=" + getStatus() + ")";
	}
}
