// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package control;

import gui.FontMetrics;
import gui.Imager;
import gui.Manager;
import gui.launch.GameCanvasInstance;
import gui.utils.GfxUtils;

import javax.microedition.lcdui.Font;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.game.GameCanvas;

import utils.StringUtils;
import data.persistence.RemoteDataBinary;
import data.persistence.UserSettingsInt;
import data.persistence.UserSettingsString;
import dev.Logger;

/**
 * Diese Klasse liefert Informationen zur Hardware und Funktionalität des Geräts
 * auf dem push&ride läuft.
 */
public class Device {
	public static final short GFX_QUALITY_HIGH = 0;
	public static final short GFX_QUALITY_MID = 10;
	public static final short GFX_QUALITY_LOW = 20;
	
	public static final boolean isAndroid;

	// /////////////////////////////////////////////////////////////////////////////////////////////////
	// SYSTEM PROPERTIES
	// /////////////////////////////////////////////////////////////////////////////////////////////////

	private static final boolean softwareKeyboardSuggesten; 

	private static int screenWidth = 0;
	private static int screenHeight = 0;
	private static int screenColorDepth = 0;

	public static void initScreen(int width, int height, int depth) {
		Device.screenWidth = width;
		Device.screenHeight = height;
		Device.screenColorDepth = depth;
	}

	static {
		String platform = getPlatform().toLowerCase();
		// SonyEricssonW890i/R1EA030 & SonyEricssonW910i/R1EA030 - basierend auf Useranfragen
		if (StringUtils.startsWith("sonyericssonw890i", platform)
				|| StringUtils.startsWith("sonyericssonw910i", platform) 
				|| Global.getDownloadUrl().startsWith("z.")) {
			GameCanvasInstance.setRequiresFlushGraphics();
		}
//		isAndroid = true;
		if (StringUtils.startsWith("microemulator-android", platform) || Global.IS_FAKE_ANDROID) {
			isAndroid = true;
		} else {
			isAndroid = false;
		}
		if(Global.ISMIDLET) {
			softwareKeyboardSuggesten = GameCanvasInstance.getInstance().hasPointerEvents();
		} else {
			softwareKeyboardSuggesten = false;
		}
	}
	
	/**
	 * aktiviert das SoftwareKeyboard, falls es nicht vorher vom User manuell
	 * deaktiviert wurde
	 */
	public static void checkForSoftwareKeyboard() {
		if (forceSoftwareKeyboardHide()) {
			return;
		}
		if (!showSoftwareKeyboard()) {
			setShowSoftwareKeyboard(true);
			Logger.info(Device.class, "enable software keyboard by touch");
		}
	}

	/**
	 * Gibt die Plattform des aktuellen Gerätes zurück. Leerer String falls es
	 * einen Fehler gab, aber die Plattform nicht gelesen werden konnte.
	 */
	public static String getPlatform() {
		String p;
		if (Global.ISMIDLET) {
			p = getUserSettingStringHelper(UserSettingsString.SETTINGS_ID_PLATFORM, "microedition.platform");
		} else {
			p = "JAVA -> NoMidlet -> Global.ISMIDLET=false";
		}
		return (p != null) ? p : "";
	}

	public static boolean hasMessaging() {
		String messaging = getUserSettingStringHelper(UserSettingsString.SETTINGS_ID_MESSAGING, "wireless.messaging.version");
		return messaging != null;
	}

	/*
	Diese Methoden werden zur Zeit nicht gebraucht, können aber später aktiviert werden. 
	public static boolean hasBluetooth() {
		return getUserSettingStringHelper(UserSettingsString.SETTINGS_ID_BLUETOOTH, "bluetooth.api.version") != null;
	}*/

	/*public static boolean hasObex() {
		return getUserSettingStringHelper(UserSettingsString.SETTINGS_ID_OBEX, "obex.api.version") != null;
	}*/

	/*public static boolean hasInquiryScan() {
		return getUserSettingStringHelper(UserSettingsString.SETTINGS_ID_INQUIRY_SCAN,
			"bluetooth.connected.inquiry.scan") != null;
	}*/

	private static String getUserSettingStringHelper(int id, String property) {
		String value = UserSettingsString.get(id);
		// Wenn es kein Midlet ist, sind keine sinnvollen Werte zu erwarten.
		if (value == null || !Global.ISMIDLET) {
			return null;
		}
		if (value.equals(UserSettingsString.NULL_VALUE)) {
			value = System.getProperty(property);
			UserSettingsString.set(id, value);
		}
		return value;
	}

	// /////////////////////////////////////////////////////////////////////////////////////////////////
	// DEVICE SETTINGS
	// /////////////////////////////////////////////////////////////////////////////////////////////////

	private static boolean firstInfo = true;

	public static int getGfxQuality() {
		// Global kann alle überschreiben
		if (Global.OVERRIDE_GFX_SET_LOW_QUALITY) {
			if (firstInfo) {
				Logger.warn(Device.class, "Override Gfx Low Quality");
			}
			firstInfo = false;
			return GFX_QUALITY_LOW;
		}

		// Wert laden und gegen Null-Value überprüfen.
		// Falls null, lade default Wert.
		int value = UserSettingsInt.get(UserSettingsInt.SETTINGS_ID_GFX_QUALITY);
		if (value == UserSettingsInt.NULL_VALUE) {
			value = gfxQualitySuggestion();
			UserSettingsInt.set(UserSettingsInt.SETTINGS_ID_GFX_QUALITY, value);
		}

		return value;
	}

	public static void setGfxQuality(short gfxQuality) {
		switch (gfxQuality) {
		case GFX_QUALITY_HIGH:
		case GFX_QUALITY_LOW:
		case GFX_QUALITY_MID:
			break;
		default:
			Logger.error(Device.class, "GFX_QUALITY_DOESNT_EXIST");
			gfxQuality = GFX_QUALITY_LOW;
		}

		UserSettingsInt.set(UserSettingsInt.SETTINGS_ID_GFX_QUALITY, gfxQuality);
	}

	/**
	 * forceSoftwareKeyboardHide ist true, falls vom User die Softwaretastatur
	 * manuell deaktiviert wurde und kann nicht mehr automatisch aktiviert
	 * werden
	 */
	private static boolean forceSoftwareKeyboardHide() {
		int value = UserSettingsInt.get(UserSettingsInt.SETTINGS_ID_HIDDENSOFTWAREKEYBOARD);
		if (value == UserSettingsInt.NULL_VALUE || (value != 1 && value != 0)) {
			value = 0;
			UserSettingsInt.set(UserSettingsInt.SETTINGS_ID_HIDDENSOFTWAREKEYBOARD, value);
		}

		return (value == 1) ? true : false;
	}

	public static boolean showSoftwareKeyboard() {
		// Wert laden und gegen Null-Value überprüfen
		// Falls null lade default Wert
		int value = UserSettingsInt.get(UserSettingsInt.SETTINGS_ID_SOFTWAREKEYBOARD);
		if (value == UserSettingsInt.NULL_VALUE || (value != 1 && value != 0)) {
			value = softwareKeyboardSuggestion() ? 1 : 0;
			UserSettingsInt.set(UserSettingsInt.SETTINGS_ID_SOFTWAREKEYBOARD, value);
		}

		return (value == 1) ? true : false;
	}

	public static void setShowSoftwareKeyboard(boolean newState) {
		UserSettingsInt.set(UserSettingsInt.SETTINGS_ID_SOFTWAREKEYBOARD, (newState) ? 1 : 0);
		if (newState == false) {
			setForceSoftwareKeyboardHide();
		}
	}

	private static void setForceSoftwareKeyboardHide() {
		UserSettingsInt.set(UserSettingsInt.SETTINGS_ID_HIDDENSOFTWAREKEYBOARD, 1);
	}

	private static short gfxQualitySuggestion() {
		// leider kann sich getRuntime().totalMemory() während der Laufzeit ändern - daher ist default die Mid-Quality
		String lowercasePlatform = getPlatform().toLowerCase();

		// override
		String url = Global.getBinaryInfo()[Global.BINARY_INFO_URL];
		if (url.startsWith("d.") || url.startsWith("g.") || url.startsWith("z.")) {
			Logger.info(Device.class, "set Gfx low wegen URL");
			return GFX_QUALITY_LOW;
		}

		if (lowercasePlatform.equals(("Jbed").toLowerCase())) {
			return GFX_QUALITY_MID;
		}

		if (Device.isAndroid) {
			return GFX_QUALITY_MID;
		}

		// check memory
		long totalMemeory = Pushandride.STARTUP_TOTAL_MEMORY;

		long memHigh = suggestTotalMemoryHigh();
		if (totalMemeory >= memHigh) {
			return GFX_QUALITY_HIGH;
		}

		// default
		return GFX_QUALITY_MID;
	}

	public static long suggestBufferMemory() {
		if (screenWidth == 0) {
			Logger.fatal(Device.class, "no Screen Info");
			return 0;
		}
		int mem = screenWidth * screenHeight * screenColorDepth;
		mem *= 2; // 2 Buffers
		return mem;
	}

	public static long suggestPictureMemory() {
		if (screenWidth == 0) {
			Logger.fatal(Device.class, "no Screen Info");
			return 0;
		}
		int mem = screenWidth * screenHeight * screenColorDepth;
		return mem;
	}

	public static long suggestDataSetMemory() {
		return 800000 + 400000;
	}

	public static long suggestTotalMemoryHigh() {
		return suggestBufferMemory() + suggestPictureMemory() + suggestDataSetMemory();
	}

	public static long suggestTotalMemoryMid() {
		return suggestPictureMemory() + suggestDataSetMemory();
	}

	public static long suggestTotalMemoryLow() {
		return suggestDataSetMemory();
	}

	/**
	 * Devices, welche hier true zurueckliefern, werden haeufig mit einem Stick
	 * bedient und benoetigen folglich eine Software-Tastatur Vorsicht: Nicht
	 * alle zurueckgegebenen Geraete werden mit dem Stick bedient!
	 * 
	 * @return true wenn das Software Keyboard verwendet werden soll.
	 */
	private static boolean softwareKeyboardSuggestion() {
		return softwareKeyboardSuggesten;
	}

	/**
	 * set timeOffset in push&ride time (minutes)
	 */
	public static void setTimeOffset(int timeOffset) {
		UserSettingsInt.set(UserSettingsInt.SETTINGS_TIME_OFFSET, timeOffset);
	}

	/**
	 * return timeOffset in push&ride time (minutes)
	 */
	public static int getTimeOffset() {
		// Wert laden und gegen Null-Value überprüfen.
		// Falls null, lade default Wert.
		if (Global.ISMIDLET) {
			int value = UserSettingsInt.get(UserSettingsInt.SETTINGS_TIME_OFFSET);
			if (value == UserSettingsInt.NULL_VALUE) {
				value = timeOffsetSuggestion();
				UserSettingsInt.set(UserSettingsInt.SETTINGS_TIME_OFFSET, value);
			}
			return value;
		} else {
			return 0;
		}
	}

	private static int timeOffsetSuggestion() {
		String lowercasePlatform = getPlatform().toLowerCase();
		if (StringUtils.startsWith("HTC Touch".toLowerCase(), lowercasePlatform) || StringUtils.startsWith("sunmicrosystems_wtk", lowercasePlatform)) {
			return 60;
		} else {
			return 0;
		}
	}

	// /////////////////////////////////////////////////////////////////////////////////////////////////
	// printDeviceInfo()
	// /////////////////////////////////////////////////////////////////////////////////////////////////

	public static String printDeviceInfo() {
		String out = "";
		out += "getPlatform=" + Device.getPlatform();
		out += "hasMessaging=" + Device.hasMessaging();
		out += "isGfxHighQuality=" + Device.getGfxQuality();
		out += "showSoftwareKeyboard=" + Device.showSoftwareKeyboard();
		return out;
	}

	synchronized public static void drawInfoTo(Graphics g, GameCanvas gc) {
		Font font = Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, Font.SIZE_SMALL);
		g.setFont(font);
		int fontHeight = font.getHeight();

		// clear background
		g.setGrayScale(255);
		g.fillRect(0, 0, gc.getWidth(), gc.getHeight());

		g.setGrayScale(0);

		int y = 0;

		int xMem = font.stringWidth("Total Memory: " + StringUtils.addDots(Runtime.getRuntime().totalMemory()));

		g.drawString("Info-Screen", 0, y, Graphics.LEFT | Graphics.TOP);
		y += fontHeight;

		y += fontHeight / 3;

		g.drawString("Platform: " + Device.getPlatform(), 0, y, Graphics.LEFT | Graphics.TOP);
		y += fontHeight;
		g.drawString("Total Memory:", 0, y, Graphics.LEFT | Graphics.TOP);
		g.drawString(StringUtils.addDots(Runtime.getRuntime().totalMemory()), xMem, y, Graphics.RIGHT | Graphics.TOP);
		y += fontHeight;
		g.drawString("Free Memory:", 0, y, Graphics.LEFT | Graphics.TOP);
		g.drawString(StringUtils.addDots(Runtime.getRuntime().freeMemory()), xMem, y, Graphics.RIGHT | Graphics.TOP);
		y += fontHeight;
		g.drawString("Total Start:", 0, y, Graphics.LEFT | Graphics.TOP);
		g.drawString(StringUtils.addDots(Pushandride.STARTUP_TOTAL_MEMORY), xMem, y, Graphics.RIGHT | Graphics.TOP);
		y += fontHeight;
		g.drawString("Free Start:", 0, y, Graphics.LEFT | Graphics.TOP);
		g.drawString(StringUtils.addDots(Pushandride.STARTUP_FREE_MEMORY), xMem, y, Graphics.RIGHT | Graphics.TOP);
		y += fontHeight;

		y += fontHeight / 3;

		g.drawString("MemHigh:", 0, y, Graphics.LEFT | Graphics.TOP);
		g.drawString(StringUtils.addDots(Device.suggestTotalMemoryHigh()), xMem, y, Graphics.RIGHT | Graphics.TOP);
		y += fontHeight;
		g.drawString("MemMid:", 0, y, Graphics.LEFT | Graphics.TOP);
		g.drawString(StringUtils.addDots(Device.suggestTotalMemoryMid()), xMem, y, Graphics.RIGHT | Graphics.TOP);
		y += fontHeight;
		g.drawString("MemLow:", 0, y, Graphics.LEFT | Graphics.TOP);
		g.drawString(StringUtils.addDots(Device.suggestTotalMemoryLow()), xMem, y, Graphics.RIGHT | Graphics.TOP);
		y += fontHeight;

		y += fontHeight / 3;

		int width = gc.getWidth();
		g.drawString("width:", 0, y, Graphics.LEFT | Graphics.TOP);
		g.drawString(StringUtils.addDots(width), xMem * 10 / 13, y, Graphics.RIGHT | Graphics.TOP);
		y += fontHeight;
		int height = gc.getHeight();
		g.drawString("height:", 0, y, Graphics.LEFT | Graphics.TOP);
		g.drawString(StringUtils.addDots(height), xMem * 10 / 13, y, Graphics.RIGHT | Graphics.TOP);
		y += fontHeight;
		int pixel = width * height;
		g.drawString("pixel:", 0, y, Graphics.LEFT | Graphics.TOP);
		g.drawString(StringUtils.addDots(pixel), xMem * 10 / 13, y, Graphics.RIGHT | Graphics.TOP);
		y += fontHeight;

		y += fontHeight / 3;

		int i = gfxQualitySuggestion();
		String quality;
		switch (i) {
		case 0:
			quality = "high";
			break;
		case 10:
			quality = "mid";
			break;
		case 20:
			quality = "low";
			break;
		default:
			quality = "unbekannt";
		}

		g.drawString("sug. quality: " + quality, 0, y, Graphics.LEFT | Graphics.TOP);
		y += fontHeight;

		gc.flushGraphics();
		gc.repaint();
	}

	public static boolean increaseHeightForTouch() {
		int value = UserSettingsInt.get(UserSettingsInt.SETTINGS_ID_TOUCH_SIZE);
		if (value == UserSettingsInt.NULL_VALUE || (value != 1 && value != 0)) {
			if (showSoftwareKeyboard()) {
				int extendedHeight = GfxUtils.getHeightForTouch(FontMetrics.getHeight(Manager.ITEMS_FONT_SIZE));
				value = 1;
				UserSettingsInt.set(UserSettingsInt.SETTINGS_ID_TOUCH_SIZE, value);
				int keyHeight = Imager.getKeyImage().getHeight();
				int freeHeight = Device.screenHeight - Manager.getHeaderHeight() - Manager.getFooterHeight() - keyHeight - 4 * extendedHeight;
				if (freeHeight <= 0) {
					Logger.info(Device.class, "override IncreaseHeightForTouch (disable)");
					value = 0;
					Imager.clearKeyImage();
				}
			} else {
				value = 0;
			}
			UserSettingsInt.set(UserSettingsInt.SETTINGS_ID_TOUCH_SIZE, value);
		}

		return (value == 1) ? true : false;
	}

	public static void setTouchSize(boolean newState) {
		UserSettingsInt.set(UserSettingsInt.SETTINGS_ID_TOUCH_SIZE, (newState) ? 1 : 0);
	}

	private static String uniqueId = null;
	public static final String UNIQUE_ID_UNKNOWN = "UNKNOWN";
	
	/**
	 * gibt die uniqueId des Device zurück (wird vergeben vom Server beim Erstellen eines individuellen Datensatzes)
	 */
	public static String getUniqueId() {
		if (uniqueId == null) {
			uniqueId = RemoteDataBinary.getUniqueIdFromRecordStore();
		}
		return uniqueId;
	}

	/**
	 * setzt die uniqueId des Device (wird vergeben vom Server beim Erstellen eines individuellen Datensatzes)
	 */
	public static void setUniqueId(String uniqueId) {
		if (uniqueId.equals(UNIQUE_ID_UNKNOWN)) {
			Logger.warn(Device.class, "Unique ID sollte nicht UNKNOWN sein");
		} else {
			if (!getUniqueId().equals(UNIQUE_ID_UNKNOWN) && !getUniqueId().equals(uniqueId)) {
				Logger.warn(Device.class, "Unique ID sollte sich nicht ändern");
			} 
		}
		
		Device.uniqueId = uniqueId;
		RemoteDataBinary.saveUniqueIdInRecordStore(uniqueId);
	}

}
