// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package control.connections.bluetooth;

import java.util.Timer;
import java.util.TimerTask;

import javax.bluetooth.RemoteDevice;

import dev.Logger;

public class BluetoothManager {
	protected static final boolean SHOW_WATCHDOG_STATE = false;

	public BluetoothManager(IBluetoothSearchDevicesCallback searchDeviceCallback, IBluetoothSendFileViaObexCallback sendFileViaObexCallback, IBluetoothStateCallback bluetoothStateCallback) {
		this.searchDeviceCallback = searchDeviceCallback;
		this.sendFileViaObexCallback = sendFileViaObexCallback;
		this.stateCallback = bluetoothStateCallback;
	}

	public static final int STATE_IDLE = 0;
	public static final int STATE_SEARCH_FOR_DEVICES = 1;
	public static final int STATE_QUERY_NAMES = 2;
	public static final int STATE_SEND_JAR = 3;
	public static final int STATE_SEND_JAD = 4;
	
	public static final int SEND_STATE_NONE = 0;
	public static final int SEND_STATE_SEARCH_OBEX = 1;
	public static final int SEND_STATE_CONNECTING = 2;
	public static final int SEND_STATE_SENDING = 3;
	
	public static final int FAILURE_BLUETOOTH_NOT_WORKING = 0;
	public static final int FAILURE_BLUETOOTH_EXCEPTION = 1;
	public static final int FAILURE_BLUETOOTH_NO_OBEX = 2;
	public static final int FAILURE_EXCEPTION = 3;
	public static final int FAILURE_TIMEOUT = 4;

	private int state = STATE_IDLE;
	private int sendState = SEND_STATE_NONE;

	private boolean errorstate = false;

	private BluetoothSearchDevices bc = null;
	private BluetoothSendFileViaObex bsfvo = null;
	private final IBluetoothSearchDevicesCallback searchDeviceCallback;
	private final IBluetoothSendFileViaObexCallback sendFileViaObexCallback;
	private final IBluetoothStateCallback stateCallback;

	private BluetoothManager getBluetoothManager() {
		return this;
	}

	public void searchForDevices() {
		startBluetoothAction(STATE_SEARCH_FOR_DEVICES);

		Thread queryNamesThread = new Thread(new Runnable() {
			public void run() {
				try {
					bc = new BluetoothSearchDevices(getBluetoothManager());
					bc.searchForDevices();
				} catch (Exception e) {
					failure(getClass(), e, " Thread - searchForDevices", FAILURE_BLUETOOTH_NOT_WORKING);
				}
			}
		});
		queryNamesThread.start();
	}

	private void queryNames(final RemoteDevice[] remoteDevices) {
		startBluetoothAction(STATE_QUERY_NAMES);

		Thread queryNamesThread = new Thread(new Runnable() {
			public void run() {
				try {
					for (int i = 0; i < remoteDevices.length; i++) {
						String name = remoteDevices[i].getFriendlyName(true);
						getBluetoothManager().foundName(remoteDevices[i], name);
					}
				} catch (Exception e) {
					failure(getClass(), e, " Thread - queryNames", FAILURE_EXCEPTION);
					return;
				}
				getBluetoothManager().finishedQueryNames();
			}
		});
		queryNamesThread.start();
	}

	public static long getSizeOfJar(BluetoothManager bluetoothManager) {
		return BluetoothSendFileViaObex.getSizeOfJar(bluetoothManager);
	}

	public void sendJar(final RemoteDevice device) {
		sendFile(device, STATE_SEND_JAR);
	}

	public void sendJad(final RemoteDevice device) {
		sendFile(device, STATE_SEND_JAD);
	}

	private void sendFile(final RemoteDevice device, final int type) {
		startBluetoothAction(type);

		// bevor die Datei geschickt werden kann muss der OBEX-Dienst gefunden werden
		Thread searchObexServiceThread = new Thread(new Runnable() {
			public void run() {
				try {
					new BluetoothSearchObexService(getBluetoothManager(), device);
				} catch (Exception e) {
					failure(getClass(), e, " Thread - searchObexService " + type, FAILURE_EXCEPTION);
					return;
				}
			}
		});
		sendState = SEND_STATE_SEARCH_OBEX;
		searchObexServiceThread.start();
	}

	public void cancelAction() {
		try {
			switch (state) {
			case STATE_SEARCH_FOR_DEVICES:
				if (bc != null) {
					bc.cancelSearch();
				}
				break;
			case STATE_SEND_JAD:
			case STATE_SEND_JAR:
				if (bsfvo != null) {
					bsfvo.cancel();
				}
				break;
			}
		} catch (Exception e) {
			// beim Abbrechen werden Exceptions nicht behandelt
		}

		finishBluetoothAction();
	}

	synchronized public void failure(Class c, Exception e, String message, int errorType) {
		if (errorstate) {
			Logger.error(getClass(), "double Error");
			return;
		}
		errorstate = true;
		finishBluetoothAction();
		stateCallback.failure(c, e, message, errorType);
	}

	// CALLBACK SEARCH DEVICES

	public void finishedDeviceSearch(RemoteDevice[] devices, int inqueryState) {
		finishBluetoothAction();
		searchDeviceCallback.finishedDeviceSearch(devices, inqueryState);
		queryNames(devices);
	}

	public void foundDevice(RemoteDevice device) {
		updateBluetoothAction();
		searchDeviceCallback.foundDevice(device);
	}

	// CALLBACK QUERY NAME
	public void foundName(RemoteDevice device, String name) {
		updateBluetoothAction();
		searchDeviceCallback.friendyName(device, name);
	}

	private void finishedQueryNames() {
		finishBluetoothAction();
	}

	// CALLBACK SearchObexService
	public void finishedObexServiceSearch(final String connectionUrl, boolean success) {
		if (!success) {
			finishBluetoothAction();
			failure(getClass(), null, "Das augewählte Gerät unterstüzt nicht den OBEX-Dienst zur Datenübertragung", FAILURE_BLUETOOTH_NO_OBEX);
		} else {
			updateBluetoothAction();
			final int type;
			switch (state) {
			case STATE_SEND_JAR:
				type = BluetoothSendFileViaObex.SENDJAR;
				break;
			case STATE_SEND_JAD:
				type = BluetoothSendFileViaObex.SENDJAD;
				break;
			default:
				type = -1;
			}
			Thread sendFileThread = new Thread(new Runnable() {
				public void run() {
					try {
						sendState = SEND_STATE_CONNECTING;
						bsfvo = new BluetoothSendFileViaObex(getBluetoothManager(), connectionUrl, type);
						bsfvo.sendFile();
					} catch (Exception e) {
						failure(getClass(), e, " Thread - sendFile " + type, FAILURE_EXCEPTION);
						return;
					}
				}
			});
			sendFileThread.start();
		}
	}

	// CALLBACK GLOBAL

	public void setDebugState(Class c, String message) {
		//		System.out.println("\nSTATE: " + c + " " + message + "\n");
		stateCallback.setState(c + " " + message);
	}

	public void finishedBluetoothSendFileViaObex(boolean success, int type) {
		finishBluetoothAction();
		sendFileViaObexCallback.finishedBluetoothSendFileViaObex(success, type);
	}

	synchronized private void startBluetoothAction(int newState) {
		if (state != STATE_IDLE) {
			Logger.error(this, "nicht idle: " + state);
			cancelAction();
		}
		watchDogCreate();
		this.state = newState;
	}

	synchronized private void finishBluetoothAction() {
		watchDogDisable();
		state = STATE_IDLE;
		sendState = SEND_STATE_NONE;
	}

	synchronized private void updateBluetoothAction() {
		watchDogUpdate();
	}

	Timer watchdogTimer = null;
	WatchDogTimerTask watchdogTimerTask = null;

	synchronized public void watchDogCreate() {
		if (watchdogTimer != null) {
			watchdogTimer.cancel();
		}

		watchdogTimer = new Timer();
		watchdogTimerTask = new WatchDogTimerTask(this);
		watchdogTimer.scheduleAtFixedRate(watchdogTimerTask, UPDATE_INTERVAL, UPDATE_INTERVAL);
	}

	synchronized public void watchDogUpdate() {
		if (watchdogTimer == null) {
			Logger.error(getClass(), "Watchdog exisitert nicht");
			return;
		}

		watchdogTimerTask.update();
	}

	synchronized private void watchDogDisable() {
		if (watchdogTimer != null) {
			watchdogTimer.cancel();
		}
	}

	synchronized protected void timeOut() {
		cancelAction();
		failure(getClass(), null, "timeout", FAILURE_TIMEOUT);
	}

	synchronized public int getState() {
		return state;
	}

	synchronized public long getSendPos() {
		return bsfvo.getCurrentPos();
	}

	synchronized public int getSendState() {
		return sendState;
	}
	
	public void callBackSendingFile() {
		sendState = SEND_STATE_SENDING;
	}
	

	private static final long UPDATE_INTERVAL = 5000;
	public static final long TIMEOUT = 30000;

}

class WatchDogTimerTask extends TimerTask {
	private long lastUpdate;
	private final BluetoothManager bluetoothManager;

	public WatchDogTimerTask(BluetoothManager bluetoothManager) {
		lastUpdate = System.currentTimeMillis();
		if (BluetoothManager.SHOW_WATCHDOG_STATE) {
			System.out.println("Watchdog create");
		}
		this.bluetoothManager = bluetoothManager;
	}

	synchronized public final void run() {
		if (BluetoothManager.SHOW_WATCHDOG_STATE) {
			System.out.println("Watchdog t " + (System.currentTimeMillis() - BluetoothManager.TIMEOUT - lastUpdate));
		}
		if (System.currentTimeMillis() - BluetoothManager.TIMEOUT >= lastUpdate) {
			bluetoothManager.timeOut();
			cancel();
		}
	}

	synchronized public void update() {
		if (BluetoothManager.SHOW_WATCHDOG_STATE) {
			System.out.println("Watchdog Update");
		}
		lastUpdate = System.currentTimeMillis();
	}

}
