// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package control.connections.bluetooth;

import java.util.Vector;

import javax.bluetooth.BluetoothStateException;
import javax.bluetooth.DeviceClass;
import javax.bluetooth.DiscoveryAgent;
import javax.bluetooth.DiscoveryListener;
import javax.bluetooth.LocalDevice;
import javax.bluetooth.RemoteDevice;
import javax.bluetooth.ServiceRecord;

public class BluetoothSearchDevices implements DiscoveryListener {

	private final BluetoothManager bluetoothManager;
	private final BluetoothSearchDevices owner;
	private int inqueryState;

	private LocalDevice local;
	private Vector foundDevices = new Vector();
	private Vector foundRecords = new Vector();
	private Thread searchingForDevicesThread;

	public BluetoothSearchDevices(BluetoothManager bluetoothManager) {
		this.bluetoothManager = bluetoothManager;

		try {
			local = LocalDevice.getLocalDevice();
		} catch (BluetoothStateException BSe) {
			bluetoothManager.failure(getClass(), BSe, null, BluetoothManager.FAILURE_BLUETOOTH_NOT_WORKING);
		}

		owner = this;
	}

	synchronized public void searchForDevices() {
		inqueryState = INQUIRY_ERROR;
		// retrieve the discovery agent
		searchingForDevicesThread = new Thread(new Runnable() {
			public void run() {
				try {
					DiscoveryAgent agent = local.getDiscoveryAgent();
					// ok, start a new search then
					foundDevices.removeAllElements();
					synchronized (searchingForDevicesThread) {
						agent.startInquiry(DiscoveryAgent.GIAC, owner);
						searchingForDevicesThread.wait(); // until devices are found
					}

					// finished - report to callback function
					RemoteDevice[] devices = new RemoteDevice[foundDevices.size()];
					foundDevices.copyInto(devices);
					bluetoothManager.finishedDeviceSearch(devices, inqueryState);
				} catch (Exception e) {
					bluetoothManager.failure(owner.getClass(), e, null, BluetoothManager.FAILURE_BLUETOOTH_NOT_WORKING);
					return;
				}
			}
		});
		searchingForDevicesThread.start();
	}

	/**
	 * Wird aufgerufen sobald ein neues Geraet gefunden wurde - Speichert das
	 * gefundene Geraet, falls dieses noch nicht eingetragen wurde
	 */
	synchronized public void deviceDiscovered(RemoteDevice rd, DeviceClass dc) {
		// same device may be found several times during single search
		if (foundDevices.indexOf(rd) == -1) {
			foundDevices.addElement(rd);
			bluetoothManager.foundDevice(rd);
		}
	}

	/**
	 * 
	 * @param transactionID
	 * @param records
	 */
	synchronized public void servicesDiscovered(int transactionID, ServiceRecord[] records) {
		for (int i = 0; i < records.length; i++)
			foundRecords.addElement(records[i]);
	}

	/**
	 * 
	 * @param transactionID
	 * @param responseCode
	 *            - SERVICE_SEARCH_COMPLETED, SERVICE_SEARCH_TERMINATED,
	 *            SERVICE_SEARCH_ERROR,SERVICE_SEARCH_NO_RECORDS or
	 *            SERVICE_SEARCH_DEVICE_NOT_REACHABLE
	 */
	synchronized public void serviceSearchCompleted(int transactionID, int responseCode) {
		//		throw new UnsupportedOperationException("Not supported yet.");
	}

	/**
	 * Wird aufgerufen sobald die Suche abgeschlossen wurde Verarbeitet die
	 * gefundenen Daten
	 * 
	 * @param inquiryState
	 *            - INQUIRY_COMPLETED, INQUIRY_TERMINATED, or INQUIRY_ERROR
	 */
	synchronized public void inquiryCompleted(int inquiryState) {
		this.inqueryState = inquiryState;
		synchronized (searchingForDevicesThread) {
			searchingForDevicesThread.notify();
		}
	}

	synchronized public void cancelSearch() {
		local.getDiscoveryAgent().cancelInquiry(this);
	}

}
