// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package control.reminder;

import java.util.Date;
import java.util.Timer;

import model.DatumZeit;
import model.Route;
import control.Global;

public class Reminder {
	
	public static final int REMINDER_RESERVE_ZEIT = 1;
	public static final boolean showDebugData = false;

	private static Timer timer;
	private static Route route = Route.NULL;
	private static int minutenVorAbfahrt;

	/**
	 * Ueberprueft ob zu der uebergebenen Route abzüglich der Reservezeit ein Erinnerungsmodul gestartet
	 * werden kann oder ob die Zeit bereits zu weit fortgeschritten ist
	 * 
	 * @param route
	 * @return
	 */
	public static boolean isRemindable(Route route) {
		return isRemindable(route, Global.REMINDER_LAUFZEIT[0]);
	}

	/**
	 * Ueberprueft ob es zu der uebergebenen Route und vorLaufZeit noch sinnvoll ist ein Erinnerungs-modul zu
	 * starten oder ob die Zeit bereits zu weit fortgeschritten ist
	 * 
	 * @param route
	 * @param vorLaufZeit
	 * @return
	 */
	private static boolean isRemindable(Route route, int vorLaufZeit) {
		return (route.getAbfahrtszeit() - vorLaufZeit - Reminder.REMINDER_RESERVE_ZEIT - DatumZeit.Now() > 0);
	}

	/**
	 * Liefert die für diese Route noch sinnvollen Zeitspannen in Muten zurück, für die ein Reminder aktiviert
	 * werden kann
	 * 
	 * @param route
	 * @return
	 */
	public static int[] getPossibleReminderValues(Route route) {
		int cnt = 0;

		int maxTime = route.getAbfahrtszeit() - DatumZeit.Now() - Reminder.REMINDER_RESERVE_ZEIT;

		if (maxTime < 0) {
			maxTime = -1;
		}

		for (int i = 0; i < Global.REMINDER_LAUFZEIT.length; i++) {
			if (Global.REMINDER_LAUFZEIT[i] <= maxTime) {
				cnt++;
			}
		}

		int[] values = new int[cnt];

		for (int i = 0; i < cnt; i++) {
			values[i] = Global.REMINDER_LAUFZEIT[i];
		}

		return values;
	}

	/**
	 * Ueberprueft ob zu der uebergebenen Route bereits ein Reminder gestartet wurde
	 * 
	 * @param route
	 * @return Reminder zu der uebergebenen Route bereits vorhanden?
	 */
	public static boolean yetExists(Route route) {
		return isSameRoute(route, Reminder.route);
	}

	/**
	 * Gibt die Vorlaufzeit an, welche fuer diesen Reminder benutzt wird. Falls zu der uebergebenen Route kein
	 * Reminder gefunden wurde wird -1 zurueckgegeben
	 * 
	 * @param route
	 * @return Vorlaufzeit des Reminders zur uebergebenen Route bzw -1 falls kein Reminder zur Route gefunden
	 *         wurde
	 */
	public static int getMinutenVorAbfahrt() {
		return Reminder.minutenVorAbfahrt;
	}

	/**
	 * Stoppt und loescht einen Reminder zur entsprechenden Route, falls ein solcher vorhanden ist.
	 * 
	 * @param route
	 */
	
	public static void reset() {
		if (timer != null) {
			timer.cancel();
			route = Route.NULL;
			timer = null;
		}
	}

	/**
	 * Wenn bereits ein der uebergebenen Route entsprechender Reminder gestartet wurde und eine Aenderung
	 * entsprechend der uebergebenen vorLaufZeit moeglich ist, wird die VorLaufZeit des bestehenden Reminders
	 * abgaendert
	 * 
	 * @param route
	 * @param vorLaufZeit
	 * @return erfolgreiche Abaenderung
	 */
	public static void enable(Route route, int minutenVorAbfahrt) {
		reset();
		if (isRemindable(route, minutenVorAbfahrt) || Global.REMINDER_ALWAYS_VISIBLE) {
			Reminder.route = route;
			Reminder.minutenVorAbfahrt = minutenVorAbfahrt;
			timer = new Timer();
			int reminderZeit = route.getAbfahrtszeit() - minutenVorAbfahrt;
			if (showDebugData) {
				printReminderData("enable");
				System.out.println(" ^^^ reminderZeit: " + DatumZeit.toTimeString(reminderZeit));
			}
			timer.schedule(new ReminderTask(route), new Date(DatumZeit.convertToDeviceTime(reminderZeit)));
		}
	}

	/**
	 * Gibt Informationen über gestartete und gespeicherte Reminder aus
	 * 
	 * @param s
	 *            - Woher die Anfrage kommt (z.B. Start oder Stop)
	 */
	public static void printReminderData(String s) {
		System.out.println("----------------" + s + "------------------");
		Route r = Reminder.route;
		System.out.println(DatumZeit.toTimeString(r.getAbfahrtszeit()) + " - " + Reminder.getMinutenVorAbfahrt() + " - " + r.getHaltestelleVon() + " - " + r.getHaltestelleNach());
	}

	private static boolean isSameRoute(Route route, Route otherRoute) {
		boolean uebereinstimmung = (route.getAbfahrtszeit() == otherRoute.getAbfahrtszeit()) && (route.getHaltestelleVon().getId() == otherRoute.getHaltestelleVon().getId())
			&& (route.getHaltestelleNach().getId() == otherRoute.getHaltestelleNach().getId());
		return uebereinstimmung;
	}
}
