// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package data;

import java.util.Vector;

import model.AbstractLocation;
import model.DatumZeit;
import model.Fahrt;
import model.Haltestelle;
import model.IFahrt;
import model.Linie;
import model.Ort;
import model.Verkehrsmittel;
import utils.StringUtils;
import control.Global;
import dev.Logger;
import dev.Memory;
import dev.Statistics;

public class Data {
	private static Haltestelle[] alleHaltestellen = null;
	private static Linie[] alleLinien = null;
	private static IFahrt[] alleFahrten = null;
	private static Ort[] alleOrte = null;
	private static String dataVersion = "nicht initialisiert";
	private static short[][] crossTable = null;
	private static Haltestelle[][] haltestellenOfLinie; 
	
	// Strukturversion
	public static final int VERSION = 2;
	// Update Nr innerhalb der Struktur-Version
	public static final int UPDATE = 3;
	
	// Ablaufzeitpunkt des Datensatzes
	private static int DATENSATZ_GUELTIG_BIS = DatumZeit.NULL;
	private static int DATENSATZ_GUELTIG_VON = DatumZeit.NULL;
	
	public static final int NULL_ID = -1;
	
	public static final int DATSET_UNKNOWN = -1;
	public static final int DATSET_KLEIN = 0;
	public static final int DATSET_ALLES = 1;
	public static final int DATSET_INDIVIDUELL = 2;
	
	private static int datasetType = DATSET_UNKNOWN;
	
	public static void setDatasetType(int datasetType) {
		Data.datasetType = datasetType;
	}

	/**
	 * Wird von der Startklasse aufgerufen um die ganzen Daten zu laden.
	 */

	public static void loadDataStagePre(String datafile) {
		FileData.loadDataStagePre(datafile);
	}

	public static void loadDataStageFinal() {
		FileData.loadDataStageFinal();
	}

	public static boolean isLoaded() {
		return FileData.isLoaded();
	}

	/**
	 * getAlleOrte (ALLE Orte)
	 * 
	 * @return Ort[] Array mit allen Orten
	 */
	public static Ort[] getAlleOrte() {
		return Data.alleOrte;
	}

	public static void setAlleOrte(Ort[] orte) {
		Data.alleOrte = orte;
	}

	/**
	 * Setzt die Haltestellen und ordnet jeder Haltestelle eine ID zu.
	 * 
	 * @param Haltestelle
	 *            [] Array mit allen Haltestellen
	 */
	public static void setAlleHaltestellen(Haltestelle[] haltestellen) {
		Data.alleHaltestellen = haltestellen;

		// Ordnet jeder Haltestelle eine ID zu
		for (int i = 0; i < Data.alleHaltestellen.length; i++) {
			Data.alleHaltestellen[i].setId(i);
		}
	}

	/**
	 * getOrt (EIN Ort)
	 * 
	 * @param id
	 *            des Ortes
	 * @return Ort mit der id
	 */
	public static Ort getOrt(int id) {
		return Data.alleOrte[id];
	}

	/**
	 * Gibt die Haltestelle für die INTERNE Push&Ride des Datensatzes zurück.
	 * 
	 * @param pushandrideId
	 *            der Haltestelle
	 * @return Haltestelle
	 */
	public static Haltestelle getHaltestelle(int pushandrideId) {
		if (pushandrideId < 0 || pushandrideId >= Data.alleHaltestellen.length) {
			return Haltestelle.NULL;
		} else {
			return Data.alleHaltestellen[pushandrideId];
		}
	}

	/**
	 * Gibt alle Haltestellen als Array zurück.<br>
	 * Der Index des Arrays ist die interne ID unseres Datensatzes.
	 * 
	 * @return Haltestelle[] Array mit allen Haltestellen
	 */
	public static Haltestelle[] getAlleHaltestellen() {
		return Data.alleHaltestellen;
	}

	/**
	 * setLinien
	 * 
	 * @param Linie
	 *            [] Array mit allen Linien
	 */
	public static void setAlleLinien(Linie[] linien) {
		Data.alleLinien = linien;
	}

	/**
	 * getLinie (EINE Linie)
	 * 
	 * @param id
	 *            der Linie
	 * @return Linie mit der id
	 */
	public static Linie getLinie(int id) {
		return Data.alleLinien[id];
	}

	/**
	 * getLinien (ALLE Linien)
	 * 
	 * @return Linie[] Array mit allen Linien
	 */
	public static Linie[] getAlleLinien() {
		return Data.alleLinien;
	}

	/**
	 * setAlleFahrten
	 * 
	 * @param fahrten
	 *            Array mit allen Fahrten
	 */
	public static void setAlleFahrten(IFahrt[] fahrten) {
		Data.alleFahrten = fahrten;
	}

	/**
	 * getAlleFahrten
	 * 
	 * @return Array mit allen Fahrten
	 */
	public static IFahrt[] getAlleFahrten() {
		return Data.alleFahrten;
	}

	/**
	 * Setzt die Version der Daten
	 * 
	 * @param version
	 *            String
	 */
	public static void setVersion(String version) {
		Data.dataVersion = version;
	}

	/**
	 * Gibt die Version der Datenfiles zurück
	 * 
	 * @return String
	 */
	public static String getVersion() {
		return Data.dataVersion;
	}

	/**
	 * liefert eine Liste mit allen Orten und Haltestellen zusammen
	 */
	public static AbstractLocation[] getAlleLocations() {
		Ort[] orte = getAlleOrte();
		Vector locations = new Vector();
		for (int ortsId = 0; ortsId < orte.length; ortsId++) {
			Ort ort;
			ort = orte[ortsId];
			locations.addElement(ort);

			if (ort.getAlleHaltestellen().size() <= 0) {
				continue;
			}

				Haltestelle[] haltestellenInOrt = new Haltestelle[ort.getAlleHaltestellen().size()];
			ort.getAlleHaltestellen().copyInto(haltestellenInOrt);

			// falls nur eine Haltestelle im Ort  verfügbar ist und dieser Bahnhof heisst - nicht anzeigen
			if (ort.getAlleHaltestellen().size() == 1 && haltestellenInOrt[0].getScreenName().equals("Bahnhof")) {
				continue;
			}

			for (int haltestellenId = 0; haltestellenId < haltestellenInOrt.length; haltestellenId++) {
				locations.addElement(haltestellenInOrt[haltestellenId]);
			}
		}
		AbstractLocation[] returnValue = new AbstractLocation[locations.size()];
		locations.copyInto(returnValue);
		return returnValue;
	}

	/**
	 * Lädt den Datensatz mit einem Funktionsaufruf
	 */
	public static void loadDataComplete(String datafile) {
		FileData.loadDataComplete(datafile);
	}

	/**
	 * gibt alle Linien zurück, die zur Berechnung der Route zwischen dem
	 * haltestellenPaar verwendet werden können
	 */
	public static Vector getMoeglicheLinienOfHaltestellenPaar(Haltestelle von, Haltestelle nach) {
		if (crossTable == null) {
			return null;
		}
		// alle Linien von Von und Nach ermitteln
		Linie[] linien = Data.getAlleLinien();

		Vector vonLinien = new Vector();
		Vector nachLinien = new Vector();

		IFahrt fahrtVon;
		IFahrt[] fahrtenOfVonHaltestelle = von.getAlleFahrten();
		for (int fahrtNo = 0; fahrtNo < fahrtenOfVonHaltestelle.length; fahrtNo++) {
			fahrtVon = fahrtenOfVonHaltestelle[fahrtNo];
			Linie linieVon = fahrtVon.getLinie();
			if (!vonLinien.contains(linieVon)) {
				vonLinien.addElement(linieVon);
			}
		}

		IFahrt fahrtNach;
		IFahrt[] fahrtenOfNachHaltestelle = nach.getAlleFahrten();
		for (int fahrtNo = 0; fahrtNo < fahrtenOfNachHaltestelle.length; fahrtNo++) {
			fahrtNach = fahrtenOfNachHaltestelle[fahrtNo];

			Linie linieNach = fahrtNach.getLinie();
			if (!nachLinien.contains(linieNach)) {
				nachLinien.addElement(linieNach);
			}
		}

		// die Schnittmenge zwischen den verwendeten Linien ermitteln

		if (vonLinien.size() == 0) {
			Logger.error(Linie.class, "vonLinien.size = 0");
			// return new Linie[0];
			return null;
		}
		if (nachLinien.size() == 0) {
			Logger.error(Linie.class, "nachLinien.size = 0");
			// return new Linie[0];
			return null;
		}

		Vector resultVektor = new Vector();

		// erste Kombination hinzufügen
		short[] moeglicheLinien;
		moeglicheLinien = getMoeglicheLinienOfLinienpaar((Linie) vonLinien.elementAt(0), (Linie) nachLinien.elementAt(0));

		for (int i = 0; i < moeglicheLinien.length; i++) {
			resultVektor.addElement(linien[moeglicheLinien[i]]);
		}

		// die Schnittmenge mit den anderen bilden
		for (int vonNo = 0; vonNo < vonLinien.size(); vonNo++) {
			for (int nachNo = 0; nachNo < nachLinien.size(); nachNo++) {
				if (vonNo == 0 && nachNo == 0) {
					continue;
				}
				moeglicheLinien = getMoeglicheLinienOfLinienpaar((Linie) vonLinien.elementAt(vonNo), (Linie) nachLinien.elementAt(nachNo));
				Vector newResult = new Vector();
				for (int i = 0; i < moeglicheLinien.length; i++) {
					if (resultVektor.contains(linien[moeglicheLinien[i]])) {
						newResult.addElement(linien[moeglicheLinien[i]]);
					}
				}
				resultVektor = newResult;
			}
		}

		// Start / Ziellinien hinzufügen
		for (int vonNo = 0; vonNo < vonLinien.size(); vonNo++) {
			if (!resultVektor.contains(vonLinien.elementAt(vonNo))) {
				resultVektor.addElement(vonLinien.elementAt(vonNo));
			}
		}
		for (int nachNo = 0; nachNo < nachLinien.size(); nachNo++) {
			if (!resultVektor.contains(nachLinien.elementAt(nachNo))) {
				resultVektor.addElement(nachLinien.elementAt(nachNo));
			}
		}

		return resultVektor;
	}

	/**
	 * gibt alle Linien zurück, die zur Berechnung der Route zwischen den beiden
	 * Linien verwendet werden müssen
	 */
	private static short[] getMoeglicheLinienOfLinienpaar(Linie von, Linie nach) {
		return (crossTable[von.getId() * getAlleLinien().length + nach.getId()]);
	}

	public static void setCrossTable(short[][] crossTable) {
		Data.crossTable = crossTable;
	}

	public static Linie[] getAlleLinienLaufwege() {
		return getLaufwegeOfLinien(getAlleLinien());
	}

	public static int getTotalMemory() {
		int memory = 0;
		int currentMem;
		currentMem = getMemOfAlleOrte();
		memory += currentMem;

		currentMem = getMemOfAlleHaltestellen();
		memory += currentMem;
		
		currentMem = getMemOfAlleLinien();
		memory += currentMem;
		
		currentMem = getMemOfAlleFahrten();
		memory += currentMem;
		
		return memory;
	}
	
	public static String createDataSetInfo() {
		StringBuffer result = new StringBuffer("");
		result.append("Datensatz-Version: " + dataVersion + "\n");

		result.append("\n");

		int captionlength = 20;
		int memlength = 14;
		int intlength = 6;
		int memory = 0;
		int currentMem;
		currentMem = getMemOfAlleOrte();
		memory += currentMem;
		result.append(StringUtils.fillStringBack("#Orte", ' ', captionlength) + StringUtils.fillInt(getAlleOrte().length, ' ', intlength) + " Mem: " + StringUtils.fillInt(currentMem, ' ', memlength) + "\n");

		currentMem = getMemOfAlleHaltestellen();
		memory += currentMem;
		result.append(StringUtils.fillStringBack("#Haltestellen", ' ', captionlength) + StringUtils.fillInt(getAlleHaltestellen().length, ' ', intlength) + " Mem: " + StringUtils.fillInt(currentMem, ' ', memlength) + "\n");
		
		currentMem = getMemOfAlleLinien();
		memory += currentMem;
		result.append(StringUtils.fillStringBack("#Alle Linien", ' ', captionlength) + StringUtils.fillInt(getAlleLinien().length, ' ', intlength) + " Mem: " + StringUtils.fillInt(currentMem, ' ', memlength) + "\n");
		
		currentMem = getMemOfAlleFahrten();
		memory += currentMem;
		result.append(StringUtils.fillStringBack("#Fahrten", ' ', captionlength) + StringUtils.fillInt(getAlleFahrten().length, ' ', intlength) + " Mem: " + StringUtils.fillInt(currentMem, ' ', memlength) + "\n");
		int memLaufwegeFahrten = getMemOfLaufwegeFahrten();
		int anzLaufwegeFahrten = getAnzLaufwegeFahrten();
		result.append(StringUtils.fillStringBack("^^ davon Laufwege", ' ', captionlength) + StringUtils.fillInt(anzLaufwegeFahrten, ' ', intlength) + " Mem: " + StringUtils.fillInt(memLaufwegeFahrten, ' ', memlength) + "\n");
		
		result.append(StringUtils.fillStringBack("#Abfahrtszeiten", ' ', captionlength - 3) + StringUtils.fillInt( getNumberOfAbfahrtszeiten(), ' ', intlength + 3) + "\n");

		result.append(StringUtils.fillStringBack("Total Memory", ' ', captionlength + intlength) + "      " + StringUtils.fillInt(memory, ' ', memlength) + "\n");

		result.append("\n");
		if (1==1) return result.toString();
		Statistics statistics = new Statistics("Data.statistics");
		Statistics statistics2 = new Statistics("Data.statistics2");
		Statistics statistics3 = new Statistics("Data.statistics3");

		result.append("Statistik: Name Anzahl Mittelwert Standardabweichung\n");
		result.append("\n");
		int lenAnzahl = 5;
		int lenVorKomma = 3;
		int lenNachKomma = 2;


		result.append("+++ Orte +++\n");
		Ort[] alleOrte = getAlleOrte();
		for (int ortNo = 0; ortNo < alleOrte.length; ortNo++) {
			statistics.add(alleOrte[ortNo].getAlleHaltestellen().size());
		}

		result.append(StringUtils.fillStringBack(" ^ Haltestellen", ' ', captionlength));
		result.append(" " + StringUtils.formatDouble(statistics.getAnzahl(), lenAnzahl, 0));
		result.append(" " + StringUtils.formatDouble(statistics.getMittelwert(), lenVorKomma, lenNachKomma));
		result.append(" " + StringUtils.formatDouble(statistics.getStandardabweichung(), lenVorKomma, lenNachKomma));
		result.append("\n");
		
		result.append("+++ Haltestellen +++\n");
		Haltestelle[] alleHaltestellen = getAlleHaltestellen();
		
		statistics.clear();
		statistics2.clear();
		statistics3.clear();
		
		for (int haltestellenNo = 0; haltestellenNo < alleHaltestellen.length; haltestellenNo++) {
			Linie[] linien = getLinienOfFahrten(alleHaltestellen[haltestellenNo].getAlleFahrten());
			statistics.add(getEchteLinieOfLinien(linien).length);
			statistics2.add(getLaufwegeOfLinien(linien).length);
			statistics3.add(alleHaltestellen[haltestellenNo].getAlleFahrten().length);
		}

		result.append(StringUtils.fillStringBack(" ^ Linien (echt)", ' ', captionlength));
		result.append(" " + StringUtils.formatDouble(statistics.getAnzahl(), lenAnzahl, 0));
		result.append(" " + StringUtils.formatDouble(statistics.getMittelwert(), lenVorKomma, lenNachKomma));
		result.append(" " + StringUtils.formatDouble(statistics.getStandardabweichung(), lenVorKomma, lenNachKomma));
		result.append("\n");

		result.append(StringUtils.fillStringBack(" ^ Linien (Laufwege)", ' ', captionlength));
		result.append(" " + StringUtils.formatDouble(statistics2.getAnzahl(), lenAnzahl, 0));
		result.append(" " + StringUtils.formatDouble(statistics2.getMittelwert(), lenVorKomma, lenNachKomma));
		result.append(" " + StringUtils.formatDouble(statistics2.getStandardabweichung(), lenVorKomma, lenNachKomma));
		result.append("\n");
		result.append(StringUtils.fillStringBack(" ^ Fahrten", ' ', captionlength));
		result.append(" " + StringUtils.formatDouble(statistics3.getAnzahl(), lenAnzahl, 0));
		result.append(" " + StringUtils.formatDouble(statistics3.getMittelwert(), lenVorKomma, lenNachKomma));
		result.append(" " + StringUtils.formatDouble(statistics3.getStandardabweichung(), lenVorKomma, lenNachKomma));
		result.append("\n");

		result.append("+++ Linie (echt) +++\n");
		statistics.clear();
		Linie[] alleLinien = getAlleLinien();
		for (int linienNo = 0; linienNo < alleLinien.length; linienNo++) {
			if (alleLinien[linienNo].getVerkehrsmittel() == Verkehrsmittel.LAUFWEG) {
				continue;
			}
			statistics.add(getFahrtenOfLinie(alleLinien[linienNo]).length);
		}

		result.append(StringUtils.fillStringBack(" ^ Fahrten", ' ', captionlength));
		result.append(" " + StringUtils.formatDouble(statistics.getAnzahl(), lenAnzahl, 0));
		result.append(" " + StringUtils.formatDouble(statistics.getMittelwert(), lenVorKomma, lenNachKomma));
		result.append(" " + StringUtils.formatDouble(statistics.getStandardabweichung(), lenVorKomma, lenNachKomma));
		result.append("\n");


		result.append("+++ Fahrten (echt) +++\n");
		statistics.clear();
		for (int fahrtNo = 0; fahrtNo < alleFahrten.length; fahrtNo++) {
			IFahrt fahrt = alleFahrten[fahrtNo];
			if (fahrt.isLaufweg()) {
				continue;
			}
			statistics.add(fahrt.getAlleHaltestellen().length);
		}

		result.append(StringUtils.fillStringBack(" ^ Haltestellen", ' ', captionlength));
		result.append(" " + StringUtils.formatDouble(statistics.getAnzahl(), lenAnzahl, 0));
		result.append(" " + StringUtils.formatDouble(statistics.getMittelwert(), lenVorKomma, lenNachKomma));
		result.append(" " + StringUtils.formatDouble(statistics.getStandardabweichung(), lenVorKomma, lenNachKomma));
		result.append("\n");

		return result.toString();
	}

	public static Linie[] getLaufwegeOfLinien(Linie[] linien) {
		Vector v = new Vector();
		for (int i = 0; i < linien.length; i++) {
			if (linien[i].getVerkehrsmittel() == Verkehrsmittel.LAUFWEG) {
				v.addElement(linien[i]);
			}
		}
		Linie[] result = new Linie[v.size()];
		v.copyInto(result);
		return result;
	}

	public static Linie[] getEchteLinieOfLinien(Linie[] linien) {
		Vector v = new Vector();
		for (int i = 0; i < linien.length; i++) {
			if (linien[i].getVerkehrsmittel() != Verkehrsmittel.LAUFWEG) {
				v.addElement(linien[i]);
			}
		}
		Linie[] result = new Linie[v.size()];
		v.copyInto(result);
		return result;
	}

	public static Linie[] getLinienOfFahrten(IFahrt[] fahrten) {
		Vector linien = new Vector();
		for (int i = 0; i < fahrten.length; i++) {
			if (!linien.contains(fahrten[i].getLinie())) {
				linien.addElement(fahrten[i].getLinie());
			}
		}
		
		Linie[] result = new Linie[linien.size()];
		linien.copyInto(result);
		
		return result;
	}
	
	public static IFahrt[] getFahrtenOfLinie(Linie linie) {
		Vector v = new Vector();
		IFahrt[] alleFahrten = getAlleFahrten();
		for (int i = 0; i < alleFahrten.length; i++) {
			if (alleFahrten[i].getLinie() == linie) {
				v.addElement(alleFahrten[i]);
			}
		}
		IFahrt[] result = new IFahrt[v.size()];
		v.copyInto(result);
		return result;
	}
	
	public static int getMemOfAlleHaltestellen() {
		Haltestelle[] alleHaltestellen = getAlleHaltestellen();
		int mem = Memory.getArrayBaseSize() + alleHaltestellen.length * Memory.getObjSize();
		for (int i = 0; i < alleHaltestellen.length; i++) {
			mem += alleHaltestellen[i].getMemory();
		}
		return mem;
	}
	
	public static int getMemOfAlleLinien() {
		Linie[] alleLinien = getAlleLinien();
		int mem = Memory.getArrayBaseSize() + alleLinien.length * Memory.getObjSize();
		for (int i = 0; i < alleLinien.length; i++) {
			mem += alleLinien[i].getMemory();
		}
		return mem;
	}
	
	public static int getMemOfAlleFahrten() {
		IFahrt[] alleFahrten = getAlleFahrten();
		int mem = Memory.getArrayBaseSize() + alleFahrten.length * Memory.getObjSize();
		for (int i = 0; i < alleFahrten.length; i++) {
			mem += alleFahrten[i].getMemory();
		}
		return mem;
	}
	
	public static int getAnzLaufwegeFahrten() {
		IFahrt[] alleFahrten = getAlleFahrten();
		int anz = 0;
		for (int i = 0; i < alleFahrten.length; i++) {
			if (alleFahrten[i].isLaufweg()) {
				anz++;
			}
		}
		return anz;
	}

	public static int getMemOfLaufwegeFahrten() {
		IFahrt[] alleFahrten = getAlleFahrten();
		int mem = Memory.getArrayBaseSize() + alleFahrten.length * Memory.getObjSize();
		for (int i = 0; i < alleFahrten.length; i++) {
			if (alleFahrten[i].isLaufweg()) {
				mem += alleFahrten[i].getMemory();
			}
		}
		return mem;
	}
	
	public static int getMemOfAlleOrte() {
		Ort[] alleOrte = getAlleOrte();
		int mem = Memory.getArrayBaseSize() + alleOrte.length * Memory.getObjSize();
		for (int i = 0; i < alleOrte.length; i++) {
			mem += alleOrte[i].getMemory();
		}
		return mem;
	}
	
	public static long getNumberOfAbfahrtszeiten() {
		long cnt = 0;
		IFahrt[] alleFahrten = getAlleFahrten();
		for (int i = 0; i < alleFahrten.length; i++) {
			IFahrt iFahrt = alleFahrten[i]; 
			if (!iFahrt.isLaufweg()) {
				Fahrt fahrt = (Fahrt) iFahrt;
				cnt += fahrt.getAllAbfahrtszeitenByDay(Global.TAG_WOCHENTAG_FERIEN).length * iFahrt.getAlleHaltestellen().length;
				cnt += fahrt.getAllAbfahrtszeitenByDay(Global.TAG_SAMSTAG).length * iFahrt.getAlleHaltestellen().length;
				cnt += fahrt.getAllAbfahrtszeitenByDay(Global.TAG_SONN_FEIERTAG).length * iFahrt.getAlleHaltestellen().length;
			}
		}
		return cnt;
	}

	public static void setGueltigVon(int zeitGueltigVon) {
		Data.DATENSATZ_GUELTIG_VON = zeitGueltigVon;
	}

	public static void setGueltigBis(int zeitGueltigBis) {
		Data.DATENSATZ_GUELTIG_BIS = zeitGueltigBis;	
	}
	
	public static int getGueltigVon() {
		return Data.DATENSATZ_GUELTIG_VON;
	}

	public static int getGueltigBis() {
		return Data.DATENSATZ_GUELTIG_BIS;	
	}
	
	/**
	 * ID des Verkehrsverbundes!
	 * 
	 * @param uniqueId
	 * @return
	 */
	public static Haltestelle getHaltestelleByUniqueId(String uniqueId) {
		int id = uniqueIdToId(uniqueId);
		if (id == NULL_ID) {
			return Haltestelle.NULL;
		}
		return alleHaltestellen[id];
	}
	
	// FIXME: miese Performance - checken wo es verwendet wird und evtl. cachen
	public static int uniqueIdToId(String uniqueId) {
		for (int i = 0; i < alleHaltestellen.length; i++) {
			Haltestelle h = alleHaltestellen[i];
			if (h.getUniqueId().equals(uniqueId)){
				return i;
			}
		}
		return NULL_ID;	
	}
	
	public static int getIdOfHaltestelle(Ort o) {
		Ort[] alleOrte = Data.getAlleOrte();
		for (int i = 0; i < alleOrte.length; i++) {
			if (o == alleOrte[i]) {
				return i;
			}
		}
		return NULL_ID;
	}

	public static int getIdOfFahrt(IFahrt f) {
		IFahrt[] alleFahrten = Data.getAlleFahrten();
		for (int i = 0; i < alleFahrten.length; i++) {
			if (f == alleFahrten[i]) {
				return i;
			}
		}
		return NULL_ID;
	}

	private static void addUniqueToVector(Vector v, IFahrt fahrt) {
		Haltestelle[] haltestellen = fahrt.getAlleHaltestellen();
		for (int i = 0; i < haltestellen.length; i++) {
			Haltestelle haltestelle = haltestellen[i]; 
			if (!v.contains(haltestelle)) {
				v.addElement(haltestelle);
			}
		}
	}
	
	public static void createHaltestellenOfLinienArray() {
		Vector[] tmp = new Vector[getAlleLinien().length];
		haltestellenOfLinie = new Haltestelle[getAlleLinien().length][];

		for (int i = 0; i < tmp.length; i++) {
			tmp[i] = new Vector();
		}

		for (int i = 0; i < getAlleFahrten().length; i++) {
			IFahrt fahrt = getAlleFahrten()[i];
			if (fahrt.isLaufweg()) {
				continue;
			}
			addUniqueToVector(tmp[fahrt.getLinie().getId()], fahrt);
		}
		
		for (int i = 0; i < haltestellenOfLinie.length; i++) {
			haltestellenOfLinie[i] = new Haltestelle[tmp[i].size()];
			tmp[i].copyInto(haltestellenOfLinie[i]);
		}
	}

	public static Haltestelle[] getHaltestellenOfLinie(Linie linie) {
		if (linie == Linie.LAUFWEG_LINIE) {
			Logger.error(Data.class, "getHaltestellenOfLinie( !Laufweg! )");
			return new Haltestelle[0];
		}
		return haltestellenOfLinie[linie.getId()];
	}

	public static int getCurrentDatasetType() {
		return datasetType;
	}
}
