// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package data;

import java.io.DataInput;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.UTFDataFormatException;
import java.util.Vector;

import model.Haltestelle;
import model.IFahrt;

public class MyDataInputStream extends InputStream implements DataInput {

	private final boolean USE_CACHING = true;
	private final Vector cacheString;
	private final Vector cacheFahrzeiten; // byte[]
	private final Vector cacheAbfahrtszeiten; // short[]
	private final Vector cacheFahrtenOfHaltestelle; // short[]
	private final Vector cacheHaltestellenOfFahrt; // short[]

	private final int CACHE_OFFSET = 10000;

	public MyDataInputStream(InputStream in) {
		this.in = in;
		if (USE_CACHING) {
			cacheString = new Vector();
			cacheFahrzeiten = new Vector();
			cacheAbfahrtszeiten = new Vector();
			cacheFahrtenOfHaltestelle = new Vector();
			cacheHaltestellenOfFahrt = new Vector();
		} else {
			cacheString = null;
			cacheFahrzeiten = null;
			cacheAbfahrtszeiten = null;
			cacheFahrtenOfHaltestelle = null;
			cacheHaltestellenOfFahrt = null;
		}
	}

	public int read() throws IOException {
		return in.read();
	}

	public final int read(byte b[]) throws IOException {
		return in.read(b, 0, b.length);
	}

	public final int read(byte b[], int off, int len) throws IOException {
		return in.read(b, off, len);
	}

	public final void readFully(byte b[]) throws IOException {
		readFully(b, 0, b.length);
	}

	public final void readFully(byte b[], int off, int len) throws IOException {
		if (len < 0)
			throw new IndexOutOfBoundsException();
		int count;
		for (int n = 0; n < len; n += count) {
			count = read(b, off + n, len - n);
			if (count < 0)
				throw new EOFException();
		}

	}

	public final int skipBytes(int n) throws IOException {
		int total = 0;
		for (int cur = 0; total < n && (cur = (int) skip(n - total)) > 0; total += cur)
			;
		return total;
	}

	public final boolean readBoolean() throws IOException {
		int a = read();
		if (a < 0)
			throw new EOFException();
		else
			return a != 0;
	}

	public final byte readByte() throws IOException {
		int a = read();
		if (a < 0)
			throw new EOFException();
		else
			return (byte) a;
	}

	public final int readUnsignedByte() throws IOException {
		int a = read();
		if (a < 0)
			throw new EOFException();
		else
			return a;
	}

	public final short readShort() throws IOException {
		return (short) readUnsignedShort();
	}

	public final int readUnsignedShort() throws IOException {
		int a = read();
		int b = read();
		if ((a | b) < 0)
			throw new EOFException();
		else
			return (a << 8) + (b << 0);
	}

	public final char readChar() throws IOException {
		return (char) readUnsignedShort();
	}

	public final int readInt() throws IOException {
		int a = read();
		int b = read();
		int c = read();
		int d = read();
		if ((a | b | c | d) < 0)
			throw new EOFException();
		else
			return (a << 24) + (b << 16) + (c << 8) + (d << 0);
	}

	public final long readLong() throws IOException {
		return ((long) readInt() << 32) + ((long) readInt() & 4294967295L);
	}

	public final float readFloat() throws IOException {
		return Float.intBitsToFloat(readInt());
	}

	public final double readDouble() throws IOException {
		return Double.longBitsToDouble(readLong());
	}

	public final String readUTF() throws IOException {
		return readUTF(((DataInput) (this)));
	}

	public final String readUTF(DataInput in) throws IOException {
		int utflen = in.readUnsignedShort();

		// caching:
		if (USE_CACHING) {
			if (utflen >= CACHE_OFFSET) {
				return (String) cacheString.elementAt(utflen - CACHE_OFFSET);
			}
		}

		char str[] = new char[utflen];
		byte bytearr[] = new byte[utflen];
		int count = 0;
		int strlen = 0;
		in.readFully(bytearr, 0, utflen);
		do {
			if (count >= utflen)
				break;
			int c = bytearr[count] & 255;
			switch (c >> 4) {
			case 0: // '\0'
			case 1: // '\001'
			case 2: // '\002'
			case 3: // '\003'
			case 4: // '\004'
			case 5: // '\005'
			case 6: // '\006'
			case 7: // '\007'
			{
				count++;
				str[strlen++] = (char) c;
				break;
			}

			case 12: // '\f'
			case 13: // '\r'
			{
				if ((count += 2) > utflen)
					throw new UTFDataFormatException();
				int char2 = bytearr[count - 1];
				if ((char2 & 192) != 128)
					throw new UTFDataFormatException();
				str[strlen++] = (char) ((c & 31) << 6 | char2 & 63);
				break;
			}

			case 14: // '\016'
			{
				if ((count += 3) > utflen)
					throw new UTFDataFormatException();
				int char2 = bytearr[count - 2];
				int char3 = bytearr[count - 1];
				if ((char2 & 192) != 128 || (char3 & 192) != 128)
					throw new UTFDataFormatException();
				str[strlen++] = (char) ((c & 15) << 12 | (char2 & 63) << 6 | (char3 & 63) << 0);
				break;
			}

			case 8: // '\b'
			case 9: // '\t'
			case 10: // '\n'
			case 11: // '\013'
			default: {
				throw new UTFDataFormatException();
			}
			}
		} while (true);

		String returnValue = new String(str, 0, strlen);

		if (USE_CACHING) {
			cacheString.addElement(returnValue);
		}

		return returnValue;
	}

	public long skip(long n) throws IOException {
		return in.skip(n);
	}

	public int available() throws IOException {
		return in.available();
	}

	public void close() throws IOException {
		in.close();

		// explizites Freigeben der Caches
		if (USE_CACHING) {
			cacheAbfahrtszeiten.removeAllElements();
			cacheFahrzeiten.removeAllElements();
			cacheHaltestellenOfFahrt.removeAllElements();
			cacheFahrtenOfHaltestelle.removeAllElements();
			cacheString.removeAllElements();
		}
	}

	public synchronized void mark(int readlimit) {
		in.mark(readlimit);
	}

	public synchronized void reset() throws IOException {
		in.reset();
	}

	public boolean markSupported() {
		return in.markSupported();
	}

	protected InputStream in;

	public int[] readByteFahrzeitenArray() throws IOException {
		int length = readUnsignedShort();
		if (USE_CACHING) {
			if (length >= CACHE_OFFSET) {
				return (int[]) cacheFahrzeiten.elementAt(length - CACHE_OFFSET);
			}
		}
		// byte[] data = new byte[length];
		// readFully(data);
		int[] data = new int[length];
		for (int i = 0; i < length; i++) {
			data[i] = readUnsignedByte();
		}

		cacheFahrzeiten.addElement(data);
		return data;
	}

	public short[] readShortAbfahrtszeitenArray() throws IOException {
		int length = readUnsignedShort();
		if (USE_CACHING) {
			if (length >= CACHE_OFFSET) {
				return (short[]) cacheAbfahrtszeiten.elementAt(length - CACHE_OFFSET);
			}
		}
		short[] data = new short[length];
		for (int i = 0; i < length; i++) {
			data[i] = readShort();
		}
		cacheAbfahrtszeiten.addElement(data);
		return data;
	}

	public IFahrt[] readShortFahrtenOfHaltestelleArray() throws IOException {
		int length = readUnsignedShort();
		if (USE_CACHING) {
			if (length >= CACHE_OFFSET) {
				return (IFahrt[]) cacheFahrtenOfHaltestelle.elementAt(length - CACHE_OFFSET);
			}
		}
		IFahrt[] data = new IFahrt[length];
		for (int i = 0; i < length; i++) {
			data[i] = Data.getAlleFahrten()[readShort()];
		}
		cacheFahrtenOfHaltestelle.addElement(data);
		return data;
	}

	public Haltestelle[] readShortHaltestellenOfFahrtArray() throws IOException {
		int length = readUnsignedShort();
		if (USE_CACHING) {
			if (length >= CACHE_OFFSET) {
				return (Haltestelle[]) cacheHaltestellenOfFahrt.elementAt(length - CACHE_OFFSET);
			}
		}
		Haltestelle[] data = new Haltestelle[length];
		for (int i = 0; i < length; i++) {
			data[i] = Data.getHaltestelle(readShort());
		}
		cacheHaltestellenOfFahrt.addElement(data);
		return data;
	}
}
