// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package data.persistence;

import javax.microedition.rms.InvalidRecordIDException;
import javax.microedition.rms.RecordStore;
import javax.microedition.rms.RecordStoreException;
import javax.microedition.rms.RecordStoreFullException;
import javax.microedition.rms.RecordStoreNotFoundException;
import javax.microedition.rms.RecordStoreNotOpenException;

import utils.StringUtils;
import control.Global;
import dev.Logger;

/**
 * Persistence API
 */
public class Persistence {

	// Show info messages regarding RecordStore?
	private static final boolean LOCAL_DEBUG_OUTPUT = false;

	// RecordStore Name (Bei größeren Änderungen am Design des RecordStores eins hochzählen)
	// http://java.sun.com./javame/reference/apis/jsr037/javax/microedition/rms/RecordStore.html
	// max 32. unicode chars
	private static final String CONF_RS_NAME = "RushRide_Persistence8";


	// SLOT IDs
	protected static final int SLOT_LASTHALTESTELLEN = 1;
	protected static final int SLOT_SETTINGS_INT = 2;
	protected static final int SLOT_FAVORITEN = 3;
	//protected static final int SLOT_ROUTEN = 4; <- nicht wiederverwenden
	protected static final int SLOT_STATS = 5;
	protected static final int SLOT_REMINDERS = 6;
	protected static final int SLOT_SETTINGS_STRING = 7;
	protected static final int SLOT_DATASET_TYPE = 8;
	protected static final int SLOT_DATASET_CUSTOM_LINES = 9;
	protected static final int SLOT_DATASET_CUSTOM_BINARY = 10;
	protected static final int SLOT_UNIQUE_ID = 11;

	// Array Größe aktualisieren!!! (= maxSlotID + 1)
	private static final int NUM = 12;
	
	// Zeit bevor Dirty-Daten in den RecordStore geschrieben werden (in ms)
	public static final int RS_COMMIT_IMMEDIATLY = -1;
	public static final int RS_COMMIT_AFTER_300ms = 300;

	// CACHING
	private static boolean[] dirty = new boolean[NUM];
	private static boolean[] loaded = new boolean[NUM];
	private static byte[][] cache = new byte[NUM][];
	private static boolean isInitialized = false;

	// Record-Store Write Thread
	private static CommitThread commitThread = null;

	/**
	 * Initializes the Persistence Layer
	 */
	private static void init() {
		if (Persistence.isInitialized) {
			return;
		}

		// Erstelle Datenstrukturen
		for (int i = 0; i < NUM; i++) {
			Persistence.dirty[i] = false;
			Persistence.loaded[i] = false;
			Persistence.cache[i] = null;
		}
		Persistence.isInitialized = true;
	}

	/**
	 * Löscht die verschiedenen SecondLevel Caches welche in persistence package verwendet werden.
	 */
	private static void clearAllCaches() {
		UserSettingsInt.clearCache();
		UserSettingsString.clearCache();
		isInitialized = false;
	}

	/**
	 * Commits all data to persistent RecordStore
	 */
	protected synchronized static void commit() {
		// RecordStore funktioniert nur bei Midlets
		if(!Global.ISMIDLET) {
			return;
		}
		
		Persistence.init();
		localDebugOutput("### Persistence.commit()");
		RecordStore rs = null;
		try {
			// Open or create RecordStore
			try {
				rs = RecordStore.openRecordStore(Persistence.CONF_RS_NAME, true);
			} catch (RecordStoreException ex) {
				Persistence.localDebugOutput("RecordStore could not be opened! " + ex.getMessage());
				Logger.errorException(Persistence.class, ex, "RS konnte nicht geoeffnet werden.");
				return;
			}
	
			// Copy data to avoid read/write conflicts on cache.
			byte[][] toStore = new byte[Persistence.NUM][];
			for (int i = 0; i < NUM; i++) {
				if (Persistence.dirty[i] && Persistence.cache[i] != null) {
					toStore[i] = Persistence.cache[i];
					Persistence.dirty[i] = false;
				} else {
					toStore[i] = null;
				}
			}
	
			// Store data to RecordStore
			for (int i = 0; i < NUM; i++) {
				if (toStore[i] != null) {
					Persistence.writeToRecordStore(i, toStore[i], rs);
				}
			}
	
			// Close RecordStore
			try {
				rs.closeRecordStore();
			} catch (RecordStoreException ex) {
				Persistence.localDebugOutput("RecordStore could not be closed! " + ex.getMessage());
				Logger.errorException(Persistence.class, ex, "en0ote9www");
			}
		
		} catch (Exception ex) {
			Logger.errorException(Persistence.class, ex, "dsfih3pss98os");
			emergencyCloseRS(rs);
		}
		
		localDebugOutput("### Persistence.commit() DONE\n");
	}

	/**
	 * Stores data to Persistence. (Write is delayed through cache!)
	 * 
	 * @param slot
	 *            constant (Persistence.SLOT_*)
	 */
	protected synchronized static void set(int slot, byte[] data, int commitDelay) {
		Persistence.init();

		// Check slot number
		if (slot > Persistence.NUM) {
			Persistence.localDebugOutput("Unkown slot number, please use constants in Persistence! Slot = "
				+ slot);
			return;
		}

		// Cachen und speichern
		Persistence.cache[slot] = data;
		Persistence.dirty[slot] = true;
		Persistence.loaded[slot] = true;

		// Commit wird in einem extra Thread, evtl. mit Verzögerung aufgerufen.
		if (commitThread == null) {
			commitThread = new CommitThread();
			commitThread.setCommitDelay(commitDelay);
			commitThread.start();
		} else {
			commitThread.setCommitDelay(commitDelay);
			synchronized (commitThread) {
				commitThread.notify();
			}
		}
	}

	/**
	 * Returns the data in the given slot. This method is cached!
	 * 
	 * @param slot
	 *            constant (Persistence.SLOT_*)
	 * @return data byte[]
	 */
	protected static byte[] get(int slot) {
		Persistence.init();

		// Load to cache if necessary
		if (!Persistence.loaded[slot]) {
			Persistence.localDebugOutput("Need to load data from slot = " + slot);
			Persistence.cache[slot] = Persistence.getFromRecordStore(slot);
			Persistence.loaded[slot] = true;
			Persistence.dirty[slot] = false;
		} else {
			Persistence.localDebugOutput("Found data in Cache for slot = " + slot);
		}

		return Persistence.cache[slot];
	}

	/**
	 * Loescht den gesamten RecordStore, mit allen Daten.
	 */
	public static void deleteRecordStore() {
		isInitialized = false;
		try {
			// Get all record stores we own
			String[] recordStores = RecordStore.listRecordStores();

			// First: Löscht den aktuell verwendeten RecordStore
			if (StringUtils.containsFull(RecordStore.listRecordStores(), Persistence.CONF_RS_NAME)) {
				RecordStore.deleteRecordStore(Persistence.CONF_RS_NAME);
				Persistence.localDebugOutput("Deleted Record current Store");
			}

			// Second: Löscht alle RecordStores auf die dieses Midlet zugreifen kann. Das machen wir damit wir
			// alte RS, aus alten Versionen löschen.
			recordStores = RecordStore.listRecordStores();
			if (recordStores == null) {
				return;
			}
			for (int i = 0; i < recordStores.length; i++) {
				RecordStore.deleteRecordStore(recordStores[i]);
				Persistence.localDebugOutput("Deleted Record Store with name = " + recordStores[i]);
			}
		} catch (RecordStoreException ex) {
			Persistence.localDebugOutput("RecordStore could not be deleted! " + ex.getMessage());
			Logger.errorException(Persistence.class, ex, "RS konnte nicht geloescht werden.");
		}

		// Third: Clear Caches
		clearAllCaches();
	}

	private static byte[] getFromRecordStore(int slot) {
		// Wir können nur laden, wenn es ein Midlet ist.
		if(!Global.ISMIDLET) {
			return new byte[0];
		}
		
		RecordStore rs = null;
		try {
			// Open the RS if one RS with this name exists.
			if (StringUtils.containsFull(RecordStore.listRecordStores(), Persistence.CONF_RS_NAME)) {
				rs = RecordStore.openRecordStore(Persistence.CONF_RS_NAME, false);
			} else {
				Persistence.localDebugOutput("RS not existing. Slot = " + slot);
				return new byte[0];
			}

			// Number of Records
			int num = rs.getNumRecords();

			// Read data from the RS
			byte[] out = new byte[0];
			if (slot <= num) {
				out = rs.getRecord(slot);
			}

			// Close RS
			rs.closeRecordStore();

			// Messaging
			if (out == null || out.length == 0) {
				out = new byte[0];
				Persistence.localDebugOutput("No Data in this RS Slot = " + slot);
			} else {
				Persistence.localDebugOutput(out.length + " Bytes read from Slot = " + slot);
			}

			// Return
			return out;

		} catch (RecordStoreFullException ex) {
			Persistence.localDebugOutput("Record Store full! (" + ex.toString() + ")");
			Logger.errorException(Persistence.class, ex, "5erpojw");
			emergencyCloseRS(rs);
			return new byte[0];

		} catch (RecordStoreNotOpenException ex) {
			Persistence.localDebugOutput("Could not open RecordStore. (" + ex.toString() + ")");
			Logger.errorException(Persistence.class, ex, "slfo8j3");
			emergencyCloseRS(rs);
			return new byte[0];

		} catch (InvalidRecordIDException ex) {
			Persistence.localDebugOutput("Record Id not found. (" + ex.toString() + ")");
			Logger.errorException(Persistence.class, ex, "awera3refara3e");
			emergencyCloseRS(rs);
			return new byte[0];

		} catch (RecordStoreException ex) {
			Persistence.localDebugOutput("Unknown RecordStoreException. (" + ex.toString() + ")");
			Logger.errorException(Persistence.class, ex, "adq3wreda3wer");
			emergencyCloseRS(rs);
			return new byte[0];

		} catch (Throwable ex) {
			Persistence.localDebugOutput("Throwable: " + ex.toString());
			Logger.errorException(Persistence.class, ex, "esrl983frj");
			emergencyCloseRS(rs);
			return new byte[0];
		}
	}
	
	/**
	 * Diese Methode soll sicherstellen das der RecordStore auch im Fehlerfall jeweils richtig geschlossen
	 * wird und beim nachfolgenden Öffnen keine Probleme auftreten.
	 * 
	 * @param RecordStore to close
	 */
	private static void emergencyCloseRS(RecordStore rs) {
		try {
			if(rs != null) {
				rs.closeRecordStore();
			}
		} catch(Throwable rsex) {
			// Wir wollen diese Exception nicht weiter werfen, weil wenn es bei Schließen
			// eine gibt stehen die Chancen gut das es schon zu ist, also nicht mehr geschlossen
			// werden kann...
			Persistence.localDebugOutput("Emergency close of RecordStore failed! " + rsex.getMessage());
			Logger.errorException(Persistence.class, rsex, "dsfih3pRE98os");
		}
	}

	private static boolean writeToRecordStore(int slot, byte[] obj, RecordStore rs) {
		try {
			// Caculate numer of Lines
			int num = rs.getNumRecords();
			
			// Check RS Structure
			if (slot > num) {
				// Need to build RecordStore structure
				Persistence.localDebugOutput("Building RecordStore:");
				while (slot > num) {
					rs.addRecord(new byte[0], 0, 0);
					num = rs.getNumRecords();
					Persistence.localDebugOutput(" + Building Slot " + num);
				}
			}

			// Write
			if (slot <= num) {
				rs.setRecord(slot, obj, 0, obj.length);
				Persistence.localDebugOutput("Wrote " + obj.length + " Bytes to Slot " + slot);
				Persistence.localDebugOutput("Size=" + rs.getSize() + " Byte");
				Persistence.localDebugOutput("SizeAvailable=" + rs.getSizeAvailable() + " Byte");
			} else {
				Persistence.localDebugOutput("ERROR: RS Structure is not valid!");
			}

			return true;

		} catch (RecordStoreNotFoundException ex) {
			Persistence.localDebugOutput("Could not find RecordStore. (" + ex.toString() + ")");
			Logger.errorException(Persistence.class, ex, "Rd.illo83i");
			return false;

		} catch (RecordStoreNotOpenException ex) {
			Persistence.localDebugOutput("Could not open RecordStore. (" + ex.toString() + ")");
			Logger.errorException(Persistence.class, ex, "Rd.illssdweo83i");
			return false;

		} catch (RecordStoreException ex) {
			Persistence.localDebugOutput("RecordStore error in set2RecordStore " + ex.getMessage());
			Logger.errorException(Persistence.class, ex, "Rd.ill242o83i");
			return false;
		}
	}

	/**
	 * Beendet den Record-Store Thread regulär.
	 */
	public static void finishRSThread() {
		commit();
		if (commitThread != null) {
			commitThread.finishThread();
		}
	}

	private static void localDebugOutput(String msg) {
		if (Persistence.LOCAL_DEBUG_OUTPUT && msg != null && msg.trim().length() > 0) {
			System.out.println("# " + msg);
		}
	}
}

/**
 * Dieser Thread schreibt Einstellungs- und Daten-Aenderungen in den Record-Store.
 */
class CommitThread extends Thread {

	private int commitDelay = Persistence.RS_COMMIT_IMMEDIATLY;
	private boolean finishFlag = false;

	/**
	 * Zeit nach welcher die Aenderung in den RS geschrieben wird. Methode muss aufgerufen werden, bevor das
	 * Objekt per notify benachrichtigt wird.
	 * 
	 * @param timeBeforeWriting
	 *            Zeit in ms
	 */
	public void setCommitDelay(int timeBeforeWriting) {
		this.commitDelay = timeBeforeWriting;
	}

	public void finishThread() {
		finishFlag = true;
		synchronized (this) {
			this.notify();
		}
	}

	public void run() {
		while (true) {
			// Wartet falls gewünscht mit dem Schreiben
			if (commitDelay != Persistence.RS_COMMIT_IMMEDIATLY) {
				try {
					Thread.sleep(commitDelay);
				} catch (InterruptedException ie) {
					ie.printStackTrace();
				}
			}

			// Commited die Daten
			Persistence.commit();

			// Wartet bis zum nächsten Notify
			try {
				synchronized (this) {
					this.wait();
				}
			} catch (InterruptedException ie) {
				ie.printStackTrace();
			}

			// Beendet den Thread, falls gewünscht
			if (finishFlag) {
				break;
			}
		}
	}
}
