// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package gui;

import gui.utils.GfxUtils;

import javax.microedition.lcdui.Font;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;

/**
 * Fontmetrics erstellt aus allen möglichen Schriftgrößen eine Tabelle für Ober- / Mittel- /
 * Unterlänge. Dazu zeichnet es die Schrift auf einen Canvas und untersucht die Pixel
 */
public final class FontMetrics {
	// Abstand von oben bis zur Schrift
	private static int[] topSpace = { -1, -1, -1 };
	// Oberlänge
	private static int[] topHeight = { -1, -1, -1 };
	// Mittellänge
	private static int[] midHeight = { -1, -1, -1 };
	// Unterlänge
	private static int[] bottomHeight = { -1, -1, -1 };
	// Abstand von unten bis zur Schrift
	private static int[] bottomSpace = { -1, -1, -1 };
	// Komplette Höhe (Font.getHeight()) 
	private static int[] totalHeight = { -1, -1, -1 };

	// calculated font_y_offsets 
	private static int[] font_y_offset = { 0, 0, 0 };

	private final static int SCAN_TOP2BOTTOM = 0;
	private final static int SCAN_BOTTOM2TOP = 1;

	public static final int[] sizes = { Font.SIZE_SMALL, Font.SIZE_MEDIUM, Font.SIZE_LARGE };

	private static final String[] sizesStrings = { "SMALL", "MEDIUM", "LARGE" };

	private static final int[] types = { Font.FACE_SYSTEM, Font.FACE_MONOSPACE, Font.FACE_PROPORTIONAL };

	private static final String[] typesStrings = { "SYSTEM", "MONOSPACE", "PROPORTIAL" };

	private static final int[] styles = { Font.STYLE_PLAIN, Font.STYLE_BOLD, Font.STYLE_ITALIC, Font.STYLE_UNDERLINED };

	private static final String[] stylesStrings = { "PLAIN", "BOLD", "ITALIC", "UNDERLINED" };

	private static int sizeType2array(int sizeType) {
		switch (sizeType) {
		case Font.SIZE_SMALL:
			return 0;
		case Font.SIZE_MEDIUM:
			return 1;
		case Font.SIZE_LARGE:
			return 2;
		default:
			return -1;
		}
	}

	private static void clearImage(Image i, Graphics g) {
		g.setColor(Color.BLACK);
		g.fillRect(0, 0, i.getWidth(), i.getHeight());
	}

	private static int scanBlack(Image im, int direction) {
		int[] data = new int[(im.getHeight() - 2) * (im.getWidth() - 2)];
		im.getRGB(data, 0, im.getWidth() - 2, 1, 1, im.getWidth() - 2, im.getHeight() - 2);
		int pos;
		if (direction == SCAN_TOP2BOTTOM) {
			pos = 0;
			for (int y = 0; y < im.getHeight() - 2; y++) {
				for (int x = 0; x < im.getWidth() - 2; x++) {
					if (((data[pos++]) & 0x000000FF) > 128) {
						return y;
					}
				}
			}
		} else {
			pos = data.length - 1;
			for (int y = im.getHeight() - 1 - 2; y >= 0; y--) {
				for (int x = im.getWidth() - 1 - 2; x >= 0; x--) {
					if ((((data[pos--]) & 0x000000FF)) > 128) {
						return y;
					}
				}
			}
		}
		return -1;
	}

	public static Image createImage(String text, Font font) {
		int lineHeight = font.getHeight();
		int maxWidth = font.stringWidth(text);
		Image im = Image.createImage(maxWidth + 2, lineHeight + 2);
		Graphics g = im.getGraphics();
		g.setColor(Color.BLACK);
		g.fillRect(0, 0, im.getWidth(), im.getHeight());
		g.setColor(Color.WHITE);
		g.setFont(font);
		g.drawString(text, 1, 1, Graphics.LEFT | Graphics.TOP);
		return im;
	}
	
	public static Image createImage2(String text, Font font) {
		int lineHeight = font.getHeight();
		int maxWidth = font.stringWidth(text);
		Image im = Image.createImage(maxWidth + 2, lineHeight * 2);
		Graphics g = im.getGraphics();
		g.setColor(Color.BLACK);
		g.fillRect(0, 0, im.getWidth(), im.getHeight());
		g.setColor(Color.WHITE);
		g.setFont(font);
		g.drawString(text, 1, 0, Graphics.LEFT | Graphics.TOP);
		return im;
	}
	
	public static void init(Graphics debug_graphics) {
		int pos_x = 0;
		int pos_y = 0;
		if (debug_graphics != null) {
			debug_graphics.setColor(Color.BLACK);
			pos_y += drawText(debug_graphics, "Schrifttest: FontMetrics", pos_y);
		}

		
		for (int size = 0; size < 3; size++) {
			Font font;
			font = Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, sizes[size]);
			int lineHeight = font.getHeight();
			
			
			Image imTd = createImage("Td", font);
			int scanValue = scanBlack(imTd, SCAN_TOP2BOTTOM);
			
			int topSpace = scanValue;
			if (debug_graphics != null) {
				debug_graphics.drawImage(imTd, pos_x, pos_y, Graphics.LEFT | Graphics.TOP);
				pos_x += imTd.getWidth() + 5;
			}
			
			// for safety
			if (scanValue == -1) {
				imTd = createImage("TdTdTd", font);
				topSpace = scanBlack(imTd, SCAN_TOP2BOTTOM);
				if (debug_graphics != null) {
					debug_graphics.drawImage(imTd, pos_x, pos_y, Graphics.LEFT | Graphics.TOP);
					pos_x += imTd.getWidth() + 5;
				}
			}

			Image imO = createImage("o", font);
			int scanValueTop = scanBlack(imO, SCAN_TOP2BOTTOM);
			int scanValueMid = scanBlack(imO, SCAN_BOTTOM2TOP);
			int topHeight = scanValueTop - topSpace;
			int midHeight = scanValueMid - topSpace - topHeight + 1;
			if (debug_graphics != null) {
				debug_graphics.drawImage(imO, pos_x, pos_y, Graphics.LEFT | Graphics.TOP);
				pos_x += imO.getWidth() + 5;
			}
			
			// for safety
			if (scanValueTop == -1 || scanValueMid == -1) {
				imO = createImage("ooo", font);
				topHeight = scanBlack(imO, SCAN_TOP2BOTTOM) - topSpace;
				midHeight = scanBlack(imO, SCAN_BOTTOM2TOP) - topSpace - topHeight + 1;
				if (debug_graphics != null) {
					debug_graphics.drawImage(imO, pos_x, pos_y, Graphics.LEFT | Graphics.TOP);
					pos_x += imO.getWidth() + 5;
				}
			}

			Image imY = createImage("y", font);
			int scanValueBottom = scanBlack(imY, SCAN_BOTTOM2TOP);
			int bottomHeight = scanValueBottom - topSpace - topHeight - midHeight + 1;
			int bottomSpace = lineHeight - topSpace - topHeight - midHeight - bottomHeight;
			if (debug_graphics != null) {
				debug_graphics.drawImage(imY, pos_x, pos_y, Graphics.LEFT | Graphics.TOP);
				pos_x += imY.getWidth() + 5;
			}

			// for safety
			if (scanValueBottom == -1) {
				imY = createImage("yyy", font);
				bottomHeight = scanBlack(imY, SCAN_BOTTOM2TOP) - topSpace - topHeight - midHeight + 1;
				bottomSpace = lineHeight - topSpace - topHeight - midHeight - bottomHeight;
				if (debug_graphics != null) {
					debug_graphics.drawImage(imY, pos_x, pos_y, Graphics.LEFT | Graphics.TOP);
					pos_x += imY.getWidth() + 5;
				}

				// um ganz sicher zu gehen
				if (bottomHeight < 0) {
					bottomHeight = lineHeight - topHeight - topHeight - midHeight;
					bottomSpace = 0;

					int maxWidthX = font.stringWidth("x");
					Image imX = Image.createImage(maxWidthX, lineHeight);
					Graphics gX = imX.getGraphics();
					gX.setFont(font);
					clearImage(imX, gX);
					gX.setColor(Color.WHITE);
					gX.drawString("x", 0, 0, Graphics.LEFT | Graphics.TOP);
					if (debug_graphics != null) {
						debug_graphics.drawImage(imX, pos_x, pos_y, Graphics.LEFT | Graphics.TOP);
						pos_x += imX.getWidth() + 5;
					}
				}
			}

			// Erkennung falls die Handyschrift oben bündig ist und Berechnung der Y-Korrekturwerte
			if (topSpace == 0) {
				int newBottomSpace = ((lineHeight - (topHeight + midHeight)) / 2) - bottomHeight;

				if (newBottomSpace < 0) {
					newBottomSpace = 0;
				}

				int newUpperSpace = (lineHeight - (topHeight + midHeight + bottomHeight + newBottomSpace));

				font_y_offset[size] = +newUpperSpace;
				if (newUpperSpace != 0) {
					GfxUtils.activateFontYOffset();
				}
			}
			
			
			FontMetrics.topSpace[size] = topSpace + font_y_offset[size];
			FontMetrics.bottomSpace[size] = bottomSpace - font_y_offset[size];
			FontMetrics.topHeight[size] = topHeight;
			FontMetrics.midHeight[size] = midHeight;
			FontMetrics.bottomHeight[size] = bottomHeight;
			FontMetrics.totalHeight[size] = lineHeight;

			if (debug_graphics != null) {
				pos_x = 0;
				pos_y += lineHeight;
				debug_graphics.setColor(Color.BLACK);
				pos_y += drawText(debug_graphics, "Height: " + lineHeight + "  (" + (topHeight + midHeight + bottomHeight) + ", " + font_y_offset[size] + ")", pos_y);
				pos_y += drawText(debug_graphics, "H = " + topSpace + "+" + topHeight + "+" + midHeight + "+" + bottomHeight + "+" + bottomSpace, pos_y);
			}
			/*				
							System.out.println("Size:" + size);
							System.out.println("TopSpace:    "+topSpace);
							System.out.println("BottomSpace: "+bottomSpace);
							System.out.println("High:        "+topHeight);
							System.out.println("Mid:         "+midHeight);
							System.out.println("Low:         "+bottomHeight);			
							System.out.println("TotalHeight: "+textHeight);
							*/
		}
		if (debug_graphics != null) {
			// TESTING
			pos_y += 3;
			for (int size = 0; size < 3; size++) {
				pos_x = 0;
				Font font;
				font = Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, sizes[size]);
				Image im = createImage2("y", font);
				if (debug_graphics != null) {
					debug_graphics.drawImage(im, pos_x, pos_y, Graphics.LEFT | Graphics.TOP);
					pos_x += im.getWidth() + 5;
				}

				im = createImage2("yy", font);
				if (debug_graphics != null) {
					debug_graphics.drawImage(im, pos_x, pos_y, Graphics.LEFT | Graphics.TOP);
					pos_x += im.getWidth() + 5;
				}

				im = createImage2("yyy", font);
				if (debug_graphics != null) {
					debug_graphics.drawImage(im, pos_x, pos_y, Graphics.LEFT | Graphics.TOP);
					pos_x += im.getWidth() + 5;
				}

				im = createImage2(" y ", font);
				if (debug_graphics != null) {
					debug_graphics.drawImage(im, pos_x, pos_y, Graphics.LEFT | Graphics.TOP);
					pos_x += im.getWidth() + 5;
				}

				im = createImage("g", font);
				if (debug_graphics != null) {
					debug_graphics.drawImage(im, pos_x, pos_y, Graphics.LEFT | Graphics.TOP);
					pos_x += im.getWidth() + 5;
				}

				im = createImage("gg", font);
				if (debug_graphics != null) {
					debug_graphics.drawImage(im, pos_x, pos_y, Graphics.LEFT | Graphics.TOP);
					pos_x += im.getWidth() + 5;
				}

				im = createImage("ggg", font);
				if (debug_graphics != null) {
					debug_graphics.drawImage(im, pos_x, pos_y, Graphics.LEFT | Graphics.TOP);
					pos_x += im.getWidth() + 5;
				}

				im = createImage2("g", font);
				if (debug_graphics != null) {
					debug_graphics.drawImage(im, pos_x, pos_y, Graphics.LEFT | Graphics.TOP);
					pos_x += im.getWidth() + 5;
				}

				im = createImage2("gg", font);
				if (debug_graphics != null) {
					debug_graphics.drawImage(im, pos_x, pos_y, Graphics.LEFT | Graphics.TOP);
					pos_x += im.getWidth() + 5;
				}

				pos_y += im.getHeight();
			}

			// default text
			pos_y += 3;
			for (int size = 0; size < 3; size++) {
				Font font;
				font = Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, sizes[size]);
				debug_graphics.setFont(font);

				debug_graphics.setColor(Color.BLACK);
				debug_graphics.fillRect(0, pos_y, debug_graphics.getClipWidth(), font.getHeight());

				debug_graphics.setColor(Color.WHITE);
				pos_y += drawText(debug_graphics, "Toy ABCDJ öüä ÄÜÖ gkl", pos_y);
				pos_y += 1;
			}
		}
	}

	private static int drawText(Graphics g, String s, int y) {
		Font f;
		f = g.getFont();
		g.drawString(s, 2, y, Graphics.LEFT | Graphics.TOP);
		return f.getHeight();
	}

	public static void init() {
		init(null);
	}

	public static void drawText(Graphics g) {
		String string = "Toyd";
		for (int size = 0; size < 3; size++) {
			Font font;
			font = Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, sizes[size]);
			g.setFont(font);
			int xBase = 140;
			int yBase = 130;
			int yOffset = 0;
			for (int i = 0; i < (size); i++) {
				yOffset += 1 + totalHeight[size];
			}
			int yPos = yBase + yOffset;
			g.setColor(Color.gray2color(180));
			g.fillRect(xBase, yPos, font.stringWidth(string), totalHeight[size]);
			System.out.println("xy: (" + xBase + "," + yPos + ")");
			System.out.println("h:" + totalHeight[size]);
			g.setColor(Color.BLACK);
			g.drawString(string, xBase, yPos, Graphics.LEFT | Graphics.TOP);
		}
	}

	public static int getBottomSpace(int fontSize) {
		return bottomSpace[sizeType2array(fontSize)];
	}

	public static int getTopSpace(int fontSize) {
		return topSpace[sizeType2array(fontSize)];
	}

	public static int getHeight(int fontSize) {
		return totalHeight[sizeType2array(fontSize)];
	}

	public static int getMidHeight(int fontSize) {
		return midHeight[sizeType2array(fontSize)];
	}

	public static int getTopHeight(int fontSize) {
		return topHeight[sizeType2array(fontSize)];
	}

	public static int getBottomHeight(int fontSize) {
		return bottomHeight[sizeType2array(fontSize)];
	}

	/**
	 * Echte Schrifthöhe ohne oberen und unteren Rand
	 * 
	 * @param fontSize -
	 *            Schriftgröße (z.B. Font.SIZE_MEDIUM)
	 * @return Hoehe in Pixeln
	 */
	public static int getRealHeight(int fontSize) {
		return topHeight[sizeType2array(fontSize)] + midHeight[sizeType2array(fontSize)] + bottomHeight[sizeType2array(fontSize)];
	}

	public static Font getFont(int style, int size) {
		return Font.getFont(Font.FACE_SYSTEM, style, size);
	}

	public static Font getFont(int size) {
		return Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, size);
	}

	public static String getFontNameById(int type) {
		return typesStrings[type];
	}

	public static String getFontSizeById(int size) {
		return sizesStrings[size];
	}

	public static String getFontStyleById(int style) {
		return stylesStrings[style];
	}

	public static Font getFontById(int type, int style, int size) {
		return Font.getFont(types[type], styles[style], sizes[size]);
	}

	public static int getFontYOffset(int size) {
		return font_y_offset[sizeType2array(size)];
	}

	public static int getBaseLine(int size) {
		int a = sizeType2array(size);
		return totalHeight[a] - bottomSpace[a] - bottomHeight[a];
	}
	
	public static int getIdBySize(int size) {
		if (size == Font.SIZE_SMALL) {
			return 0;
		}
		if (size == Font.SIZE_MEDIUM) {
			return 1;
		}
		if (size == Font.SIZE_LARGE) {
			return 2;
		}
		return -1;
	}

}
