// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package gui.guis;

import gui.Imager;
import gui.items.Item;
import gui.items.ItemDetail;
import gui.items.ItemDetailHaltestelle;
import gui.items.ItemDetailHaltestellenPaar;
import gui.items.ItemDetailLaufweg;
import gui.items.ItemDetailLinie;
import gui.items.ItemDetailModifiedWalkingSpeed;
import gui.items.ItemDetailNotOptimized;
import gui.items.ItemDetailPricing;
import gui.items.ItemDetailWartezeit;

import java.util.Vector;

import javax.microedition.lcdui.Canvas;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.game.GameCanvas;

import model.DatumZeit;
import model.Route;
import model.RoutenAbschnitt;
import control.Global;
import control.Keys;
import control.locale.Locale;
import control.reminder.Reminder;

public class GuiDetail extends GuiView {
	private Route route;
	private boolean reminderAvailable = false;
	private boolean showKompakt = true;
	

	public GuiDetail(GameCanvas gc, Graphics g) {
		super(Locale.Route, gc, g);
		setRightFooterIcon(Imager.getFooterIcon(Imager.ICON_FOOTER_LISTE));
	}

	public void updateReminderString() {
		if (Reminder.yetExists(route)) {
			int zeit = Reminder.getMinutenVorAbfahrt();
			setLeftFooterString(Locale.replace(Locale.ReminderActive, 1, zeit + ""), Imager.getFooterIcon(Imager.ICON_FOOTER_REMINDER));
		} else {
			if (!isRemindable()) {
				setLeftFooterString("", null);
			} else {
				setLeftFooterString(Locale.ReminderActivate, Imager.getFooterIcon(Imager.ICON_FOOTER_REMINDER));
			}
		}
	}

	public boolean isRemindable() {
		return reminderAvailable;
	}

	public void activate(final Route route) {
		this.route = route;
		if (this instanceof GuiDetailReminder) {
			disableReminder();
		} else {
			if (Reminder.isRemindable(route)) {
				enableReminder();
			} else {
				disableReminder();
			}
			// force override when selected in Globals
			if (Global.REMINDER_ALWAYS_VISIBLE && (!isRemindable())) {
				enableReminder();
			}
		}

		updateReminderString();
		
		showKompakt = true;
		initItems(showKompakt);
	}

	private void initItems(boolean kompakt) {
		RoutenAbschnitt routenAbschnitte[] = route.getKompakteRoutenAbschnitte();

		if (routenAbschnitte.length == 0) {
			setItems(new Item[0]);
			return;
		}

		Vector v = new Vector();

		int lastValidId = -1;

		// Items aus Route erstellen

		if (route.isOptimierungAbgebrochen()) {
			v.addElement(new ItemDetailNotOptimized());
			System.out.println("adding Item");
		}

		for (int i = 0; i < routenAbschnitte.length; i++) {
			if (routenAbschnitte[i].isLaufweg()) {
				if (i == 0) {
					ItemDetail itemHaltestelle = new ItemDetailHaltestelle(routenAbschnitte[i].getVon());
					v.addElement(itemHaltestelle);
				}

				ItemDetail item = new ItemDetailLaufweg(routenAbschnitte[i]);
				v.addElement(item);
				lastValidId = -1;

				if (i == routenAbschnitte.length - 1) {
					ItemDetail itemHaltestelle = new ItemDetailHaltestelle(routenAbschnitte[i].getNach());
					v.addElement(itemHaltestelle);
				}

				continue;
			}

			if (lastValidId != -1) {
				Item item = new ItemDetailWartezeit(routenAbschnitte[lastValidId], routenAbschnitte[i]);
				v.addElement(item);
			}

			ItemDetail item = new ItemDetailLinie(routenAbschnitte[i]);
			v.addElement(item);
			if (kompakt) {
				item = new ItemDetailHaltestellenPaar(routenAbschnitte[i], route.getKompakteRoutenAbschnitte());
			} else {
				item = new ItemDetailHaltestellenPaar(routenAbschnitte[i], route.getRoutenAbschnitteWithLoops());
			}
			v.addElement(item);
			lastValidId = i;
		}
		
		if (Global.getLaufzeitFaktor() != 10) {
			v.addElement(new ItemDetailModifiedWalkingSpeed());
		}
		
		v.addElement(new ItemDetailPricing(route));

		Item[] items = new Item[v.size()];
		v.copyInto(items);

		setItems(items);
	}

	private void disableReminder() {
		reminderAvailable = false;
	}

	private void enableReminder() {
		reminderAvailable = true;
	}

	public int getDiffernzeitAbfahrt() {
		return route.getAbfahrtszeit() - DatumZeit.Now();
	}

	public Route getRoute() {
		return route;
	}

	/*--------------------------------------------------
	 * Key code event handling
	 *-------------------------------------------------*/
	public boolean keyPressed(int keyCode, Canvas canvas) {
		boolean superHandled = super.keyPressed(keyCode, canvas);
		if (superHandled) {
			return superHandled;
		}
		if (Keys.isKey(keyCode, Keys.MENU_RIGHT)) {
			showKompakt = !showKompakt;
			initItems(showKompakt);
			
			if (isVisible()) {
				drawTo(getGameCanvasGraphics());
			}
			
			return true;
		}
		return false;
	}

	public boolean pointerPressed(int x, int y) {
		boolean superHandled = super.pointerPressed(x, y);
		if (superHandled) {
			return superHandled;
		}
		if (y >= getMainY() && y < getMainY() + getMainHeight()) {
			showKompakt = !showKompakt;
			initItems(showKompakt);
			
			if (isVisible()) {
				drawTo(getGameCanvasGraphics());
			}
			
			return false;
		}
		return false;
	}

	public void scrollUp() {
		keyPressed(Keys.getKeyCode(Keys.UP), getGameCanvas());
	}

	public void scrollDown() {
		keyPressed(Keys.getKeyCode(Keys.DOWN), getGameCanvas());
	}
	
}
