// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package gui.guis;

import gui.FontMetrics;
import gui.Manager;
import gui.TimerTaskGuiUpdate;
import gui.utils.GfxUtils;

import java.util.Timer;

import javax.microedition.lcdui.Font;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.game.GameCanvas;

import control.TimerTaskRestart;

public class GuiNewDataset extends Gui implements iCallbackTimer {

	public GuiNewDataset(GameCanvas gc, Graphics g) {
		super("neuer Datensatz", gc, g);
		setRightFooterString("", null);
	}

	public static final int STATE_UNKNOWN = -1;
	public static final int STATE_PRE_DOWNLOAD = 0;
	public static final int STATE_DOWNLOAD = 1;
	public static final int STATE_RESTART = 2;
	public static final int STATE_ERROR = 3;
	public static final int STATE_INVALID_VERSION = 4;
	private int state = STATE_UNKNOWN;

	private Timer timer;

	synchronized public void drawTo(Graphics g) {

		int y_base = getMainY();

		int mainHeight = getMainHeight();

		// clear Background
		if (isVisible()) {
			g.setColor(Manager.ITEMS_COLOR_BACKGROUND);
			gui.utils.GfxUtils.fillFullRectangle(g, y_base, mainHeight + Manager.MAIN_INNER_SPACE);
		}

		int drawWidth = Manager.SCREEN_WIDTH - Manager.MAIN_LEFT_SPACE * 2;
		Font f = FontMetrics.getFont(Manager.ITEMS_FONT_SIZE);
		int x = Manager.MAIN_LEFT_SPACE;
		int y = y_base + Manager.MAIN_INNER_SPACE;
		g.setFont(f);

		g.setColor(Manager.ITEMS_COLOR_FONT);
		String[] output;
		String[] lines;
		switch (state) {
		case STATE_RESTART:
			if (downloaded == filesize && downloaded != 0) {
				output = new String[] { "Datensatz wird heruntergeladen", "", (downloaded / 1024) + "kb von " + (filesize / 1024) + "kb" };
				lines = GfxUtils.splitupInLines(f, output, drawWidth);
				GfxUtils.drawText(g, lines, x, y);

				y += (lines.length + 1) * f.getHeight();
			}
			setLeftFooterString("", null);
			output = new String[] { "aktiviere neuen Datensatz"};
			lines = GfxUtils.splitupInLines(f, output, drawWidth);
			GfxUtils.drawText(g, lines, x, y);
			y += (lines.length) * f.getHeight() + f.getHeight() / 2;
			GfxUtils.drawHCenteredText(g, "... bitte warten ...", y);
			break;
		case STATE_PRE_DOWNLOAD:
			setLeftFooterString("abbrechen", null);
			output = new String[] { "Fordere neuen Datensatz vom Server an."};
			lines = GfxUtils.splitupInLines(f, output, drawWidth);
			GfxUtils.drawText(g, lines, x, y);
			y += (lines.length) * f.getHeight() + f.getHeight() / 2;
			GfxUtils.drawHCenteredText(g, "... bitte warten ...", y);
			break;
		case STATE_DOWNLOAD:
			setLeftFooterString("abbrechen", null);
			output = new String[] { "Datensatz wird heruntergeladen", "", (downloaded / 1024) + "kb von " + (filesize / 1024) + "kb" };
			lines = GfxUtils.splitupInLines(f, output, drawWidth);
			GfxUtils.drawText(g, lines, x, y);
			break;
		case STATE_ERROR:
			setLeftFooterString("zurück", null);
			setRightFooterString("", null);
			output = new String[] { "Beim herunterladen der Datei ist leider ein Fehler aufgetreten. Bitte überprüfe deine Internetverbindung und versuche es später nocheinmal." };
			lines = GfxUtils.splitupInLines(f, output, drawWidth);
			GfxUtils.drawText(g, lines, x, y);
			break;
		case STATE_INVALID_VERSION:
			setLeftFooterString("zurück", null);
			setRightFooterString("", null);
			output = new String[] { "Es ist eine neue Version von push&ride verfügbar. Bitte aktualisiere zuerst push&ride über den Market." };
			lines = GfxUtils.splitupInLines(f, output, drawWidth);
			GfxUtils.drawText(g, lines, x, y);
			break;
		}

		repaintFooter();

		if (isVisible()) {
			gc.repaint(0, y_base, Manager.SCREEN_WIDTH, mainHeight + Manager.MAIN_INNER_SPACE);
		}
	}

	private int downloaded;
	private int filesize;

	public void triggerRestart() {
		cancelUpdateTimer();
		state = STATE_RESTART;
		if (isVisible()) {
			drawTo(getGameCanvasGraphics());
		}
		new Timer().schedule(new TimerTaskRestart(), 3000);
	}

	public void triggerDownload() {
		state = STATE_PRE_DOWNLOAD;
		if (isVisible()) {
			drawTo(getGameCanvasGraphics());
		}
	}

	public void startDownload(int fileSize) {
		downloaded = 0;
		this.filesize = fileSize;
		state = STATE_DOWNLOAD;
		if (isVisible()) {
			drawTo(getGameCanvasGraphics());
		}
		timer = new Timer();
		timer.schedule(new TimerTaskGuiUpdate(this), 500, 500);
	}

	public void updateDownload(int downloaded) {
		this.downloaded = downloaded;
	}

	public void cancelUpdateTimer() {
		if (timer != null) {
			timer.cancel();
		}
	}

	public void error() {
		cancelUpdateTimer();
		state = STATE_ERROR;
		if (isVisible()) {
			drawTo(getGameCanvasGraphics());
		}
	}

	public void invalidVersion() {
		cancelUpdateTimer();
		state = STATE_INVALID_VERSION;
		if (isVisible()) {
			drawTo(getGameCanvasGraphics());
		}
	}

	synchronized public void redraw() {
		if (isVisible()) {
			drawTo(getGameCanvasGraphics());
		}
	}

	synchronized public void callback() {
		if (state == STATE_DOWNLOAD) {
			if (isVisible()) {
				drawTo(getGameCanvasGraphics());
			}
		}
	}

	public int getState() {
		return state;
	}

	public void cancel() {
		cancelUpdateTimer();
	}
}
