// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package gui.guis;

import gui.Cursor;
import gui.FontMetrics;
import gui.ICursorable;
import gui.Imager;
import gui.Manager;
import gui.elements.ElementVonNachDynamic;
import gui.items.Item;
import gui.items.ItemLocation;
import gui.items.ItemLocationDescription;
import gui.items.ItemLocationDescriptionLastUsed;
import gui.items.ItemLocationLastUsed;
import gui.items.ItemLocationShowAll;
import gui.utils.GfxUtils;
import gui.utils.IntelligentKey;

import java.util.Stack;
import java.util.Vector;

import javax.microedition.lcdui.Canvas;
import javax.microedition.lcdui.Font;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;
import javax.microedition.lcdui.game.GameCanvas;

import model.AbstractLocation;
import model.Haltestelle;
import model.Ort;
import utils.StringUtils;
import control.Device;
import control.Global;
import control.Keys;
import data.Data;

abstract public class GuiSelectionLocation extends GuiSelection implements ICursorable {
	private IntelligentKey intelligentKey;
	// Hauptelemente, die auch für IntelligentKey verwendet werden
	private ItemLocation[] mainItems;
	private ItemLocation[] initItems;
	// zusätzliche Elemente wie Beschriftungstext oder gespeicherte Haltestellen, kein IntelligentKey
	private ItemLocation[] additionalItems;

	private ItemLocationDescription breakItem;

	public static boolean softKeysZweizeilig = true;

	private final int guiType;

	// FIXME: alles in ein Objekt packen
	private final Stack storedFirstPos;
	private final Stack storedSelectedPos;
	private final Stack storedOverFlowType;
	private final Stack storedVisibleItems;
	private final Stack storedAlpha;

	// Alle vier zusammengehörigen Guis
	protected static GuiSelectionLocation[] allGuis = new GuiSelectionLocation[4];

	public static final int TYPE_VON_ORT = 0;
	public static final int TYPE_VON_HALTESTELLE = 1;
	public static final int TYPE_NACH_ORT = 2;
	public static final int TYPE_NACH_HALTESTELLE = 3;

	private static final int KEY_STATE_BOTH = 0;
	private static final int KEY_STATE_NUMERIC = 1;
	private static final int KEY_STATE_ALPHA = 2;
	private int keyState = KEY_STATE_BOTH;
	
	private int lastKey = -1;
	private long lastKeyTime = -1;
	private int KEY_TIME = 400;
	private boolean lastKeyVisible = false;

	public static final AbstractLocation[] ausgewaehlteLocations = new AbstractLocation[4];

	protected static final boolean WITH_RESET = true;
	protected static final boolean WITHOUT_RESET = false;

	private static int keyHeight = 0;

	private final ElementVonNachDynamic guiElementVonNach;

	private boolean firstAfterNewItems;

	private final Cursor cursor;

	protected int getGuiType() {
		return guiType;
	}

	protected static GuiSelectionLocation getPreviousGuiSelectionLocation(int typ) {
		if (typ == TYPE_VON_ORT) {
			return null;
		} else {
			return allGuis[typ - 1];
		}
	}

	public int getKeyHeight() {
		return keyHeight;
	}

	public AbstractLocation getSelected() {
		if (selectedId == SELECTION_NONE) {
			return null;
		}
		return getSelectedItem().getLocation();
	}

	public ItemLocation getSelectedItem() {
		if (selectedId == SELECTION_NONE) {
			return null;
		}
		return (ItemLocation) getItems()[selectedId];
	}

	public int getHeaderHeight() {
		return Manager.HEADER_HEIGHT;
	}

	public GuiSelectionLocation(int typ, GameCanvas gc, Graphics g) {
		super("Haltestellen", gc, g);

		this.guiType = typ;

		guiElementVonNach = new ElementVonNachDynamic(this);

		breakItem = new ItemLocationDescription("---");

		allGuis[typ] = this;

		storedFirstPos = new Stack();
		storedSelectedPos = new Stack();
		storedVisibleItems = new Stack();
		storedOverFlowType = new Stack();
		storedAlpha = new Stack();

		cursor = new Cursor(Manager.ITEMS_FONT_SIZE, this);
	}

	/**
	 * erstellt die zusätzlichen Elemente neu - z.B. falls es neue gespeicherte
	 * Halstellen gibt
	 */
	protected void updateAdditionalItems() {
		Vector additionalItems = new Vector();

		if ((guiType == TYPE_VON_ORT) || (guiType == TYPE_NACH_ORT)) {
			Haltestelle haltestelleVon = null;
			if (guiType == TYPE_NACH_ORT) {
				haltestelleVon = (Haltestelle) getPreviousGuiSelectionLocation(getGuiType()).getSelected();
			}
			ItemLocationLastUsed lastHaltestellen[] = data.persistence.LastHaltestellenPaare.getHaltestellenItems();
			if (lastHaltestellen.length != 0) {
				additionalItems.addElement(new ItemLocationDescriptionLastUsed());
				for (int i = 0; i < lastHaltestellen.length; i++) {
					// den bereits ausgewählte Haltestelle auslassen
					if (guiType == TYPE_NACH_ORT) {
						if (haltestelleVon.equals((Haltestelle) lastHaltestellen[i].getLocation())) {
							continue;
						}
					}

					// keine Einträge doppelt eintragen
					if (additionalItems.contains(lastHaltestellen[i])) {
						continue;
					}

					additionalItems.addElement(lastHaltestellen[i]);
					if (additionalItems.size() == Global.NUM_RETUNED_LAST_HALTESTELLEN + 1) {
						break;
					}
				}
				additionalItems.addElement(breakItem);
			}
		}

		this.additionalItems = new ItemLocation[additionalItems.size()];
		additionalItems.copyInto(this.additionalItems);
	}

	/**
	 * aktualisiert die Gui/IntelligentKey mit neuen Locations
	 */
	protected void setInitItems(AbstractLocation[] locations) {
		// IntelligentKey von Ortsscreen wird nur einmal erstellt, da dort alle Orte/Haltestellen enthalten sind
		if (this instanceof GuiSelectionLocationOrte) {
			if (intelligentKey == null) {
				intelligentKey = new IntelligentKey(Data.getAlleLocations());
			} else {
				intelligentKey.clear();
			}
		} else {
			intelligentKey = new IntelligentKey(locations);
		}

		// items erstellen
		this.initItems = new ItemLocation[locations.length];

		for (int i = 0; i < locations.length; i++) {
			initItems[i] = new ItemLocation(locations[i], intelligentKey);
		}

		mainItems = initItems;
	}

	synchronized public boolean keyPressed(int keyCode, Canvas canvas) {
		if (keyCode == (int) 'ä' || keyCode == (int) 'Ä') {
			keyCode = (int) 'a';
		}
		if (keyCode == (int) 'ö' || keyCode == (int) 'Ö') {
			keyCode = (int) 'o';
		}
		if (keyCode == (int) 'ü' || keyCode == (int) 'Ü') {
			keyCode = (int) 'u';
		}
		if (keyCode == (int) 'ß') {
			keyCode = (int) 's';
		}
		
		firstAfterNewItems = false;

		if (Keys.isKey(keyCode, Keys.NUMBER) && (keyState == KEY_STATE_BOTH || keyState == KEY_STATE_NUMERIC)) {

			if (intelligentKey.addKey(keyCode)) {
				keyState = KEY_STATE_NUMERIC;

				storedFirstPos.push(new Short((short) firstId));
				storedSelectedPos.push(new Short((short) selectedId));
				storedOverFlowType.push(new Short((short) overflowType));
				storedVisibleItems.push(new Short((short) visibleItems));
				storedAlpha.push(null);

				setAllItems(WITHOUT_RESET);

				resetSelectionAndFirstPos();

				drawMain(getGameCanvasGraphics(), true);
			}
			return true;
		} else {
			if (Keys.isKey(keyCode, Keys.ALPHA) && (keyState == KEY_STATE_BOTH || keyState == KEY_STATE_ALPHA)) {
				char lowerCaseKey = Keys.getLowerCaseAlphaKey((char) keyCode);
				int intelligentKeyCode = intelligentKey.getKeyCode(lowerCaseKey);
				if (intelligentKey.addKey(intelligentKeyCode)) {
					keyState = KEY_STATE_ALPHA;
					storedFirstPos.push(new Short((short) firstId));
					storedSelectedPos.push(new Short((short) selectedId));
					storedOverFlowType.push(new Short((short) overflowType));
					storedVisibleItems.push(new Short((short) visibleItems));
					if (storedAlpha.isEmpty()) {
						storedAlpha.push("" + Keys.getLowerCaseAlphaKey((char) keyCode));
					} else {
						storedAlpha.push(((String) storedAlpha.lastElement()) + Keys.getLowerCaseAlphaKey((char) keyCode));
					}

					setAllItems(WITHOUT_RESET);

					resetSelectionAndFirstPos();

					drawMain(getGameCanvasGraphics(), true);
				}
				return true;
			} else {
				if (Keys.isKey(keyCode, Keys.C) || Keys.isKey(keyCode, Keys.MENU_RIGHT)) {
					if (intelligentKey.removeChar()) {
						storedAlpha.pop();
						if (storedAlpha.isEmpty()) {
							keyState = KEY_STATE_BOTH;
						}
						setAllItems(WITHOUT_RESET);

						firstId = ((Short) storedFirstPos.pop()).shortValue();
						selectedId = ((Short) storedSelectedPos.pop()).shortValue();
						overflowType = ((Short) storedOverFlowType.pop()).shortValue();
						visibleItems = ((Short) storedVisibleItems.pop()).shortValue();

						onChangeSelected();

						drawMain(getGameCanvasGraphics(), true);
					}
					return true;
				} else {
					return super.keyPressed(keyCode, canvas);
				}
			}
		}
	}

	private String createMatchString(String s) {
		StringBuffer sb = new StringBuffer();
		s = s.toLowerCase();
		for (int i = 0; i < s.length(); i++) {
			sb.append(IntelligentKey.replaceCharactersForIntelligentKeyStyle(s.charAt(i)));
		}
		return sb.toString();
	}

	/**
	 * kopiert mainItems und additionalItems in ein Array und übergibt es
	 * new/setItems(Items[])
	 */
	protected void setAllItems(boolean doResetSelectionAndFirstPos) {
		int visibleItems = getMainHeight() / breakItem.getHeight();

		int additionalItemLength = additionalItems.length;

		// Stage 1 - Daten holen je nach Eingabetyp
		
		if (keyState == KEY_STATE_BOTH) {
			mainItems = initItems;
		} else {
			additionalItemLength = 0;
		}

		if (keyState == KEY_STATE_NUMERIC) {
			mainItems = intelligentKey.getItemLocations();
		}

		if (keyState == KEY_STATE_ALPHA) {
			// zuerst Daten aus IntelligentKey holen, dann weiter filtern
			if (intelligentKey.hasValidEntries()) {
				mainItems = intelligentKey.getItemLocations();
				String searchPattern = (String) storedAlpha.lastElement();
				Vector v = new Vector();
				for (int i = 0; i < mainItems.length; i++) {
					ItemLocation item = mainItems[i];
					String locationName = item.getLocation().getScreenName().toLowerCase();
					int pos = StringUtils.firstPos(createMatchString(locationName), searchPattern);
					if (pos >= 0) {
						if (item.getIntelligentKeyMatchPos() >= 0) {
							// die neue Position darf nur gesetzt werden, falls es auf früher schon eine gab 
							item.setIntelligentKeyMatch((short)pos);
						}
						v.addElement(item);
					}
				}
				mainItems = new ItemLocation[v.size()];
				v.copyInto(mainItems);
			} else {
				mainItems = new ItemLocation[0];
			}
		}
		
		// Stage 2 - Orte hinzufügen bei Haltestellen die gematcht haben, bzw. Hinweis anzeigen falls nichts gefunden wurde
		
		if (mainItems.length != 0) {
			if (keyState != KEY_STATE_BOTH) {
				if (getTyp() == TYPE_NACH_ORT || getTyp() == TYPE_VON_ORT) {
					// Orte hinzufügen / alle Haltestellen hinzufügen

					// FIXME State-Machine funktoniert nicht richtig - ^^^ alle Haltestellen anzeigen wird manchmal nicht angezeigt
					AbstractLocation[] alleLocations = Data.getAlleLocations();
					boolean letzerOrtSichtbar = false;
					boolean letzerOrtMatched = false;
					boolean zeigeAlleHaltestellenSichtbar = false;
					boolean HaltestelleUebersprungen = false;
					boolean eineHaltestelleSichtbar = false;
					int letzteOrtsPosition = 0;

					int iMatched = 0;

					Vector finalItems = new Vector();

					for (int iCurrent = 0; iCurrent < alleLocations.length; iCurrent++) {
						AbstractLocation currentLocation = alleLocations[iCurrent];

						boolean skipLocation;
						ItemLocation matchedLocationItem;

						if (currentLocation == mainItems[iMatched].getLocation()) {
							matchedLocationItem = mainItems[iMatched];
							iMatched++;
							if (iMatched == mainItems.length) {
								iMatched = mainItems.length - 1;
							}
							skipLocation = false;
						} else {
							skipLocation = true;
							matchedLocationItem = null;
						}

						if (currentLocation instanceof Ort) {
							letzteOrtsPosition = finalItems.size();
							if (skipLocation) {
								letzerOrtSichtbar = false;
								letzerOrtMatched = false;
							} else {
								letzerOrtSichtbar = true;
								letzerOrtMatched = true;
							}
							zeigeAlleHaltestellenSichtbar = false;
							HaltestelleUebersprungen = false;
							eineHaltestelleSichtbar = false;
						}

						if (currentLocation instanceof Haltestelle) {
							if (skipLocation) {
								HaltestelleUebersprungen = true;

								if ((letzerOrtMatched == true) && (letzerOrtSichtbar == true) && (zeigeAlleHaltestellenSichtbar == false) && (HaltestelleUebersprungen == true) && (eineHaltestelleSichtbar)) {
									if (((Haltestelle) currentLocation).getOrt().getAlleHaltestellen().size() > 1) {
										finalItems.insertElementAt(new ItemLocationShowAll(((Haltestelle) currentLocation).getOrt()), letzteOrtsPosition + 1);
										zeigeAlleHaltestellenSichtbar = true;
									}
								}
							} else {
								eineHaltestelleSichtbar = true;
								if (letzerOrtSichtbar == false) {
									letzerOrtSichtbar = true;
									finalItems.addElement(new ItemLocation(((Haltestelle) currentLocation).getOrt(), IntelligentKey.PATTERN_NOT_FOUND, intelligentKey));
								}
								if ((letzerOrtMatched == true) && (letzerOrtSichtbar == true) && (zeigeAlleHaltestellenSichtbar == false) && (HaltestelleUebersprungen == true)) {
									if (((Haltestelle) currentLocation).getOrt().getAlleHaltestellen().size() > 1) {
										finalItems.insertElementAt(new ItemLocationShowAll(((Haltestelle) currentLocation).getOrt()), letzteOrtsPosition + 1);
										zeigeAlleHaltestellenSichtbar = true;
									}
								}
							}
						}
						if (!skipLocation) {
							finalItems.addElement(matchedLocationItem);
						}
					}
					mainItems = new ItemLocation[finalItems.size()];
					finalItems.copyInto(mainItems);
				}
			}
		} else {
			String[] noEntryText = { "never visible" };
			if (keyState == KEY_STATE_NUMERIC) {
				noEntryText = new String[] { "Keine Übereinstimmung gefunden! Bitte den letzten Buchstaben löschen.", "Texteingabe funktioniert ähnlich zu IntelligentKey(*). Um Beispielsweise 'Flughafen' einzugeben bitte folgende Tasten drücken (nachdem alle bisher eingegebenen Zeichen gelöscht wurden):", "1x '3' -> F", "1x '5' -> l", "1x '8' -> u", "1x '4' -> g", "...", "(*) IntelligentKey ist ein eingetragenes Warenzeichen" };
			}

			if (keyState == KEY_STATE_ALPHA) {
				noEntryText = new String[] { "Keine Übereinstimmung gefunden! Bitte den letzten Buchstaben löschen." };
			}

			Font f = FontMetrics.getFont(Manager.ITEMS_FONT_SIZE);
			noEntryText = GfxUtils.splitupInLines(f, noEntryText, Manager.SCREEN_WIDTH - 2 * Manager.MAIN_LEFT_SPACE);

			mainItems = new ItemLocation[noEntryText.length];

			for (int i = 0; i < noEntryText.length; i++) {
				mainItems[i] = new ItemLocationDescription(noEntryText[i]);
			}
		}

		// Stage 3 - alles zusammensetzen 

		int mainItemLength = mainItems.length;

		int totalLength = mainItemLength + additionalItemLength;

		if (visibleItems < totalLength) {
			setOverflowType(OVERFLOW_TYPE_ENDLESS);
			totalLength += 1;
		} else {
			setOverflowType(OVERFLOW_TYPE_ROUND);
			firstId = 0;
		}

		Item allItems[] = new Item[totalLength];
		System.arraycopy(additionalItems, 0, allItems, 0, additionalItemLength);
		System.arraycopy(mainItems, 0, allItems, additionalItemLength, mainItemLength);

		if (getOverflowType() == OVERFLOW_TYPE_ENDLESS) {
			allItems[allItems.length - 1] = breakItem;
		}

		if (doResetSelectionAndFirstPos) {
			newItems(allItems);
		} else {
			setItems(allItems);
		}

		dataValid();
	}

	synchronized protected void drawHeader(Graphics g) {
		guiElementVonNach.drawTo(g, 0, 0, firstAfterNewItems);
	}

	private int getKeyboardY() {
		return Manager.SCREEN_HEIGHT - Manager.FOOTER_HEIGHT - keyHeight + Manager.FOOTER_UPPER_SPACE;
	}

	synchronized protected void drawSoftwareKeyboard(Graphics g) {
		int ypos = getKeyboardY();
		Image i;
		if (lastKeyVisible) {
			i = Imager.getKeyPressedImage(lastKey);
		} else {
			i = Imager.getKeyImage();
		}
		g.drawImage(i, 0, ypos, Graphics.LEFT | Graphics.TOP);
		
		if (isVisible()) {
			getGameCanvas().repaint(0, ypos, i.getWidth(), i.getHeight());
		}
	}

	synchronized protected void resetSelectionAndFirstPos() {
		super.resetSelectionAndFirstPos();
		// firstId und selectedId passend setzen

		if (selectedId != GuiSelection.SELECTION_NONE) {
			// FIXME: die Addional items - ob sie sichbar sind oder nicht muss sauberer gelöst werden
			if ((this instanceof GuiSelectionLocationHaltestellen) || intelligentKey.isInUse()) {
				selectedId = 0;
				if (getOverflowType() == OVERFLOW_TYPE_ENDLESS) {
					firstId = getValidId(-1);
				} else {
					firstId = 0;
				}
				// selektiere das erste Item, das von IntelligentKey gematched wurde
				if (intelligentKey.hasValidEntries()) {
					for (int i = 0; i < mainItems.length; i++) {
						if (mainItems[i].getIntelligentKeyMatchPos() >= 0) {
							selectedId = i;
							break;
						}
					}
				} else {
					selectedId = GuiSelectionLocation.SELECTION_NONE;
				}
			} else {
				// additional Items visible;
				if (!intelligentKey.isInUse()) {
					Item[] items = getItems();
					for (int i = 0; i < items.length; i++) {
						if (items[i].isSelectable()) {
							selectedId = i;
							break;
						}
					}
				} else {
					// selektiere das erste Item, das von IntelligentKey gematched wurde
					if (intelligentKey.hasValidEntries()) {
						for (int i = 0; i < mainItems.length; i++) {
							if (mainItems[i].getIntelligentKeyMatchPos() >= 0) {
								selectedId = i + additionalItems.length;
								break;
							}
						}
					} else {
						selectedId = GuiSelectionLocation.SELECTION_NONE;
					}
				}
				if (getOverflowType() == OVERFLOW_TYPE_ENDLESS) {
					firstId = getValidId(selectedId - 1);
				} else {
					firstId = 0;
				}
			}
		}

		onChangeSelected();
	}

	protected void resetStoredData() {
		AbstractLocation von = null;
		AbstractLocation nach = null;
		keyState = KEY_STATE_BOTH;

		switch (guiType) {
		case TYPE_VON_ORT:
			break;
		case TYPE_VON_HALTESTELLE:
			von = getAusgewaehlteLocation(TYPE_VON_ORT);
			break;
		case TYPE_NACH_ORT:
			von = getAusgewaehlteLocation(TYPE_VON_HALTESTELLE);
			break;
		case TYPE_NACH_HALTESTELLE:
			von = getAusgewaehlteLocation(TYPE_VON_HALTESTELLE);
			nach = getAusgewaehlteLocation(TYPE_NACH_HALTESTELLE);
			break;
		}

		guiElementVonNach.setLocations(von, nach);

		// FIXME: sollte in die Prepare Funktion
		if (Device.showSoftwareKeyboard()) {
			if (keyHeight == 0) {
				keyHeight = Imager.getKeyImage().getHeight();
			}
		} else {
			keyHeight = 0;
		}

		intelligentKey.clear();
		storedFirstPos.removeAllElements();
		storedSelectedPos.removeAllElements();
		storedOverFlowType.removeAllElements();
		storedVisibleItems.removeAllElements();
		storedAlpha.removeAllElements();
	}

	/**
	 * Aufruf der Methode falls sich das selektierte Item ändert
	 */
	synchronized protected void onChangeSelected() {
		AbstractLocation newSelectedLocation = getSelected();
		setAusgewaehlteLocation(getGuiType(), newSelectedLocation);
		// FIXME: wenn mit dem Pointer eine Haltestelle ausgewählt wird macht er viele unnötige repaints....

		String alphaKey = null;
		if (keyState == KEY_STATE_ALPHA) {
			alphaKey = (String) storedAlpha.lastElement();
		}

		guiElementVonNach.newSelectedItem(getGameCanvasGraphics(), getSelectedItem(), intelligentKey, true, firstAfterNewItems, alphaKey);
	}

	/**
	 * liefert die Pixelposition dirket unterhalt des selektiertem Elements
	 * zurück
	 */
	public int getLineBeneathSelected() {
		int mainHeight = getMainHeight();

		int currentId = firstId;

		int y_add = 0;

		Item[] items = getItems();

		if ((items != null) && (items.length != 0)) {
			while (y_add + items[currentId].getHeight() <= mainHeight) {
				boolean selected;
				if (currentId == selectedId) {
					selected = true;
				} else {
					selected = false;
				}
				y_add += items[currentId].getHeight();
				if (selected) {
					if (items[currentId] instanceof ItemLocationShowAll) {
						y_add -= items[currentId - 1].getHeight();
					}
					return y_add + getMainY();
				}
				currentId = getValidId(currentId + 1);
			}
		}
		return -1;
	}

	public int getItemHeight() {
		if (initItems.length > 0) {
			return initItems[0].getHeight();
		}

		System.err.println("GuiSelectionLocation - ItemHeight - can't be here!");

		return -1;
	}

	public int getMainY() {
		switch (guiType) {
		case TYPE_NACH_ORT:
		case TYPE_VON_ORT:
			return super.getMainY();
		case TYPE_NACH_HALTESTELLE:
		case TYPE_VON_HALTESTELLE:
			return super.getMainY() + getItemHeight() + 1;
		}
		return -1;
	}

	public int getTyp() {
		return guiType;
	}

	/**
	 * zeichnet den Cursor neu
	 */
	public synchronized void updateCursor() {
		if (isVisible()) {
			guiElementVonNach.updateCursor(cursor, getGameCanvasGraphics());
			if (hideLastKey()) {
				drawSoftwareKeyboard(getGameCanvasGraphics());
			}
		}
	}

	private boolean hideLastKey() {
		if (!lastKeyVisible) {
			return false;
		}
		if (lastKey >= 0 && (System.currentTimeMillis() - lastKeyTime) > KEY_TIME) {
			lastKeyVisible = false;
			return true;
		}
		return false;
	}

	protected void newItems(Item items[]) {
		firstAfterNewItems = true;
		super.newItems(items);
	}

	synchronized protected void onShow() {
		if (Device.getGfxQuality() != Device.GFX_QUALITY_LOW) {
			cursor.activate();
		}
	}

	synchronized protected void onHide() {
		cursor.deactivate();
	}

	protected int getMainHeight() {
		return super.getMainHeight() - keyHeight;
	}

	protected void drawAdditionalStuff(Graphics g) {
		if (Device.showSoftwareKeyboard()) {
			drawSoftwareKeyboard(g);
		}
	}

	public boolean pointerPressed(int x, int y) {
		if (Device.showSoftwareKeyboard()) {
			//			System.out.println("pointerPressed: " + x + "," + y);
			int yKey = y - getKeyboardY();
			Image keyImage = Imager.getKeyImage();
			//			System.out.println("keyImage.getHeight(): " + keyImage.getHeight());
			int keyHeight = keyImage.getHeight();
			if ((yKey >= 0) && (yKey < keyHeight)) {
				// pos gibt die 'id' der Taste wieder, ab 0-7 sind die Buchstaben, 8 Whitespace und ähnliches, 9 Backspace
				// ab 5 beginnt die zweite Zeile
				int pos = 0;
				if (softKeysZweizeilig) {
					int oneWidthTimes10 = keyImage.getWidth() * 10 / 5;
					pos = (x * 10) / oneWidthTimes10;
					if (yKey >= keyHeight / 2) {
						pos += 5;
					}
				} else {
					int oneWidthTimes10 = keyImage.getWidth();
					pos = (x * 10) / oneWidthTimes10;
				}
				
				setLastKey(pos);
				//					System.out.println("pressed Soft-Key Field" + pos);

				if (pos < 8) {
					keyPressed(48 + 2 + pos, this.getGameCanvas());
				}
				if (pos == 8) {
					keyPressed(48, this.getGameCanvas());
				}
				if (pos == 9) {
					keyPressed(Keys.getKeyCode(Keys.C), this.getGameCanvas());
				}
				return false;
			}
		}
		return super.pointerPressed(x, y);
	}

	protected static void clearAusgewaehlteLocations() {
		for (int i = 0; i < ausgewaehlteLocations.length; i++) {
			ausgewaehlteLocations[i] = null;
		}
	}

	protected static AbstractLocation getAusgewaehlteLocation(int typ) {
		return ausgewaehlteLocations[typ];
	}

	protected static void setAusgewaehlteLocation(int typ, AbstractLocation location) {
		ausgewaehlteLocations[typ] = location;
	}

	public GameCanvas getGameCanvas() {
		return super.getGameCanvas();
	}

	public Cursor getCursor() {
		return cursor;
	}

	public int getFooterHeight() {
		if (Device.showSoftwareKeyboard()) {
			return Manager.SCREEN_HEIGHT - getKeyboardY() - Imager.getKeyImage().getHeight();
		} else {
			return super.getFooterHeight();
		}
	}
	
	private synchronized void setLastKey(int pos) {
		lastKeyTime = System.currentTimeMillis();
		lastKey = pos;
		lastKeyVisible = true;
		drawSoftwareKeyboard(getGameCanvasGraphics());
	}

}
