// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package gui.guis;

import gui.Color;
import gui.FontMetrics;
import gui.IOverlayable;
import gui.Manager;
import gui.items.Item;
import gui.items.ItemReminderTime;
import gui.utils.GfxUtils;

import javax.microedition.lcdui.Font;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.game.GameCanvas;

import model.Route;
import control.Global;
import control.reminder.Reminder;
import dev.Logger;

public class GuiSelectionReminderTime extends GuiSelection implements IOverlayable {
	private final int headerHeight;
	private final int innerSpace;
	private final int textHeight;
	private final String headerString;

	private final int info_height;
	private final int info_width;
	private final int info_y;
	private final int info_x;

	public GuiSelectionReminderTime(GameCanvas gc, Graphics g, Route route) {
		super("Zeitauswahl Reminder", gc, g);
		setOverflowType(OVERFLOW_TYPE_ROUND);

		Font f = FontMetrics.getFont(Manager.ITEMS_FONT_SIZE);

		int[] possibleTimes = Reminder.getPossibleReminderValues(route);

		int noItems = possibleTimes.length + 1;

		if (Global.REMINDER_ALWAYS_VISIBLE) {
			noItems = Global.REMINDER_LAUFZEIT.length + 1;
		}

		textHeight = f.getHeight();
		innerSpace = GfxUtils.scale(textHeight, 20);
		headerHeight = textHeight + 1;

		ItemReminderTime items[] = new ItemReminderTime[noItems];

		int info_x_space = GfxUtils.scale(Manager.SCREEN_WIDTH, 10);

		info_x = info_x_space;
		info_width = Manager.SCREEN_WIDTH - 2 * info_x_space;

		String[] headerStrings = { "Erinnerung anzeigen", "Erinnerung" };
		int no = GfxUtils.selectStringByMaxWidth(headerStrings, f, info_width, GfxUtils.WITH_SPACE);
		if (no == GfxUtils.NO_FIT) {
			Logger.error(this.getClass(), "Headerbreite zu klein um passenden Text hineinschreiben zu können");
			no = headerStrings.length - 1;
		}

		headerString = headerStrings[no];

		items[0] = new ItemReminderTime(-1, info_x + 1, info_width - 2);
		for (int i = 1; i < noItems; i++) {
			items[i] = new ItemReminderTime(Global.REMINDER_LAUFZEIT[i - 1], info_x + 1, info_width - 2);
		}

		newItems(items);

		if (noItems > 0) {
			info_height = headerHeight + items.length * items[0].getHeight() + 2 * innerSpace + 1;
		} else {
			info_height = headerHeight + 2 * innerSpace + 1;
		}

//		info_y = super.getHeaderHeight() * 5 / 4;
		info_y = (Manager.SCREEN_HEIGHT - info_height) / 2;

		if (noItems > 1) {
			moveDown(true);
		}
	}

	public int getInfoHeight() {
		return info_height;
	}

	public int getInfoWidth() {
		return info_width;
	}

	public int getInfoX() {
		return info_x;
	}

	public int getInfoY() {
		return info_y;
	}

	private final static int header_Background_Color = Color.RED;
	private final static int header_Font_Color = Color.WHITE;
	private final static int main_Background_Color = Color.WHITE;

	protected void drawHeader(Graphics g) {
		int info_add_y = 1;

		// Header
		Font fb = FontMetrics.getFont(Font.STYLE_BOLD, Manager.ITEMS_FONT_SIZE);
		g.setFont(fb);
		g.setColor(header_Background_Color);
		g.fillRect(info_x + 1, info_y + 1, info_width - 2, headerHeight - 1);
		g.setColor(header_Font_Color);
		GfxUtils.drawHCenteredText(g, headerString, info_y + info_add_y);

		// Rect
		g.setColor(Color.BLACK);
		g.drawRect(info_x, info_y, info_width - 1, info_height - 1);
	}

	synchronized protected void drawMain(Graphics g, boolean clearBackgroundAndRepaint) {
		int y_base = getMainY();
		int mainHeight = getMainHeight();
		int currentId = firstId;
		visibleItems = 0;
		int y_add = innerSpace;

		g.setColor(main_Background_Color);
		g.fillRect(info_x + 1, y_base, info_width - 2, mainHeight - 1);

		Item items[] = getItems();

		if ((items != null) && (items.length != 0)) {
			/* es werden so lange Elemente neu gezeichnet bis das nächste Element nicht mehr komplett hineinpassen würde
			 */
			while (y_add + items[currentId].getHeight() <= mainHeight - innerSpace) {
				boolean selected;
				if (currentId == selectedId) {
					selected = true;
				} else {
					selected = false;
				}
				y_add += items[currentId].draw(this, g, Manager.MAIN_LEFT_SPACE, y_base + y_add, selected, currentId);
				currentId = getValidId(currentId + 1);
				visibleItems++;
				if ((overflowType != OVERFLOW_TYPE_ENDLESS) && (visibleItems == items.length)) {
					break;
				}
			}
		}

		if (clearBackgroundAndRepaint && isVisible()) {
			gc.repaint(info_x, y_base, info_width, mainHeight + Manager.MAIN_INNER_SPACE);
		}
	}

	protected void drawFooter(Graphics g) {
		// no footer
	}

	protected int getHeaderHeight() {
		return headerHeight;
	}

	protected int getMainY() {
		return info_y + getHeaderHeight();
	}

	protected int getMainHeight() {
		return info_height - getHeaderHeight();
	}

	public int getSelectedId() {
		return selectedId;
	}

	protected void clearMainBackground(Graphics g) {
		// empty
	}

	public int getTime() {
		if (selectedId == 0) {
			return -1;
		} else {
			return Global.REMINDER_LAUFZEIT[selectedId - 1];
		}
	}
}
