// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package gui.options;

import gui.launch.GameCanvasInstance;
import gui.utils.AlertMessage;

import java.util.Timer;
import java.util.TimerTask;

import javax.bluetooth.RemoteDevice;
import javax.microedition.lcdui.ChoiceGroup;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Form;
import javax.microedition.lcdui.StringItem;

import control.Pushandride;
import control.connections.bluetooth.BluetoothManager;
import control.connections.bluetooth.BluetoothSearchDevices;
import control.connections.bluetooth.BluetoothSendFileViaObex;
import control.connections.bluetooth.IBluetoothSearchDevicesCallback;
import control.connections.bluetooth.IBluetoothSendFileViaObexCallback;
import control.connections.bluetooth.IBluetoothStateCallback;
import control.locale.Locale;
import dev.DebugInfo;
import dev.Logger;

public class BluetoothChooseDeviceScreen extends Form implements CommandListener, IBluetoothSearchDevicesCallback, IBluetoothSendFileViaObexCallback, IBluetoothStateCallback {
	protected static final Command sendJarCommand = new Command("Senden", Command.OK, 1);
	protected final Command abortCommand = new Command("Abbrechen", Command.CANCEL, 3);
	protected final Command backCommand = new Command("Zurück", Command.BACK, 3);
	protected final Command sendJadJaCommand = new Command("Ja", Command.OK, 2);
	protected final Command sendJadNeinCommand = new Command("Nein", Command.EXIT, 1);

	private ChoiceGroup cg;
	private RemoteDevice[] availableDevices;

	private final Display display;
	private final Displayable last;
	private boolean showState = false;
	private String currentState = "";
	private StringItem itemState;

	private StringItem itemProgress;
	private final int UPDATEINTERVAL = 750;

	private Timer updateTimer;
	private UpdateTimerTask updateTimerTask;
	private final BluetoothManager bluetoothManager;

	private final static String moreInfo = "Weitere Informationen zu Bluetooth findest Du in der FAQ der Homepage (pushandride.de).";

	public BluetoothChooseDeviceScreen() {
		super("Geräte-Auswahl");

		this.setCommandListener(this);
		this.addCommand(abortCommand);

		display = Display.getDisplay(Pushandride.getMidlet());
		last = display.getCurrent();
		display.setCurrent(this);

		bluetoothManager = new BluetoothManager(this, this, this);

		startSearchForDevices();
	}

	synchronized public void commandAction(Command c, Displayable d) {
		if (c == sendJarCommand) {
			this.clearItems();
			this.removeCommand(sendJarCommand);
			this.removeCommand(backCommand);
			this.addCommand(abortCommand);

			RemoteDevice remoteDevice = availableDevices[cg.getSelectedIndex()];
			String remoteName = cg.getString(cg.getSelectedIndex());

			this.append("Sende pushandride.jar (" + (BluetoothManager.getSizeOfJar(bluetoothManager) / 1024) + "kb) an " + remoteName + "\n");

			bluetoothManager.sendJar(remoteDevice);
			startUpdateTask();
		} else if (c == abortCommand) {
			if (bluetoothManager != null) {
				bluetoothManager.cancelAction();
			}
			back();
		} else if (c == sendJadNeinCommand) {
			back();
		} else if (c == sendJadJaCommand) {
			addCommand(abortCommand);
			removeCommand(sendJadJaCommand);
			removeCommand(sendJadNeinCommand);

			this.clearItems();

			RemoteDevice remoteDevice = availableDevices[cg.getSelectedIndex()];
			String remoteName = cg.getString(cg.getSelectedIndex());

			this.append("Sende pushandride.jad  an " + remoteName + "\n");

			bluetoothManager.sendJad(remoteDevice);
		} else if (c == backCommand) {
			back();
		}
	}

	synchronized protected void back() {
		stopUpdateTask();
		display.setCurrent(last);
		GameCanvasInstance.getInstance().setFullScreenMode(true);
	}

	synchronized private void startSearchForDevices() {
		bluetoothManager.searchForDevices();

		this.append(new StringItem(null, Locale.BluetoothSucheNachDevices));
		startUpdateTask();
	}

	synchronized private void startUpdateTask() {
		if (updateTimer != null) {
			updateTimer.cancel();
		}

		int state = bluetoothManager.getState();
		StringItem newItemProgress = null;
		switch (state) {
		case BluetoothManager.STATE_SEARCH_FOR_DEVICES:
			newItemProgress = new StringItem("", "...");
			break;
		case BluetoothManager.STATE_SEND_JAR:
			newItemProgress = new StringItem("", "");
			break;
		}
		if (newItemProgress == null) {
			return;
		}

		itemProgress = newItemProgress;
		this.append(itemProgress);

		updateTimer = new Timer();
		updateTimerTask = new UpdateTimerTask(this);

		updateTimer.scheduleAtFixedRate(updateTimerTask, UPDATEINTERVAL, UPDATEINTERVAL);
	}

	synchronized public void finishedDeviceSearch(RemoteDevice[] devices, int inqueryState) {
		stopUpdateTask();
		this.availableDevices = devices;
		removeCommand(abortCommand);
		addCommand(backCommand);

		if (inqueryState == BluetoothSearchDevices.INQUIRY_COMPLETED) {
			if (devices.length == 0) {
				this.clearItems();
				this.append("\nEs wurden keine erreichbaren Geräte gefunden.\nBeim Empfangsgerät muss Bluetooth eingeschalten, sowie die Sichbarkeit des Gerätes aktiviert sein.\n" + moreInfo);
			} else {
				this.addCommand(sendJarCommand);
				clearItems();
				if (devices.length == 1) {
					this.append(new StringItem("", "1 Gerät gefunden:"));
				} else {
					this.append(new StringItem("", devices.length + " Geräte gefunden:"));
				}

				cg = new ChoiceGroup("Bitte auswählen", ChoiceGroup.EXCLUSIVE);

				for (int i = 0; i < devices.length; i++) {
					RemoteDevice rd = devices[i];
					String bluetoothName = rd.getBluetoothAddress();
					cg.append(bluetoothName, null);
				}

				this.append(cg);

				//				this.delete(iBitteWarten);
			}
		} else {
			if (inqueryState != BluetoothSearchDevices.INQUIRY_TERMINATED) {
				failure(getClass(), null, "INQUIRY_STATE: " + inqueryState, BluetoothManager.FAILURE_BLUETOOTH_EXCEPTION);
			}
		}
	}

	synchronized private void stopUpdateTask() {
		if (updateTimer != null) {
			updateTimer.cancel();
		}
	}

	synchronized public void failure(Class c, Exception e, String message, int errorType) {
		stopUpdateTask();

		String titel = "Kritischer Fehler";
		String fehlermeldung = "Nummer: 404";
		switch (errorType) {
		case BluetoothManager.FAILURE_BLUETOOTH_EXCEPTION:
		case BluetoothManager.FAILURE_EXCEPTION:
			titel = "Fehler";
			fehlermeldung = "Fehler beim Senden per Bluetooth";
			break;
		case BluetoothManager.FAILURE_BLUETOOTH_NO_OBEX:
			titel = "Fehler";
			fehlermeldung = "Das Gerät, an welches gesendet werden soll, unterstützt nicht den 'OBEX push service'.";
			break;
		case BluetoothManager.FAILURE_BLUETOOTH_NOT_WORKING:
			titel = "Hinweis";
			fehlermeldung = Locale.BluetoothNichtAktiv;
			break;
		case BluetoothManager.FAILURE_TIMEOUT:
			titel = "Timeout";
			fehlermeldung = "Bluetooth hat seit über " + (BluetoothManager.TIMEOUT / 1000) + " Sekunden nicht mehr reagiert und wird daher beendet.";
			break;
		}
		if (e == null) {
			Logger.error(c, e + "\n" + message + "\n" + errorType);
		} else {
			Logger.errorException(c, e, message + "\n" + errorType);
		}

		new AlertMessage(titel, fehlermeldung, "OK").setVisible(true);
	}

	synchronized public void finishedBluetoothSendFileViaObex(boolean success, int type) {
		if (type == BluetoothSendFileViaObex.SENDJAR) {
			if (success) {
				removeCommand(abortCommand);
				clearItems();
				addCommand(sendJadJaCommand);
				addCommand(sendJadNeinCommand);
				this.append("zusätzlich die JAD-Datei senden?\n(wird nur sehr vereinzelt benötigt)\n" + moreInfo);
				return;
			}
		}
		back();
	}

	synchronized public void foundDevice(RemoteDevice device) {
		/*		String bluetoothName;
				try {
					bluetoothName = device.getFriendlyName(true);
				} catch (IOException e) {
					bluetoothName = device.getBluetoothAddress();
				}
				this.append(new StringItem("", bluetoothName));*/
	}

	synchronized public void setState(String state) {
		if (state == "") {
			return;
		}
		if (showState) {
			this.currentState = state;
			if (itemState == null) {
				itemState = new StringItem("", "\n" + this.currentState + "\n\n");
				this.insert(0, itemState);
			} else {
				itemState.setText("\n" + this.currentState + "\n");
			}
			DebugInfo.addDebugString("BTS: " + state);
			try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
			}
		}
	}

	synchronized private void clearItems() {
		this.deleteAll();
		if (showState && (itemState != null)) {
			this.append(itemState);
		}
	}

	private int cntDots = 0;
	private final String[] dotString = { ".", "..", "..." };

	synchronized protected void updateScreen() {
		// update search
		String nextText = null;
		int state = bluetoothManager.getState();

		switch (state) {
		case BluetoothManager.STATE_SEARCH_FOR_DEVICES:
			cntDots = (cntDots + 1) % 3;
			nextText = dotString[cntDots];
			break;
		case BluetoothManager.STATE_SEND_JAR:
			int sendState = bluetoothManager.getSendState();
			switch (sendState) {
			case BluetoothManager.SEND_STATE_SEARCH_OBEX:
			case BluetoothManager.SEND_STATE_CONNECTING:
				cntDots = (cntDots + 1) % 3;
				nextText = "verbinde" + dotString[cntDots];
				break;
			case BluetoothManager.SEND_STATE_SENDING:
				nextText = (bluetoothManager.getSendPos() * 100 / (BluetoothSendFileViaObex.getSizeOfJar(bluetoothManager))) + "%";
				break;
			}
			break;
		}

		if (nextText != null) {
			itemProgress.setText(nextText);
		}
	}

	synchronized public void friendyName(RemoteDevice device, String name) {
		String bluetoothName = device.getBluetoothAddress();
		for (int i = 0; i < cg.size(); i++) {
			if (cg.getString(i).equals(bluetoothName)) {
				cg.set(i, name, null);
			}
		}
	}
}

class UpdateTimerTask extends TimerTask {
	private final BluetoothChooseDeviceScreen bluetoothScreen;

	public UpdateTimerTask(BluetoothChooseDeviceScreen bluetoothScreen) {
		this.bluetoothScreen = bluetoothScreen;
	}

	synchronized public final void run() {
		bluetoothScreen.updateScreen();
	}
}
