// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package gui.utils;

import gui.Color;
import gui.FontMetrics;
import gui.Manager;
import gui.Point;

import java.io.IOException;
import java.util.Vector;

import javax.microedition.lcdui.Font;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;

import control.Device;
import dev.Logger;

public class GfxUtils {
	private static int clip_x;
	private static int clip_y;
	private static int clip_height;
	private static int clip_width;
	private static boolean FONT_Y_CORRECTION = false;
	
	/**
	 * zeichnet eine horizontale Linie über den ganzen Bildschirm
	 */
	public static void drawFullLineHoriz(Graphics g, int ypos) {
		g.drawLine(g.getClipX(), ypos, g.getClipX() + g.getClipWidth() - 1, ypos);
	}

	/**
	 * zeichnet eine horizontale Linie ab einer gegebenen X-Position
	 * 
	 * @param x -
	 *            ab wann die Linie beginnen soll
	 */
	public static void drawFullLineHorizFromX(Graphics g, int x, int ypos) {
		g.drawLine(x, ypos, g.getClipX() + g.getClipWidth() - 1, ypos);
	}

	/**
	 * zeichnet ein ausgefülltes Recheck über die ganze Bildschrimbreite
	 */
	public static void fillFullRectangle(Graphics g, int y, int height) {
		g.fillRect(g.getClipX(), y, g.getClipWidth(), height);
	}

	/**
	 * zeichnet ein ausgefülltes Recheck ab einer bestimmten X-Position über die ganze
	 * Bildschrimbreite
	 */
	public static void fillFullRectangleFromX(Graphics g, int x, int y, int height) {
		g.fillRect(x, y, g.getClipX() + g.getClipWidth() - x, height);
	}

	/**
	 * Zeichnet ein symmetisches Dreick, welches nach unten zeigt WICHTIG: Width, Height sollten
	 * ungerade sein
	 */
	public static void downTriangle(Graphics g, int center_x, int center_y, int width, int height) {
		int base_y = center_y - (height - 1) / 2;
		for (int y = 0; y < height; y++) {
			double steigung = ((width - 1) / 2.0) / (double) (height - 1);
			int lineWidthDelta = (int) ((width - 1) / 2.0 - y * steigung + 0.499);

			g.drawLine(center_x - lineWidthDelta, base_y + y, center_x + lineWidthDelta, base_y + y);
		}
	}

	/**
	 * Speichert die aktuelle clipping_position um sie später mit restoreClip() wiederherzustellen.
	 * Zudem wird das aktuelle Clip-Rect mit dem übergebenen verrechnet
	 */
	public static void doClip(Graphics g, int x, int y, int width, int height) {
		clip_x = g.getClipX();
		clip_y = g.getClipY();
		clip_width = g.getClipWidth();
		clip_height = g.getClipHeight();

		g.clipRect(x, y, width, height);
	}

	/**
	 * stellt das letzte Clip-Rect wieder her, das beim aufruf von doClip verwendet wurde
	 * 
	 * @param g
	 */
	public static void restoreClip(Graphics g) {
		g.setClip(clip_x, clip_y, clip_width, clip_height);
	}

	/**
	 * schaltet das Clipping aus
	 */
	public static void clearClip(Graphics g) {
		g.setClip(0, 0, Manager.SCREEN_WIDTH, Manager.SCREEN_HEIGHT);
	}

	/**
	 * skaliert einen Wert mittels einer übergebenen Prozentzahl
	 * 
	 * @param value -
	 *            Ausganswert
	 * @param percent -
	 *            Prozentsatz
	 * @return value * percent / 100
	 */
	public static int scale(int value, int percent) {
		return value * percent / 100;
	}

	/**
	 * skaliert einen Wert mittels einer übergebenen Prozentzahl, es kann ein Mindestwert angegeben
	 * werden
	 * 
	 * @param value -
	 *            Ausganswert
	 * @param percent -
	 *            Prozentsatz
	 * @param minimum -
	 *            Mindestwert
	 * @return max(value * percent / 100, minimum)
	 */
	public static int scaleWithMinimun(int value, int percent, int minimum) {
		return Math.max(minimum, scale(value, percent));
	}

	public static String trim(String s) {

		while (s.length() > 0) {
			while (s.charAt(0) == ' ') {
				s = s.substring(1);
			}
		}

		while (s.length() > 0) {
			while (s.charAt(s.length() - 2) == ' ') {
				s = s.substring(0, s.length() - 2);
			}
		}

		return s;
	}

	/**
	 * gibt die Breite des übergebenen Textes zurück
	 */
	public static int getTextWidth(String text, Font f) {
		return f.stringWidth(text);
	}

	/**
	 * bestimmt die x-Position für einen einen Text, mit dem übergebenen Alignment
	 * 
	 * @param text -
	 *            Ausgabetext
	 * @param alignment -
	 *            wie bei Graphics.drawString
	 */
	public static Point getTextPos(String text, Font f, Point p, int alignment) {
		int x = p.X;
		int y = p.Y;

		int textWidth = getTextWidth(text, f);
		if ((alignment & Graphics.LEFT) != 0) {
			//
		} else {
			if ((alignment & Graphics.HCENTER) != 0) {
				x = p.X - textWidth / 2;
			} else {
				if ((alignment & Graphics.RIGHT) != 0) {
					x = p.X - textWidth;
				} else {
					System.err.println("getTextPos - wrong parameter - horizontal");
					return null;
				}
			}
		}

		int textHeight = f.getHeight();

		if ((alignment & Graphics.TOP) != 0) {
			//
		} else {
			if ((alignment & Graphics.VCENTER) != 0) {
				y = p.Y - textHeight / 2;
			} else {
				if ((alignment & Graphics.BOTTOM) != 0) {
					x = p.Y - textHeight;
				} else {
					System.err.println("getTextPos - wrong parameter - vertikal");
					return null;
				}
			}
		}

		return new Point(x, y);
	}

	public static void drawText(Graphics g, String text, int x, int y) {
		if (FONT_Y_CORRECTION) {
			y += FontMetrics.getFontYOffset(g.getFont().getSize());
		}

		g.drawString(text, x, y, Graphics.LEFT | Graphics.TOP);
	}

	public static void drawText(Graphics g, String text, int x, int y, int alignment) {
		if (FONT_Y_CORRECTION) {
			y += FontMetrics.getFontYOffset(g.getFont().getSize());
		}

		g.drawString(text, x, y, alignment);
	}

	public static void drawText(Graphics g, String text, Point p) {
		int addY = 0;
		if (FONT_Y_CORRECTION) {
			addY = FontMetrics.getFontYOffset(g.getFont().getSize());
		}

		g.drawString(text, p.X, p.Y + addY, Graphics.LEFT | Graphics.TOP);
	}

	/**
	 * schreibt einen Text an die angegebene Position (genau wie Graphics.drawString - nur
	 * funktioniert es zu 100%)
	 * 
	 * @param text -
	 *            Ausgabetext
	 * @param alignment -
	 *            wie bei Graphics.drawString
	 */
	public static void drawAlignedText(String text, Graphics g, Point p, int alignment) {
		Point pos = getTextPos(text, g.getFont(), p, alignment);
		drawText(g, text, pos.X, pos.Y);
	}

	/**
	 * schreibt einen auf dem Bilschirm zentrierten Text
	 */
	public static void drawHCenteredText(Graphics g, String text, int y) {
		int x = (Manager.SCREEN_WIDTH - getTextWidth(text, g.getFont())) / 2;
		drawText(g, text, x, y);
	}

	public static void drawText(Graphics g, String[] text, int x, int y) {
		int fontHeight = g.getFont().getHeight(); 
		int yPos = y;
		
		for (int i = 0; i < text.length; i++) {
			drawText(g, text[i], x, yPos);
			yPos += fontHeight; 
		}
	}

	/**
	 * gibt die x-Position zurück, so dass der übergebene Text zentriert erscheint
	 */
	public static int getCenterXPos(String text, Font f, int x1, int x2) {
		int textWidth = getTextWidth(text, f);
		int x_left, x_right;

		if (x2 > x1) {
			x_left = x1;
			x_right = x2;
		} else {
			x_left = x2;
			x_right = x1;
		}

		return x_left + (x_right - x_left - textWidth) / 2;
	}

	/**
	 * schreibt den Text zwischen die zwei übergebenen x-Positionen
	 */
	public static void drawTextInBetween(Graphics g, String text, int x1, int x2, int y) {
		int textWidth = getTextWidth(text, g.getFont());
		int xpos;
		int x_left;
		int x_right;

		if (x2 > x1) {
			x_left = x1;
			x_right = x2;
		} else {
			x_left = x2;
			x_right = x1;
		}

		xpos = x_left + (x_right - x_left - textWidth) / 2;

		drawText(g, text, xpos, y);
	}

	/**
	 * aktviert die Schritkorrektur, so dass die Schriften vertikal mittig zentriert dargestellt
	 * werden (falls der Hersteller sie oben bündig programmiert hat)
	 */
	public static void activateFontYOffset() {
		FONT_Y_CORRECTION = true;
	}

	public static final int NO_FIT = -1;
	public static final int WITH_SPACE = 0;
	public static final int WITHOUT_SPACE = 1;

	/**
	 * wählt einen übergebenen String über die Maximale Größe aus und liefert den passendsten zurück
	 * 
	 * @param strings -
	 *            zuerst die großen, dann die kleinen
	 * @param f
	 * @param maxWidth -
	 *            maximale Breite des Textes in Pixeln
	 * @param withInnerspace -
	 *            falls true wird nochmal f.getHeight() von der Breite abgezogen
	 * @return den Nummer des passendsten Strings (falls keiner passt Utils.NO_FIT)
	 */
	public static int selectStringByMaxWidth(String[] strings, Font f, int maxWidth, int withInnerspace) {
		if (withInnerspace == WITH_SPACE) {
			maxWidth -= f.getHeight();
		}
		for (int i = 0; i < strings.length; i++) {
			if (f.stringWidth(strings[i]) <= maxWidth) {
				return i;
			}
		}
		return NO_FIT;
	}

	public static Image loadImage(String filename) {
		try {
			return Image.createImage(filename);
		} catch (IOException e) {
			Logger.error(GfxUtils.class, "Fehler beim Laden von '" + filename + "' !" + e.getMessage());
			Logger.errorException(GfxUtils.class, e, "asjjhxy787");
		}
		// failure - return dummy
		return Image.createImage(1, 1);
	}

	public static String[] splitupInLines(Font f, String[] input, int maxWidth) {
		Vector output = new Vector();
	
		for (int i = 0; i < input.length; i++) {
			input[i] = input[i].trim();
	
			int width = f.stringWidth(input[i]);
			// überprüfe ob die aktuelle Zeile zu gross ist und splitte evtl.
			if (width <= maxWidth) {
				output.addElement(input[i]);
			} else {
				String currentItem = input[i];
				int lastPos = 0;
				int lastCopyPos = 0;
				int nextPos = 0;
				int lineWidth = 0;
				boolean finish = false;
				do {
					nextPos = currentItem.indexOf(' ', lastPos);
					if (nextPos == -1) {
						finish = true;
						nextPos = currentItem.length();
					}
	
					String nextLine;
	
					nextLine = currentItem.substring(lastCopyPos, nextPos);
	
					lineWidth = f.stringWidth(nextLine);
	
					if (lineWidth > maxWidth) {
						String substring = currentItem.substring(lastCopyPos, lastPos);
						output.addElement(substring);
						lastCopyPos = lastPos;
					} else {
					}
					if (finish) {
						String substring = currentItem.substring(lastCopyPos, nextPos);
						output.addElement(substring);
					}
	
					lastPos = nextPos + 1;
				} while (!finish);
			}
		}
	
		String[] out = new String[output.size()];
		output.copyInto(out);
		return out;
	}
	
	public static int getBackgroundColorByPos(final int pos) {
		return getBackgroundColorByPos(pos, false);
	}

	public static int getBackgroundColorByPos(final int pos, final boolean forceColor) {
		if (pos % 2 == 0) {
			return Color.WHITE;
		} else {
			if (Device.increaseHeightForTouch() || forceColor) {
				return Manager.ITEMS_COLOR_BACKGROUND_LIGHT;
			} else {
				return Color.WHITE;
			}
		}
	}

	public static int getBackgroundColorByPosWithDarkBg(int pos) {
		if (pos % 2 == 0) {
			return Manager.ITEMS_COLOR_BACKGROUND_MID;
		} else {
			if (Device.increaseHeightForTouch()) {
				return Manager.ITEMS_COLOR_BACKGROUND_DARK;
			} else {
				return Manager.ITEMS_COLOR_BACKGROUND_MID;
			}
		}
	}
	
	public static int getHeightForTouch(int height) {
		return height * 18 / 10;
	}
}
