// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package gui.utils;

import javax.microedition.lcdui.Image;

public class ImageScaler {

	/**
	 * skalert ein Bild ohne Antialiasing - sehr schnell und schlecht
	 */
	public static Image scaleBresenhamsByHeight(Image source, int newHeight) {
		if (newHeight <= 0) {
			return Image.createImage(0, 0);
		}

		int sourceWidth = source.getWidth();
		int sourceHeight = source.getHeight();

		int nYFactor = (sourceHeight << 10) / newHeight;

		int newWidth = (sourceWidth << 10) / nYFactor;

		return scaleBresenhams(source, newWidth, newHeight);
	}

	/**
	 * skalert ein Bild ohne Antialiasing - sehr schnell und schlecht
	 */
	public static Image scaleBresenhams(Image source, int newWidth, int newHeight) {
		int[] rawInput = new int[source.getHeight() * source.getWidth()];
		source.getRGB(rawInput, 0, source.getWidth(), 0, 0, source.getWidth(), source.getHeight());

		int[] rawOutput = new int[newWidth * newHeight];

		int YD = (source.getHeight() / newHeight) * source.getWidth() - source.getWidth();
		int YR = source.getHeight() % newHeight;
		int XD = source.getWidth() / newWidth;
		int XR = source.getWidth() % newWidth;
		int outOffset = 0;
		int inOffset = 0;

		for (int y = newHeight, YE = 0; y > 0; y--) {
			for (int x = newWidth, XE = 0; x > 0; x--) {
				rawOutput[outOffset++] = rawInput[inOffset];
				inOffset += XD;
				XE += XR;
				if (XE >= newWidth) {
					XE -= newWidth;
					inOffset++;
				}
			}
			inOffset += YD;
			YE += YR;
			if (YE >= newHeight) {
				YE -= newHeight;
				inOffset += source.getWidth();
			}
		}
		return Image.createRGBImage(rawOutput, newWidth, newHeight, false);
	}

	private static final int maskBlue = 0x000000FF;
	private static final int maskRed = 0x00FF0000;
	private static final int maskGreen = 0x0000FF00;

	/**
	 * skalert ein Bild bilinear (mit Antialiasing) - mittellangsam - optimierte Version
	 */
	public static Image scaleBilinearByHeight(Image source, int newHeight) {
		if (newHeight <= 0) {
			return Image.createImage(0, 0);
		}

		int sourceWidth = source.getWidth();
		int sourceHeight = source.getHeight();

		int yFactor = (sourceHeight << 10) / newHeight;

		int newWidth = (sourceWidth << 10) / yFactor;

		return scaleBilinear(source, newWidth, newHeight);
	}

	/**
	 * skalert ein Bild bilinear (mit Antialiasing) - mittellangsam - optimierte Version
	 */
	public static Image scaleBilinearByWidth(Image source, int newWidth) {
		if (newWidth <= 0) {
			return Image.createImage(0, 0);
		}

		int sourceWidth = source.getWidth();
		int sourceHeight = source.getHeight();

		int xFactor = (sourceWidth << 10) / newWidth;

		int newHeight = (sourceHeight << 10) / xFactor;

		return scaleBilinear(source, newWidth, newHeight);
	}

	/**
	 * skalert ein Bild bilinear (mit Antialiasing) wobei die maximale Höhe und Breite übergeben wird - mittellangsam - optimierte Version
	 */
	public static Image scaleBilinearByMaxWidthHeight(Image source, int maxWidth, int maxHeight) {
		if (maxWidth <= 0) {
			return Image.createImage(0, 0);
		}
		if (maxHeight <= 0) {
			return Image.createImage(0, 0);
		}

		int sourceWidth = source.getWidth();
		int sourceHeight = source.getHeight();

		int xFactor = (sourceWidth << 10) / maxWidth;
		int yFactor = (sourceHeight << 10) / maxHeight;
		int factor = Math.max(xFactor, yFactor);

		int finalWidth = (sourceWidth << 10) / factor;
		int finalHeight = (sourceHeight << 10) / factor;

		return scaleBilinear(source, finalWidth, finalHeight);
	}

	/**
	 * skalert ein Bild bilinear (mit Antialiasing) - mittellangsam - optimierte Version
	 * 
	 * Maximale Breite/Höhe: 2^21 (5097152)
	 * 
	 * Genauigkeit: 
	 * xy-Faktor: 2^10 (1/1024)
	 * Farbberechnungen: 2^10 (1/1024)
	 * 
	 */
	public static Image scaleBilinear(Image source, int newWidth, int newHeight) {
		if (newHeight <= 0) {
			return Image.createImage(0, 0);
		}

		int sourceWidth = source.getWidth();

		int sourceHeight = source.getHeight();

		// sourceWidth, sourceHeight Max: 1024 - 10 bit
		int numSourceWidth = sourceWidth << 10;

		int numSourceHeight = sourceHeight << 10;

		int numXFactor = numSourceWidth / newWidth;

		int numYFactor = numSourceHeight / newHeight;

		int fraction_x;
		int fraction_y;
		int one_minus_x;
		int one_minus_y;

		int[] rawInput = new int[sourceHeight * sourceWidth];

		source.getRGB(rawInput, 0, sourceWidth, 0, 0, sourceWidth, sourceHeight);

		int[] rawOutput = new int[newWidth * newHeight];

		int ceil_x, ceil_y, floor_x, floor_y;

		int red, green, blue;

		int b1, b2;

		int c1;
		int c2;
		int c3;
		int c4;

		for (int x = 0; x < newWidth; x++)
			for (int y = 0; y < newHeight; y++) {

				int real_x = x * numXFactor;
				int real_y = y * numYFactor;

				floor_x = real_x >> 10;
				floor_y = real_y >> 10;

				ceil_x = floor_x + 1;
				if (ceil_x >= sourceWidth)
					ceil_x = floor_x;
				ceil_y = floor_y + 1;
				if (ceil_y >= sourceHeight)
					ceil_y = floor_y;

				fraction_x = real_x - (floor_x << 10);
				fraction_y = real_y - (floor_y << 10);

				one_minus_x = (1 << 10) - fraction_x;
				one_minus_y = (1 << 10) - fraction_y;

				int offsety_floor = floor_y * sourceWidth;
				int offsety_ceil = ceil_y * sourceWidth;

				c1 = rawInput[floor_x + offsety_floor];
				c2 = rawInput[ceil_x + offsety_floor];

				c3 = rawInput[floor_x + offsety_ceil];
				c4 = rawInput[ceil_x + offsety_ceil];

				// Blue

				b1 = (c1 & maskBlue) * one_minus_x + (c2 & maskBlue) * fraction_x;

				b2 = (c3 & maskBlue) * one_minus_x + (c4 & maskBlue) * fraction_x;

				// runden ?
				blue = ((one_minus_y * b1 + fraction_y * b2) >> 20) & maskBlue;

				// Green

				b1 = ((c1 & maskGreen) * one_minus_x + (c2 & maskGreen) * fraction_x) >> 8;

				b2 = ((c3 & maskGreen) * one_minus_x + (c4 & maskGreen) * fraction_x) >> 8;

				green = ((one_minus_y * b1 + fraction_y * b2) >> 12) & maskGreen;

				// Red

				b1 = ((c1 & maskRed) >> 16) * one_minus_x + ((c2 & maskRed) >> 16) * fraction_x;

				b2 = ((c3 & maskRed) >> 16) * one_minus_x + ((c4 & maskRed) >> 16) * fraction_x;

				red = ((one_minus_y * b1 + fraction_y * b2) >> 4) & maskRed;

				rawOutput[x + y * newWidth] = red | green | blue;
			}

		return Image.createRGBImage(rawOutput, newWidth, newHeight, false);
	}
}
