// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package gui.utils;

import gui.FontMetrics;
import gui.Manager;
import gui.items.ItemLocation;
import gui.items.ItemLocationDescription;
import gui.items.ItemLocationShowAll;

import java.util.Stack;
import java.util.Vector;

import javax.microedition.lcdui.Font;

import model.AbstractLocation;

/**
 * Die IntelligentKey Klasse erlaubt es innerhalb mit der IntelligentKey-Tastatur passende Namen
 * herauszufiltern. Dabei wird darauf geachtet dass die eingegebenen Tasten auf
 * Wortanfängen passen.
 * 
 * Optimierung (alles mit condensed): IntelligentKey besteht aus 9 Tasten, somit kann eine
 * Taste mit 5 Bit abgebildet werden. Damit passen 6 Tasten (30 Bit) in ein
 * Integer (31+1Bit). steht im 'condensedValue' xx 00001 00001 00010 00010 00011
 * 00000 heisst das, dass 032211 gedrückt wurde (reversed order)
 * 
 * Jede Haltestelle hat ein int[] mit condensedValues, die den Namen
 * repräsentieren. Diese Werde mit mit eingegebenen condensedValue verarbeitet
 * und bestimmt, ob ein Match enthalten ist.
 * 
 * Jeder LocationName bekommt ein ' ' vorangestellt, die Tastatureingabe bekommt
 * ebenso eine 0 (= ' ') vorangestellt. Das bezweckt dass nur Wortanfänge
 * gesucht werden.
 * 
 * Da im condensedValue von der Eingabe nur 6 Zeichen gespeichert werden können
 * geht das Programm in den normalen Modus über wenn mehr begraucht wird.
 */
public class IntelligentKey {

	public static final short PATTERN_NOT_FOUND = -1;
	public static final short PATTERN_NOT_SEARCHED = -2;

	private AbstractLocation[] initLocations;

	private static final String[] IntelligentKeyMap = { " ", " ", "abc", "def", "ghi", "jkl", "mno", "pqrs", "tuv", "wxyz", " ", " ", " " };

	private static ItemLocation[] noEntry;

	// speichert 6x eine IntelligentKey Tasten (5 Bit)
	private int condensedValue = 0;

	// Maske für x Tasten
	private static final int[] condensedMask = { (1 << (0 * 5)) - 1, (1 << (1 * 5)) - 1, (1 << (2 * 5)) - 1, (1 << (3 * 5)) - 1, (1 << (4 * 5)) - 1, (1 << (5 * 5)) - 1, (1 << (6 * 5)) - 1 };

	// speichert die gedrückten Zahlen
	private Stack keys;

	// speichert die möglichen Haltestellen
	private Stack matchedItemsOfPos;

	public IntelligentKey(AbstractLocation[] locations) {
		this.initLocations = locations;
		clear();

		// Items die zurückgegeben werden falls kein Ort / Haltestelle mehr auf
		// das IntelligentKey matched
		if (noEntry == null) {

			String[] noEntryText = { 
					"Keine Übereinstimmung gefunden! Bitte den letzten Buchstaben löschen.", 
					"Texteingabe funktioniert ähnlich zu IntelligentKey(*). Um Beispielsweise 'Flughafen' einzugeben bitte folgende Tasten drücken (nachdem alle bisher eingegebenen Zeichen gelöscht wurden):", 
					"1x '3' -> F", 
					"1x '5' -> l", 
					"1x '8' -> u", 
					"1x '4' -> g", 
					"...",
					"(*) IntelligentKey ist ein eingetragenes Warenzeichen"};
			Font f = FontMetrics.getFont(Manager.ITEMS_FONT_SIZE);
			noEntryText = GfxUtils.splitupInLines(f, noEntryText, Manager.SCREEN_WIDTH - 2 * Manager.MAIN_LEFT_SPACE);

			noEntry = new ItemLocation[noEntryText.length];

			for (int i = 0; i < noEntryText.length; i++) {
				noEntry[i] = new ItemLocationDescription(noEntryText[i]);
			}
		}

	}

	/**
	 * setzt die IntelligentKey-Klasse auf die anfangs übergebenen Items zurück
	 */
	public void clear() {
		keys = new Stack();
		keys.push("");
		matchedItemsOfPos = new Stack();
		matchedItemsOfPos.push(this.initLocations);
	}

	/**
	 * Fuegt KeyEvent in die Liste der bisherigen hinzu
	 * 
	 * @param key
	 *            - KeyEvent
	 * @return ob es ein gültige Taste übergeben wurde
	 */
	public boolean addKey(int key) {
		int intelligentKey;
		// FIXME hier sollte Key.Star Key.Doppelkreuz rein
		if ((key == 49) || (key == 42) || (key == 35)) {
			key = 48;
		}
		if ((key >= 48) && (key <= 57)) {
			if (!hasValidEntries()) {
				return false;
			}
			intelligentKey = key - 48;
			// Wenn die Ziffer-Taste belegt ist...
			if (IntelligentKeyMap[intelligentKey].length() == 0) {
				return false;
			} else {
				// berechnet den neuen condensedWert
				int oldLength = getKeysLength();
				if (oldLength <= 4) {
					condensedValue = (condensedValue & condensedMask[oldLength + 1]) | keyToInt(key, oldLength + 1);
				}

				// der klassische IntelligentKey teil
				String newString;
				newString = ((String) keys.peek()) + (char) key;
				keys.push(newString);

				matchIntelligentKey();

				return true;
			}
		}
		return false;
	}

	/**
	 * wandelt einen übergebenen Taste key an der Position pos in einen
	 * condensedWert um
	 */
	private int keyToInt(int key, int pos) {
		return (key - '0') << (pos * 5);
	}

	/**
	 * nimmt die zuletzt gültig gespeicherte Taste zurück
	 * 
	 * @return ob eine Taste zurückgenommen wurde
	 */
	public boolean removeChar() {
		if (isInUse()) {
			matchedItemsOfPos.pop();
			keys.pop();
			return true;
		} else {
			return false;
		}
	}

	/**
	 * gibt den aktuellen IntelligentKey-Match-String zurück
	 */
	public String getString() {
		return (String) keys.peek();
	}

	/**
	 * ersetzt deutsche Sonderzeichen
	 */
	public static char replaceCharactersForIntelligentKeyStyle(char c) {
		switch (c) {
		case 'ü':
			c = 'u';
			break;
		case 'ä':
			c = 'a';
			break;
		case 'ö':
			c = 'o';
			break;
		case 'ß':
			c = 's';
			break;
		}
		// alle anderen Zeichen
		if (c < 'a' || c > 'z') {
			c = ' ';
		}
		return c;
	}

	/**
	 * generiert aus dem übergebenen String einen IntelligentKey-Sting z.B. 'Stuttgart' ->
	 * '788884278'
	 */
	private static String getIntelligentKeyString(String string) {
		String s = string.toLowerCase();

		String result = "";
		for (int i = 0; i < s.length(); i++) {
			boolean replaced;
			replaced = false;
			char c = replaceCharactersForIntelligentKeyStyle(s.charAt(i));
			for (int j = 0; j < IntelligentKeyMap.length; j++) {
				for (int k = 0; k < IntelligentKeyMap[j].length(); k++) {
					if (c == IntelligentKeyMap[j].charAt(k)) {
						// c = IntelligentKeyMap[j].charAt(0);
						c = (char) (j + 48);
						replaced = true;
						break;
					}
				}
				if (replaced)
					break;
			}
			if (!replaced) {
				// System.err.println("IntelligentKey-Error: '" + string+ " >
				// '"+result+"'");
			} else {
				result = result + c;
			}
		}
		return result;
	}

	/**
	 * generiert aus einen String s die CondensedValues
	 */
	public static int[] generateCondensedValue(String s) {
		String intelligentKeyString = getIntelligentKeyString(" " + s);
		int[] result = new int[(intelligentKeyString.length() - 1) / 6 + 1];

		for (int i = 0; i < result.length; i++) {
			result[i] = 0;
		}
		for (int i = 0; i < intelligentKeyString.length(); i++) {
			result[i / 6] |= ((int) intelligentKeyString.charAt(i) - (int) '0') << ((i % 6) * 5);
			//			showMatchInt(result[i/6], "generating Byte " + (i/6) + " char " + (i % 6));
		}
		return result;
	}

	/**
	 * matchIntelligentKey berechnet die Möglichkeiten mit der zuletzt eingegebenen Taste
	 */
	private void matchIntelligentKey() {
		String intelligentKeyString;
		intelligentKeyString = (String) keys.peek();
		int intelligentKeyLength = intelligentKeyString.length();

		if (intelligentKeyLength > 0) {
			Vector matchedItems = new Vector();

			Object[] previousItemLocations;
			previousItemLocations = (Object[]) matchedItemsOfPos.peek();

			for (int locationNo = 0; locationNo < previousItemLocations.length; locationNo++) {
				boolean skipLocation;
				skipLocation = false;

				AbstractLocation currentLocation;
				String locationName;
				int searchPos;

				/* 
				 * Fallunterscheidung für matchedItemsOfPos:
				 * matchedItemsOfPos.elementAt(0) ist vom Typ AbstractLocation[]
				 * die anderen vom Typ ItemLocation[]
				 */
				if (intelligentKeyLength > 1) {
					ItemLocation currentItem = (ItemLocation) previousItemLocations[locationNo];
					currentLocation = currentItem.getLocation();
				} else {
					currentLocation = ((AbstractLocation[]) previousItemLocations)[locationNo];
				}
				locationName = currentLocation.getScreenName();

				if (intelligentKeyLength > 1) {
					ItemLocation currentItem = (ItemLocation) previousItemLocations[locationNo];
					currentLocation = currentItem.getLocation();
					if (currentItem instanceof ItemLocationShowAll) {
						continue;
					}

					locationName = currentLocation.getScreenName();

					int nameLength = locationName.length();

					int lastMatchPos = currentItem.getIntelligentKeyMatchPos();

					if (lastMatchPos == IntelligentKey.PATTERN_NOT_FOUND) {
						skipLocation = true;
					}

					if ((!skipLocation) && (nameLength < intelligentKeyLength)) {
						skipLocation = true;
					}

					if ((!skipLocation) && (nameLength < lastMatchPos + intelligentKeyLength)) {
						skipLocation = true;
					}

					if (lastMatchPos == IntelligentKey.PATTERN_NOT_SEARCHED) {
						searchPos = 0;
					} else {
						searchPos = lastMatchPos;
					}
				} else {
					searchPos = 0;
				}

				int pos = PATTERN_NOT_FOUND;

				if (!skipLocation) {
					if (intelligentKeyLength <= 5) { // optimierte Suche mit condensedValues
						int matchLength = intelligentKeyLength + 1;
						// optimierte Version
						/*						System.out.println("");
												System.out.println(" ------------------------------------------------- ");
												System.out.println(""); */
						//						System.out.println("matchlength = " + matchLength);
						//						System.out.println("currentLocation " + currentLocation.getScreenName() + "  id " + Data.getHaltestellenId((Haltestelle) currentLocation));
						int[] locationCondensedValues = currentLocation.getIntelligentKeyCondensedValues();
						//						for (int i = locationCondensedValues.length - 1; i >= 0; i--) {
						//							showCondensedValue(locationCondensedValues[i], "byte " + i);
						//						}
						int maxMatchPos = locationName.length() + 1 - matchLength;
						//						System.out.println("maxMatchPos = " + maxMatchPos);

						int userCondensedValue = condensedValue & condensedMask[matchLength];
						//						showMatchInt(currentKeyInt, " currentKeyInt");

						int startIntNo = (searchPos) / 6;
						//						System.out.println("startInt = " + startInt);
						int startCharNo = (searchPos) % 6;
						//						System.out.println("startCharNo = " + startCharNo);
						int endInt = (searchPos + matchLength - 1) / 6;
						//						System.out.println("endInt = " + endInt);

						int locationCondensedValue = locationCondensedValues[startIntNo];
						//						showMatchInt(locationInt, "locationInt");

						int verifyCondensedValue = locationCondensedValue >> (startCharNo * 5);
						//						showMatchInt(verifyInt, "verifyInt");

						if (endInt != startIntNo) {
							//							System.out.println("endInt != startInt");
							int endIntPos = (searchPos + matchLength - 1) % 6;
							//							System.out.println("endIntPos = " + endIntPos);
							int secondLocationCondensedValue = locationCondensedValues[endInt];
							//							showMatchInt(secondLocationInt, "secondLocationInt");
							secondLocationCondensedValue = secondLocationCondensedValue << ((6 - startCharNo) * 5) & condensedMask[6];
							//							showMatchInt(secondLocationInt, "secondLocationInt");
							verifyCondensedValue |= secondLocationCondensedValue;
							//							showMatchInt(verifyInt, "verifyInt - mit second");
						}

						int validChars = 6;

						//						showCondensedValue((verifyCondensedValue & condensedMask[matchLength]), "verifyInt & keyMask[matchLength]");
						//						showCondensedValue(userCondensedValue, "currentKeyInt");

						if (((verifyCondensedValue & condensedMask[matchLength]) ^ userCondensedValue) == 0) {
							//							System.out.println("matched");
							pos = searchPos;
						} else {
							//							System.out.println("no match");
							pos = PATTERN_NOT_FOUND;
							for (int startMatchPos = searchPos + 1; startMatchPos <= maxMatchPos; startMatchPos++) {
								//								System.out.println("startMatchPos = " + startMatchPos);
								//								System.out.println("validChars = " + validChars);
								verifyCondensedValue = verifyCondensedValue >> 1 * 5;
								validChars--;
								//								System.out.println("validChars = " + validChars);
								//								showCondensedValue(verifyCondensedValue, "verifyCondensedValue");
								if (validChars - matchLength < 0) {
									//									System.out.println("require Data");
									int charNoToAquire = startMatchPos + matchLength - 1;
									//									System.out.println("charNoToAquire = " + charNoToAquire);
									int newIntNo = charNoToAquire / 6;
									//									System.out.println("newInt = " + newIntNo);
									int newCharNo = charNoToAquire % 6;
									//									System.out.println("newCharNo = " + newCharNo);
									int newLocationCondensedValue = locationCondensedValues[newIntNo];
									//									showCondensedValue(newLocationCondensedValue, "newLocationCondensedValue");
									newLocationCondensedValue = newLocationCondensedValue >> (newCharNo * 5);
									//									showCondensedValue(newLocationCondensedValue, "newLocationCondensedValue");
									newLocationCondensedValue = newLocationCondensedValue << ((matchLength - 1) * 5);
									//									showCondensedValue(newLocationCondensedValue, "newLocationCondensedValue");
									verifyCondensedValue = (verifyCondensedValue & condensedMask[matchLength - 1]) | (newLocationCondensedValue & condensedMask[6]);
									//									showCondensedValue(verifyCondensedValue, "verifyCondensedValue");
									validChars += Math.min(6 - newCharNo, 6 - matchLength + 1);
									//									System.out.println("validChars = " + validChars);
								}
								//								showCondensedValue(verifyCondensedValue & condensedMask[matchLength], "verifyInt & keyMask[matchLength]");
								//								showCondensedValue(userCondensedValue, "currentKeyInt");
								if (((verifyCondensedValue & condensedMask[matchLength]) ^ userCondensedValue) == 0) {
									pos = startMatchPos;
									//									System.out.println("match at pos = " + pos);
									break;
								} else {
									//									System.out.println("no Match");
								}
							}
							if (pos == PATTERN_NOT_FOUND) {
								skipLocation = true;
							}
						}

					} else { // Standard IntelligentKey matching

						String IntelligentKeyLocationName = getIntelligentKeyString(locationName);

						pos = IntelligentKeyLocationName.indexOf(intelligentKeyString, searchPos);

						if (pos == PATTERN_NOT_FOUND) {
							skipLocation = true;
						}

						// nur den Match verwenden falls ein neues Wort beginnt
						// durch replaceCharactersForIntelligentKeyStyle werden auch -.() in ' ' -> Tastencode 0 gewandelt 
						if (pos > 0) {
							// der Match muss auf ein neues Wort fallen
							// falls nicht muss nochmal nach ' '+IntelligentKeyString gesucht werden
							if (IntelligentKeyLocationName.charAt(pos - 1) != '0') {
								pos = IntelligentKeyLocationName.indexOf("0" + intelligentKeyString, searchPos);
								if (pos == PATTERN_NOT_FOUND) {
									skipLocation = true;
								}
								// korrektur der Position da vorne noch die '0' stehen muss 
								pos += 1;
							}
						}
					}
				}

				if (!skipLocation) {
					matchedItems.addElement(new ItemLocation(currentLocation, (short) pos, this));
				}
			}

			ItemLocation[] tmp = new ItemLocation[matchedItems.size()];
			matchedItems.copyInto(tmp);
			matchedItemsOfPos.push(tmp);
		}

		return;
	}

	/**
	 * liefert ein ItemLocation[] zurück, das alle passenden Einträge zu den
	 * aktuellen gedrückten Taste liefert
	 */
	public ItemLocation[] getItemLocations() {
		if (matchedItemsOfPos.size() == 1) {
			System.err.println("Position 1 enthält keine ItemLocation ! [getItemLocations]");
			return null;
		}

		return (ItemLocation[]) matchedItemsOfPos.peek();
	}

	/**
	 * wird IntelligentKey verwendet ?
	 */
	public boolean isInUse() {
		if (matchedItemsOfPos.size() > 1) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * kann IntelligentKey Items zurückliefern ?
	 */
	public boolean hasValidEntries() {
		if (!isInUse()) {
			return true;
		}

		if (matchedItemsOfPos.size() == 1) {
			System.err.println("Position 1 enthält keine ItemLocation ! [noValidEntries]");
			return false;
		}

		if (((ItemLocation[]) matchedItemsOfPos.peek()).length == 0) {
			return false;
		} else {
			return true;
		}
	}

	/**
	 * gibt die Anzahl der eingegeben Tasten zurück
	 */
	public int getKeysLength() {
		return keys.size() - 1;
	}

	/**
	 * gibt den String zurück der im Header angezeigt werden soll
	 */
	public String getIntelligentKeyHeaderString() {
		// Falls kein gültiger String gefunden wurde
		String result;

		String key = (String) keys.peek();
		result = "";
		for (int i = 0; i < key.length() - 1; i++) {
			result += IntelligentKeyMap[key.charAt(i) - 48].charAt(0);
		}
		ItemLocation[] locations = (ItemLocation[]) matchedItemsOfPos.peek();
		if (locations.length == 0) {
			result += "?";
		} else {
			result += IntelligentKeyMap[key.charAt(key.length() - 1) - 48].charAt(0);
		}

		return result.toUpperCase();
	}

	/**
	 * Debug-Funktion zeigt den übergebenen condensedValue lesbar an
	 */
	private static void showCondensedValue(int condensedValue, String s) {
		String out = "";
		for (int i = 31; i >= 0; i--) {
			if (i % 5 == 4) {
				out = out + ".";
			}
			if ((condensedValue & (1 << i)) != 0) {
				out = out + "1";
			} else {
				out = out + "0";
			}
		}
		System.out.println(out + " " + s);
		out = " ^ ";
		for (int i = 5; i >= 0; i--) {
			int value = condensedValue >> (5 * i);
			value &= condensedMask[1];
			char c = (char) ((int) '0' + value);
			out = out + "    " + c;
		}
		System.out.println(out);

	}

	public int getKeyCode(char lowerCaseKey) {
		String s = getIntelligentKeyString(lowerCaseKey + "");
		if (s.length() != 1) {
			return (byte) ' ';
		}
		int result = (byte) s.charAt(0);
		return result;
	}
}
