// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package model;

import java.util.Calendar;
import java.util.TimeZone;
import java.util.Vector;

import utils.DynamicIntMap;
import utils.StringUtils;
import control.Device;
import control.Global;
import dev.bootProfiler.ZeitMessung;

public final class DatumZeit {

	// NULL-Wert
	public static final int NULL = -1;

	// Vordefinierte Werte
	public static final int MIN_IN_MS = 60000;
	public static final long MIN_IN_MS_AS_LONG = 60000;
	public static final int DAY_IN_MIN = 1440;
	
	private static final String[] weekDayStrings = { "Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag" };

	/**
	 * Berechnet den Offset der Zeit die auf dem Handy Disply angezeigt wird, zu
	 * der currentTimeInMilis, damit wir nicht immer über den GMT Calendar gehen
	 * müssen.
	 */
	private final static long displayToCurrentMillisOffset;
	private static final short[] monthData;

	static {
		Calendar calCurrentPhoneDisplayTime = Calendar.getInstance();
		long now = System.currentTimeMillis();

		Calendar calCurrentTimeInGmt = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
		calCurrentTimeInGmt.set(Calendar.YEAR, calCurrentPhoneDisplayTime.get(Calendar.YEAR));
		calCurrentTimeInGmt.set(Calendar.MONTH, calCurrentPhoneDisplayTime.get(Calendar.MONTH));
		calCurrentTimeInGmt.set(Calendar.DAY_OF_MONTH, calCurrentPhoneDisplayTime.get(Calendar.DAY_OF_MONTH));
		calCurrentTimeInGmt.set(Calendar.HOUR_OF_DAY, calCurrentPhoneDisplayTime.get(Calendar.HOUR_OF_DAY));
		calCurrentTimeInGmt.set(Calendar.MINUTE, calCurrentPhoneDisplayTime.get(Calendar.MINUTE));
		calCurrentTimeInGmt.set(Calendar.SECOND, calCurrentPhoneDisplayTime.get(Calendar.SECOND));
		calCurrentTimeInGmt.set(Calendar.MILLISECOND, calCurrentPhoneDisplayTime.get(Calendar.MILLISECOND));

		displayToCurrentMillisOffset = calCurrentTimeInGmt.getTime().getTime() - now;
		
		monthData = new short[] { 306, 337, 0, 31, 61, 92, 122, 153, 184, 214, 245, 275 };
	}

	/**
	 * Aktuelle push&ride Zeit
	 */
	public static int Now() {
		if (Global.SET_TIME_TO_DEFAULT_VALUE) {
			return DatumZeit.createDatumZeit(2010, 12, 17, 07, 00);
		} else {
			// Muss jedes Mal geholt werden, da sich der Offset durch Änderung in den Settings ändern kann.
			int timeOffsetInMin = Device.getTimeOffset();

			// Wendet den Offset auf currentTimeMillis an, damit diese der Display Time entsprechen. Danach
			// werden die Milisekunden in Minuten umgerechnet und der vom User eingestellte timeOffset
			// angewendet.
			int result = (int) ((System.currentTimeMillis() + displayToCurrentMillisOffset) / MIN_IN_MS_AS_LONG) + timeOffsetInMin;

			return result;
		}
	}

	/**
	 * Konvertiert eine push&ride Zeit zurück in die Zeit des Gerätes.
	 */
	public static long convertToDeviceTime(int datumZeit) {
		Calendar calPhoneTime = Calendar.getInstance();
		calPhoneTime.set(Calendar.YEAR, getYear(datumZeit));
		calPhoneTime.set(Calendar.MONTH, getMonthOfYear(datumZeit) - 1);
		calPhoneTime.set(Calendar.DAY_OF_MONTH, getDayOfMonth(datumZeit));
		calPhoneTime.set(Calendar.HOUR_OF_DAY, getHourOfDay(datumZeit));
		calPhoneTime.set(Calendar.MINUTE, getMinutesOfHour(datumZeit));
		calPhoneTime.set(Calendar.SECOND, 0);
		calPhoneTime.set(Calendar.MILLISECOND, 0);
		long result = calPhoneTime.getTime().getTime();
		int timeOffset = Device.getTimeOffset();
		if (timeOffset != 0) {
			result -= timeOffset * MIN_IN_MS_AS_LONG;
		}

		return result;
	}

	private static final ZeitMessung zeitMessungCreateDatumZeit = ZeitMessung.getInstance("DatumZeit.createDatumZeit");

	/**
	 * Erstellt eine neue push&ride Zeit
	 * 
	 * @param year
	 * @param month
	 *            1-12
	 * @param day
	 *            1-31
	 * @param h
	 *            0-24
	 * @param min
	 *            0-60
	 * @return
	 */
	public static int createDatumZeit(int year, int month, int day, int h, int min) {
		zeitMessungCreateDatumZeit.start();
		int datumZeit = getDatumZeitFromYearMonthDay(year, month, day);
		datumZeit += h * 60 + min;
		zeitMessungCreateDatumZeit.stop();
		return datumZeit;
	}

	// ///////////////////////////////////////////////////////////////////////////////////////////////////
	// GETTER (Jahr, Monat, Tag, Stunde, Minute, ...)
	// ///////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Gibt das aktuelle Jahr zurück.
	 * 
	 * @return int
	 */
	public static int getCurrentYear() {
		return convertDatumZeit(Now()).year;
	}

	/**
	 * Extrahiert das Jahr aus der übergebenen push&ride Zeit.
	 * 
	 * @param datumZeit
	 */
	public static int getYear(int datumZeit) {
		return convertDatumZeit(datumZeit).year;
	}

	/**
	 * Extrahiert den Monat aus der übergebenen push&ride Zeit.
	 * <p>
	 * 1=Januar bis 12=Dezember
	 * 
	 * @param datumZeit
	 */
	public static int getMonthOfYear(int datumZeit) {
		return convertDatumZeit(datumZeit).month;
	}

	/**
	 * Extrahiert den Tag des Monats aus der übergebenen push&ride Zeit.
	 * 
	 * @param datumZeit
	 */
	public static int getDayOfMonth(int datumZeit) {
		return convertDatumZeit(datumZeit).day;
	}

	/**
	 * Extrahiert den Tag (die erste Minute des Tags) aus der übergebenen
	 * push&ride Zeit.
	 * <p>
	 * Das ist in push&ride der Identifieren für einen Tag, der dann auch
	 * eindeutig einem Datum zugerodnet werden kann.
	 * 
	 * @param datumZeit
	 */
	public static int getDay(int datumZeit) {
		return datumZeit - (datumZeit % DAY_IN_MIN);
	}

	/**
	 * Extrahiert die Stunde des Tages (im 24-Stunden Format) aus der
	 * übergebenen push&ride Zeit.
	 * 
	 * @param datumZeit
	 */
	public static int getHourOfDay(int datumZeit) {
		return getMinutesOfDay(datumZeit) / 60;
	}

	/**
	 * Extrahiert die Minute (seit Beginn des Tages!) aus der übergebenen
	 * push&ride Zeit.
	 * 
	 * @param datumZeit
	 */
	public static int getMinutesOfDay(int datumZeit) {
		return datumZeit % DAY_IN_MIN;
	}

	/**
	 * Extrahiert die Minute (der aktuellen Stunde, also wie in der normalen
	 * Uhrzeit) aus der übergebenen push&ride Zeit.
	 * 
	 * @param datumZeit
	 */
	public static int getMinutesOfHour(int datumZeit) {
		return datumZeit % 60;
	}

	// ///////////////////////////////////////////////////////////////////////////////////////////////////
	// DAY TYPE UND WOCHENTAGE
	// ///////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Returns the weekday of a given date time integer 0-Sonntag 6-Samstag
	 */
	public static int getWeekday(int datumZeit) {
		return ((datumZeit / DAY_IN_MIN) + 4) % 7;
	}

	/**
	 * Returns the weekday of a given date time integer 0-Sonntag 6-Samstag
	 */
	public static String getWeekdayAsString(int datumZeit) {
		return DatumZeit.weekDayStrings[DatumZeit.getWeekday(datumZeit)];
	}

	// Second Level Cache
	private static DynamicIntMap cache_getDayType_second = new DynamicIntMap(10);

	// First Level Cache - Bei Größe 4 haben wir 90% FirstCache Hits
	private static int[] cache_getDayType_first_keys = { -1, -1, -1, -1, -1 };
	private static final int cache_getDayType_first_size = cache_getDayType_first_keys.length;
	private static int[] cache_getDayType_first_values = new int[cache_getDayType_first_size];
	private static int cache_getDayType_first_replacePointer = 0;

	// Stats
	public static int counterCacheMisses1 = 0;
	public static int counterCacheHits1 = 0;
	public static int counterCacheMisses2 = 0;
	public static int counterCacheHits2 = 0;

	/**
	 * Gibt den Typ des Tages zurück, mit dem der passende Fahrplan ausgewählt
	 * wird. Mit Super-Caching
	 * 
	 * @param datumZeit
	 * @return TAG_SONN_FEIERTAG, TAG_SAMSTAG, TAG_WOCHENTAG_FERIEN,
	 *         TAG_WOCHENTAG_SCHULE
	 */
	synchronized public static int getDayType(int datumZeit) {
		int day = DatumZeit.getDay(datumZeit);

		// FIRST LEVEL CACHE
		for (int i = 0; i < cache_getDayType_first_keys.length; i++) {
			if (cache_getDayType_first_keys[i] == day) {
				counterCacheHits1++;
				return cache_getDayType_first_values[i];
			}
		}
		counterCacheMisses1++;

		// SECOND LEVEL CACHE
		// Cache Hit
		int cachePosition = cache_getDayType_second.lookupPosition(day);
		if (cachePosition != DynamicIntMap.KEY_NOT_FOUND) {
			counterCacheHits2++;
			int dayType = cache_getDayType_second.get(cachePosition);

			// First Level Cache Management (1/2)
			cache_getDayType_first_keys[cache_getDayType_first_replacePointer] = day;
			cache_getDayType_first_values[cache_getDayType_first_replacePointer] = dayType;
			cache_getDayType_first_replacePointer = (cache_getDayType_first_replacePointer + 1) % cache_getDayType_first_size;
			return dayType;

		} else {
			// Cache Miss
			counterCacheMisses2++;
			int dayType = -1;
			int weekday = DatumZeit.getWeekday(datumZeit);
			if (weekday == 0 || isFeiertag(datumZeit)) {
				dayType = Global.TAG_SONN_FEIERTAG;
			} else if (weekday == 6) {
				dayType = Global.TAG_SAMSTAG;
			} else if (isSchulFerienTag(datumZeit)) {
				dayType = Global.TAG_WOCHENTAG_FERIEN;
			} else {
				dayType = Global.TAG_WOCHENTAG_SCHULE;
			}
			
			// Second Level Cache Management
			cache_getDayType_second.add(day, dayType);

			// First Level Cache Management (2/2)
			cache_getDayType_first_keys[cache_getDayType_first_replacePointer] = day;
			cache_getDayType_first_values[cache_getDayType_first_replacePointer] = dayType;
			cache_getDayType_first_replacePointer = (cache_getDayType_first_replacePointer + 1) % cache_getDayType_first_size;

			return dayType;
		}
	}

	private static int[] feiertageCache = null;

	public static boolean isFeiertag(int datumZeit) {
		if (feiertageCache == null) {
			ZeitMessung profileIsFeiertag = ZeitMessung.getStartedInstance("DatumZeit.isFeiertag() - init");

			Vector feiertage = new Vector();
			// Feste Feiertage (machen wir für das letzte, das aktuell und nächste Jahr rein)
			for (int year = getCurrentYear() - 1; year <= getCurrentYear() + 1; year++) {
				// Neujahr 01.01.
				feiertage.addElement(new Integer(DatumZeit.createDatumZeit(year, 1, 1, 0, 0)));

				// Heilige Drei Könige 06.01.
				feiertage.addElement(new Integer(DatumZeit.createDatumZeit(year, 1, 6, 0, 0)));

				// Tag der Arbeit 01.05.
				feiertage.addElement(new Integer(DatumZeit.createDatumZeit(year, 5, 1, 0, 0)));

				// Tag der Deutschen Einheit 03.10.
				feiertage.addElement(new Integer(DatumZeit.createDatumZeit(year, 10, 3, 0, 0)));

				// Allerheiligen 01.11.
				feiertage.addElement(new Integer(DatumZeit.createDatumZeit(year, 11, 1, 0, 0)));

				// 1. Weihnachtstag 25.12.
				feiertage.addElement(new Integer(DatumZeit.createDatumZeit(year, 12, 25, 0, 0)));

				// 2. Weihnachtstag 26.12.
				feiertage.addElement(new Integer(DatumZeit.createDatumZeit(year, 12, 26, 0, 0)));
			}

			// Feiertage BW: http://www.feiertage.net/frei-tage.php
			// Fr. 22.04.2011 Karfreitag
			feiertage.addElement(new Integer(DatumZeit.createDatumZeit(2011, 4, 22, 0, 0)));
			// Mo. 25.04.2011 Ostermontag
			feiertage.addElement(new Integer(DatumZeit.createDatumZeit(2011, 4, 25, 0, 0)));
			// Do. 02.06.2011 Christi Himmelfahrt
			feiertage.addElement(new Integer(DatumZeit.createDatumZeit(2011, 6, 2, 0, 0)));
			// Mo. 13.06.2011 Pfingstmontag
			feiertage.addElement(new Integer(DatumZeit.createDatumZeit(2011, 6, 13, 0, 0)));
			// Do. 23.06.2011 Fronleichnam
			feiertage.addElement(new Integer(DatumZeit.createDatumZeit(2011, 6, 23, 0, 0)));

			// Fr. 06.04.2012 Karfreitag
			feiertage.addElement(new Integer(DatumZeit.createDatumZeit(2012, 4, 6, 0, 0)));
			// Mo. 09.04.2012 Ostermontag
			feiertage.addElement(new Integer(DatumZeit.createDatumZeit(2012, 4, 9, 0, 0)));
			// Do. 17.05.2012 Christi Himmelfahrt
			feiertage.addElement(new Integer(DatumZeit.createDatumZeit(2012, 5, 17, 0, 0)));
			// Mo. 28.05.2012 Pfingstmontag
			feiertage.addElement(new Integer(DatumZeit.createDatumZeit(2012, 5, 28, 0, 0)));
			// Do. 07.06.2012 Fronleichnam
			feiertage.addElement(new Integer(DatumZeit.createDatumZeit(2012, 6, 7, 0, 0)));
			
			// Und in ein Array cachen
			feiertageCache = new int[feiertage.size()];
			for (int i = 0; i < feiertage.size(); i++) {
				feiertageCache[i] = ((Integer) feiertage.elementAt(i)).intValue();
			}

			profileIsFeiertag.stop();
		}

		// Prüft ob es ein Feiertag ist
		for (int i = 0; i < feiertageCache.length; i++) {
			if (feiertageCache[i] <= datumZeit && datumZeit < feiertageCache[i] + DAY_IN_MIN) {
				return true;
			}
		}
		return false;
	}

	private static int[] schultageCache_FerienVon = null;
	private static int[] schultageCache_FerienBis = null;

	public static boolean isSchulFerienTag(int datumZeit) {
		if (schultageCache_FerienVon == null || schultageCache_FerienBis == null) {
			ZeitMessung profileIsSchultag = ZeitMessung.getStartedInstance("DatumZeit.isSchultag() - init");
			Vector ferienVon = new Vector();
			Vector ferienBis = new Vector();

			// Schulferien BW
			// + VVS Ferienplaner: http://www.vvs.de/service/downloads/verschiedenes/
			// Weihnachtsferien 23.12.2010 - 09.01.2011
			ferienVon.addElement(new Integer(DatumZeit.createDatumZeit(2010, 12, 23, 0, 0)));
			ferienBis.addElement(new Integer(DatumZeit.createDatumZeit(2011, 1, 9, 0, 0)));

			// Faschingsferien ????? 

			// Osterferien 21.04. - 30.04.
			ferienVon.addElement(new Integer(DatumZeit.createDatumZeit(2011, 4, 21, 0, 0)));
			ferienBis.addElement(new Integer(DatumZeit.createDatumZeit(2011, 4, 30, 0, 0)));
			// Pfingstferien 14.06. - 25.06.
			ferienVon.addElement(new Integer(DatumZeit.createDatumZeit(2011, 6, 14, 0, 0)));
			ferienBis.addElement(new Integer(DatumZeit.createDatumZeit(2011, 6, 25, 0, 0)));
			// Sommerferien 28.07. - 10.09.
			ferienVon.addElement(new Integer(DatumZeit.createDatumZeit(2011, 7, 28, 0, 0)));
			ferienBis.addElement(new Integer(DatumZeit.createDatumZeit(2011, 9, 10, 0, 0)));
			// Herbstferien 31.10. - 06.11.
			ferienVon.addElement(new Integer(DatumZeit.createDatumZeit(2011, 10, 31, 0, 0)));
			ferienBis.addElement(new Integer(DatumZeit.createDatumZeit(2011, 11, 6, 0, 0)));
			// Weihnachtsferien 23.12.2011 - 07.01.2012
			ferienVon.addElement(new Integer(DatumZeit.createDatumZeit(2011, 12, 23, 0, 0)));
			ferienBis.addElement(new Integer(DatumZeit.createDatumZeit(2012, 1, 6, 0, 0)));

			// Und in ein Array cachen
			schultageCache_FerienVon = new int[ferienVon.size()];
			for (int i = 0; i < ferienVon.size(); i++) {
				schultageCache_FerienVon[i] = ((Integer) ferienVon.elementAt(i)).intValue();
			}
			schultageCache_FerienBis = new int[ferienBis.size()];
			for (int i = 0; i < ferienBis.size(); i++) {
				schultageCache_FerienBis[i] = ((Integer) ferienBis.elementAt(i)).intValue();
			}

			profileIsSchultag.stop();
		}

		// Prüft ob es ein Schultag ist
		for (int i = 0; i < schultageCache_FerienVon.length; i++) {
			if (schultageCache_FerienVon[i] <= datumZeit && datumZeit < schultageCache_FerienBis[i] + DAY_IN_MIN) {
				return true;
			}
		}
		return false;
	}

	// ///////////////////////////////////////////////////////////////////////////////////////////////////
	// FORMATTER
	// ///////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Formatiert eine push&ride Zeit im Format: 12:34
	 * 
	 * @param datumZeit
	 */
	public static String toTimeString(int datumZeit) {
		return StringUtils.fillInt(getHourOfDay(datumZeit), '0', 2) + ":" + StringUtils.fillInt(getMinutesOfHour(datumZeit), '0', 2);
	}

	/**
	 * Formatiert eine push&ride Zeit im Format: 12:34 12.12.1234
	 * 
	 * @param datumZeit
	 */
	public static String toDateTimeString(int datumZeit) {
		return toTimeString(datumZeit) + " " + toDateYearString(datumZeit);
	}

	/**
	 * Formatiert eine push&ride Zeit im Format: 12:34 12.12.1234
	 * 
	 * @param datumZeit
	 */
	public static String toDateTimeTypeString(int datumZeit) {
		return toDateTimeString(datumZeit) + " (" + dayTypeToString(getDayType(datumZeit)) + ")";
	}

	/**
	 * Formatiert eine push&ride Zeit im Format: 12:34 12.12.1234 (Schule)
	 * 
	 * @param datumZeit
	 */
	private static String dayTypeToString(int dayType) {
		switch (dayType) {
			case Global.TAG_SAMSTAG: return "Samstag";
			case Global.TAG_SONN_FEIERTAG: return "Sonn-/Ferientag";
			case Global.TAG_WOCHENTAG_FERIEN: return "Ferien";
			case Global.TAG_WOCHENTAG_SCHULE: return "Schule";
			default: return "unbekannt";
		}
	}

	/**
	 * Formatiert eine push&ride Zeit im Format: 22.11.
	 * 
	 * @param datumZeit
	 */
	public static String toDateString(int datumZeit) {
		return StringUtils.fillInt(getDayOfMonth(datumZeit), '0', 2) + "." + StringUtils.fillInt(getMonthOfYear(datumZeit), '0', 2) + ".";
	}

	/**
	 * Formatiert eine push&ride Zeit im Format: 22.11.2008
	 * 
	 * @param datumZeit
	 */
	public static String toDateYearString(int datumZeit) {
		return toDateString(datumZeit) + getYear(datumZeit);
	}

	// ///////////////////////////////////////////////////////////////////////////////////////////////////
	// HELPER
	// ///////////////////////////////////////////////////////////////////////////////////////////////////


	// modified code from http://saxon.svn.sourceforge.net/svnroot/saxon/ 

	/**
	 * Erstellt eine neue push&ride Zeit aus einem Datum
	 * 
	 * @param year
	 *            1970-xxxx
	 * @param month
	 *            1-12
	 * @param day
	 *            1-31
	 * @return
	 */
	private static int getDatumZeitFromYearMonthDay(int year, int month, int day) {
		int z = year - (month < 3 ? 1 : 0);
		short f = monthData[month - 1];
		int result = day + f + 365 * z + z / 4 - z / 100 + z / 400 - 719469;
		return result * DAY_IN_MIN;
	}

	/**
	 * Erstellt ein Datum aus einer push&ride Zeit
	 */
	private static DateValue convertDatumZeit(int datumZeit) {
		datumZeit /= DAY_IN_MIN;
		int L = datumZeit + 2509157;
		int n = (4 * L) / 146097;
		L = L - (146097 * n + 3) / 4;
		int i = (4000 * (L + 1)) / 1461001;
		L = L - (1461 * i) / 4 + 31;
		int j = (80 * L) / 2447;
		int d = L - (2447 * j) / 80;
		L = j / 11;
		int m = j + 2 - (12 * L);
		int y = 100 * (n - 49) + i + L;
		return new DateValue(y, (byte) m, (byte) d);
	}

}

final class DateValue {
	protected int year; 
	protected byte month;
	protected byte day;

	/**
	 * Constructor given a year, month, and day. Performs no validation.
	 * 
	 * @param year
	 *            The year, 1970-xxxx
	 * @param month
	 *            The month, 1-12
	 * @param day
	 *            The day 1-31
	 */

	public DateValue(int year, byte month, byte day) {
		this.year = year;
		this.month = month;
		this.day = day;
	}

}
