// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package model;

import utils.DynamicIntMap;
import control.Global;
import data.Data;
import dev.Logger;
import dev.Memory;

public class Fahrt implements IFahrt {
	// NULL-Werte
	public static final Fahrt NULL = new Fahrt(null, null, null, null, null, null, null);
	
	// Arrays mit den geordneten Abfahrtszeiten an der ersten Haltestelle, für die verschiedenen Tagarten
	private final short[] abfahrtszeitenWochentagSchule;
	private final short[] abfahrtszeitenWochentagFerien;
	private final short[] abfahrtszeitenSamstag;
	private final short[] abfahrtszeitenSonnFeiertag;
	
	// Fahrzeiten zwischen den Haltestellen.
	// Die Haltestellen ID (int) ist der Key, der Value ist ein int mit den Minuten, die bis zur *nächsten*
	// Haltestelle benötigt werden.
	private final DynamicIntMap fahrzeiten;
	
	// Kumulierte Fahrzeiten ZU den Haltestellen!
	// Also der Key ist die Haltestellen ID (int), der Value ein int, wieviel Zeit von der ersten Haltestelle
	// zur übergeben Haltestelle benötigt wird.
	private final DynamicIntMap fahrzeitenKumuliert;

	// Auf welcher Linie befindet sich diese Fahrt
	private final Linie linie;
	
	// Array der Haltestellen (geordnet, Position 0 = erste Haltestelle)
	private final Haltestelle[] haltestellen;

	/**
	 * Konstruktor fuer die Fahrts. Damit die Variablen als final deklariert werden koennen, kann eine Linie
	 * nur ueber diesen Konstruktor angelegt werden.
	 * 
	 * @param haltestellen
	 *            - Array der Haltestellen (geordnet, Position 0 = erste Haltestelle)
	 * @param abfahrtszeiten
	 *            - Abfahrtszeiten an der ersten Haltestelle, geordnet
	 * @param fahrZeiten
	 *            - Fahrzeiten zwischen den Haltestellen Position 0 = Fahrzeit von der Haltestelle 0 zur
	 *            Haltestelle 1 im Haltestellen Array
	 */
	public Fahrt(Linie linie, Haltestelle[] haltestellen, short[] abfahrtsZeitenWochentagSchule,
		short[] abfahrtsZeitenWochentagFerien, short[] abfahrtsZeitenSamstag,
		short[] abfahrtsZeitenSonnFeiertage, int[] fahrZeiten) {

		if (haltestellen == null) {
			// NULL FAHRT
			this.linie = null;
			this.haltestellen = null;
			this.abfahrtszeitenWochentagSchule = null;
			this.abfahrtszeitenWochentagFerien = null;
			this.abfahrtszeitenSamstag = null;
			this.abfahrtszeitenSonnFeiertag = null;
			this.fahrzeiten = null;
			this.fahrzeitenKumuliert = null;
			return;
		}
		
		fahrzeiten = new DynamicIntMap(haltestellen.length - 1);
		fahrzeitenKumuliert = new DynamicIntMap(haltestellen.length);

		// Testet ob die Anzahl der Haltestellen mit der Anzahl der uebergeben
		// Fahrzeiten uebereinstimmt.
		if (haltestellen.length - 1 == fahrZeiten.length) {

			// Zuweisungen
			this.linie = linie;
			this.haltestellen = haltestellen;
			this.abfahrtszeitenWochentagSchule = abfahrtsZeitenWochentagSchule;
			this.abfahrtszeitenWochentagFerien = abfahrtsZeitenWochentagFerien;
			this.abfahrtszeitenSamstag = abfahrtsZeitenSamstag;
			this.abfahrtszeitenSonnFeiertag = abfahrtsZeitenSonnFeiertage;
			
			// Kumulierte Fahrzeiten berechnen
			int zeit = 0;
			
			// fahrzeit zur ersten Haltestelle = 0
			fahrzeitenKumuliert.add(haltestellen[0].getId(), zeit);

			// verarbeitung fzeiten
			for (int i = 0; i < fahrZeiten.length; i++) {
				// Fahrzeiten in Array
				fahrzeiten.add(haltestellen[i].getId(), fahrZeiten[i]);
				
				// fahrzeit zur (i+1)-ten Haltestelle
				// Zeit addieren die bis zur nächsten Haltestelle gebraucht
				// wird.
				zeit += fahrZeiten[i];
				
				// Abfangen der letzten Haltestelle
				if (i + 1 < this.haltestellen.length) {
					fahrzeitenKumuliert.add(haltestellen[i + 1].getId(), zeit);
				}
			}

			// Falls die Anzahl der Haltestellen und Fahrzeiten nicht stimmt.
		} else {
			// Leere Zuweisungen
			this.linie = null;
			this.haltestellen = null;
			this.abfahrtszeitenWochentagSchule = null;
			this.abfahrtszeitenWochentagFerien = null;
			this.abfahrtszeitenSamstag = null;
			this.abfahrtszeitenSonnFeiertag = null;

			// Fehlermeldung Konsole
			Logger.error(this, "Eingabefehler:\nArray mit Haltestellen nicht " + "konsistent mit "
				+ "Fahrzeitenarray! Die Daten wurden nicht in " + "der Linie angelegt! Das Fahrzeiten muss "
				+ "bei Haltestellen genau n-1 Fahrzeiten zwischen den Haltestellen angeben.");
		}
	}

	/**
	 * Gibt den Name zurück der beim Routen Ergebnis angezeigt werden soll.
	 */
	public String getScreenName() {
		// FIXME: Eigentlich sollte nicht Linie.EndHaltestellen sondern die
		// tatsächliche EndHaltestellen ausgegeben werden
		if (this.linie == null) {
			return "EMPTY";
		} else {
			String name = this.linie.getScreenName();
			for (int i = name.length() - 1; i >= 0; i--) {
				if (name.charAt(i) == '-') {
					name = name.substring(i + 1, name.length()).trim();
					break;
				}
			}
			return name;
		}
	}

	/**
	 * Gibt die Linie dieser Fahrt zurück.
	 * 
	 * @return Linie der Fahrt
	 */
	public Linie getLinie() {
		return this.linie;
	}

	/**
	 * Gibt die nachfolgende Haltestelle der uebergebenen Haltestelle zurueck
	 * 
	 * @param Haltestelle
	 *            Objekt fuer das die naechste Haltestelle bestimmt werden soll
	 * @return Die naechste Haltestelle oder
	 *         Global.ERR_OBJECT_HALTESTELLE_IST_ENDHALTESTELLE oder
	 *         Global.ERR_OBJECT_KEINE_HALTESTELLE_DIESER_LINIE;
	 */
	public Haltestelle getNextHaltestelle(Haltestelle haltestelle) {
		for (int i = 0; i < this.haltestellen.length; i++) {
			if (this.haltestellen[i] == haltestelle) {
				if (i + 1 < this.haltestellen.length) {
					return this.haltestellen[i + 1];
				} else {
					// Sonderbehandlung für ersteHaltestelle == letzteHaltestelle (Kreis)
					// nicht nötig, da von vorne gesucht wird
					
					return Haltestelle.ERR_HALTESTELLE_IST_ENDHALTESTELLE;
				}
			}
		}
		return Haltestelle.ERR_KEINE_HALTESTELLE_DIESER_LINIE;
	}

	/**
	 * getNextHaltestelleREVERSE Gibt die nachfolgende Haltestelle der
	 * uebergebenen Haltestelle zurueck
	 * 
	 * @param Haltestelle
	 *            Objekt fuer das die naechste Haltestelle bestimmt werden soll
	 * @return Die naechste Haltestelle oder
	 *         Global.ERR_OBJECT_HALTESTELLE_IST_ENDHALTESTELLE; oder
	 *         Global.ERR_OBJECT_KEINE_HALTESTELLE_DIESER_LINIE;
	 */
	public Haltestelle getNextHaltestelleReverse(Haltestelle haltestelle) {
		for (int i = this.haltestellen.length - 1; i >= 0; i--) {
			if (this.haltestellen[i] == haltestelle) {
				if (i - 1 >= 0) {
					return this.haltestellen[i - 1];
				} else {
					// Sonderbehandlung für ersteHaltestelle == letzteHaltestelle (Kreis)
					// nicht nötig, da von hinten gesucht wird
					return Haltestelle.ERR_HALTESTELLE_IST_ENDHALTESTELLE;
				}
			}
		}
		return Haltestelle.ERR_KEINE_HALTESTELLE_DIESER_LINIE;
	}

	/**
	 * getTimeToNextHaltestelle
	 * 
	 * @param Haltestelle
	 *            von der die Fahzeit zur nächsten Haltestelle bestimmt werden
	 *            soll.
	 * @return int Differenzeit in min
	 * 0 falls die Haltestelle nicht auf der Fahrt liegt.
	 */
	public int getTimeToNextHaltestelle(Haltestelle haltestelleVon) {
		int pos = this.fahrzeiten.lookupPosition(haltestelleVon.getId());
		if (pos == DynamicIntMap.KEY_NOT_FOUND) {
			Logger.error(this, "Haltestelle in getTimeToNextHaltestelle nicht gefunden");
			return 0;
		} else {
			return this.fahrzeiten.get(pos);
		}
	}


	/**
	 * getTimeAtNextHaltestelle Gibt die Zeit zurueck wann die Fahrt, von einer
	 * bestimmten Zeit ab, an der nächsten Haltestelle auf der Fahrt ankommt.
	 * Die zuletzt benutzte Linie muss übergeben werden damit er evtl.
	 * Umsteigezeiten berücksichtigen kann.
	 * <p>
	 * Achtung: Die Funktion gibt -1 zurück wenn es keine Möglichkeit gibt mit dieser Fahrt zur nächsten
	 * Haltestelle zu fahren. 
	 */
	public int getTimeAtNextHaltestelle(Haltestelle act, int zeit, Linie lastUsedLinie) {
		int abfahrtszeit = this.getAbfahrtAfterTime(act, zeit + IFahrtStatic.getUmsteigezeit(lastUsedLinie, this.getLinie(), act));

		// Falls keine Zeit gefunden wurde, wenn eine Fahrt faehrt diese
		// abbrechen.
		if (abfahrtszeit == -1) {
			return -1;
		}

		// Fahrzeit addieren und ausgeben
		return abfahrtszeit + this.getTimeToNextHaltestelle(act);
	}

	/**
	 * getTimeAtNextHaltestelleREVERSE Gibt die Zeit zurueck wann die Fahrt, von
	 * einer bestimmten Zeit ab, an der nächsten Haltestelle auf der Fahrt
	 * ankommt. Die zuletzt benutzte Linie muss übergeben werden damit er evtl.
	 * Umsteigezeiten berücksichtigen kann.
	 */
	public int getTimeAtNextHaltestelleReverse(Haltestelle act, int zeit, Linie lastUsedLinie) {
		int abfahrtszeit = this.getAnkunftBeforeTime(act, zeit - IFahrtStatic.getUmsteigezeit(lastUsedLinie, this.getLinie(), act));

		// Falls keine Zeit gefunden wurde, wenn eine Fahrt faehrt diese
		// abbrechen.
		if (abfahrtszeit == -1) {
			return -1;
		}

		// Bestimmt die Fahrzeit zur vorherigen Haltestelle, indem es diese
		// bestimmt
		// und dann "vorwärts" erfragt was die Fahrzeit wäre.
		Haltestelle reverse = this.getNextHaltestelleReverse(act);
		if (reverse == Haltestelle.ERR_HALTESTELLE_IST_ENDHALTESTELLE) {
			Logger.error(this, "In Fahrt.getTimeAtNextHaltestelleReverse() called for last Haltestelle on "
				+ "this Fahrt, this leads to wrong output!");
		}
		int result = abfahrtszeit - this.getTimeToNextHaltestelle(reverse);
		return result;
	}

	/**
	 * Gibt die Zeit zurück, die von der ersten bis zu dieser Haltestelle
	 * benötigt wird. Also den Offset den man auf die Abfahrtszeiten in der
	 * ersten Haltestelle addieren muss um die Abfahrtszeit in der gewünschten
	 * Haltestelle zu erhalten.
	 * 
	 * @return int Zeit in min oder -1 (keine Haltestelle dieser Linie)
	 */
	public int getTimeOffsetFromFirstHaltestelle(Haltestelle haltestelle) {
		int pos = this.fahrzeitenKumuliert.lookupPosition(haltestelle.getId());
		if (pos != DynamicIntMap.KEY_NOT_FOUND) {
			return this.fahrzeitenKumuliert.get(pos);
		} else {
			Logger.error(this, "Keine Haltestelle dieser Linie!");
			return -1;
		}
	}

	/**
	 * Gibt die zum übergenen Tag passenden Abfahrtszeiten zurück.
	 * 
	 * @param Tag
	 *            Konstante aus Global
	 * @return Zeit[] Abfahrtszeiten
	 */
	public short[] getAllAbfahrtszeitenByDay(int Tag) {
		switch (Tag) {
		case Global.TAG_WOCHENTAG_SCHULE:
			return abfahrtszeitenWochentagSchule;
		case Global.TAG_WOCHENTAG_FERIEN:
			return abfahrtszeitenWochentagFerien;
		case Global.TAG_SAMSTAG:
			return abfahrtszeitenSamstag;
		case Global.TAG_SONN_FEIERTAG:
			return abfahrtszeitenSonnFeiertag;
		}
		System.err.println("Fehler in Fahrt.getAllAbfahrtszeitenByDay(): Der übergebe Tag-Konstante ist ungültig.");
		return null;
	}

	/**
	 * Gibt zurück, wann die nächste Bahn der Fahrt, NACH der gegeben Zeit, an
	 * der gegeben Haltestelle los fährt.
	 * 
	 * @param Haltestelle
	 *            von der es losgehen soll.
	 * @param Zeit
	 *            nach der die Bahn an der Haltestelle fahren soll.
	 * @return Zeit wann die naechste Bahn der Fahrt an der Haltestelle faehrt.
	 */
	public int getAbfahrtAfterTime(Haltestelle haltestelle, int datumZeit) {

		// Bestimmt den offset den die Haltestelle von der ersten Haltestelle
		// entfernt ist
		int offsetToFirstHaltestelle = this.getTimeOffsetFromFirstHaltestelle(haltestelle);
		
		// Überprüfung, ob die Endhaltestelle gewählt wurde
		if (haltestelle == haltestellen[haltestellen.length - 1]) {
			return -1;
		}

		int dayOffsetCurrent = 0;

		// Versucht ein Ergebnis zu finden.
		// Aber maximal bis eine Woche im voraus, also das Fahrten die einmal
		// die Woche sind noch funktionieren.
		while (dayOffsetCurrent < 8) {
			// Nur beim erstem mal machen (dayOffsetCurrent == 0) alle anderen
			// Tage werden auch so abgeklappert.
			int last = -1;
			if (dayOffsetCurrent == 0) {
				last = this.getAbfahrtAfterTime_Helper(datumZeit, offsetToFirstHaltestelle, dayOffsetCurrent - 1);
			}
			int current = this.getAbfahrtAfterTime_Helper(datumZeit, offsetToFirstHaltestelle, dayOffsetCurrent);

			// Sonderfall: Es kann vorkommen, dass eine Zeit mit dem Plan von
			// current kleiner ist als eine aus dem Plan von last.
			if (last != -1 && current != -1) {
				if (current < last) {
					return current;
				} else {
					return last;
				}
			}

			// Wenn Ergebnisse nicht paarweise vorhanden, einzeln ausgeben.
			if (last != -1) {
				return last;
			}
			if (current != -1) {
				return current;
			}

			// Nichts gefunden, also naechstet Tag.
			dayOffsetCurrent++;
		}

		// Falls nichts gefunden wurde -1
		return -1;
	}

	/**
	 * Helfer fuer Fahrt.getAbfahrtAfterTime (Haltestelle haltestelle, int
	 * datumZeit) Berechnet die erste Abfahrt, an einem bestimmten Tag, nach
	 * einer bestimmten Zeit. Gibt -1 zurueck, falls nichts gefunden wurde.
	 * 
	 * @param datumZeit
	 * @param offsetToFirstHaltestelle
	 * @param dayOffset
	 * @return
	 */
	private int getAbfahrtAfterTime_Helper(int datumZeit, int offsetToFirstHaltestelle, int dayOffset) {
		int currentDay = DatumZeit.getDay(datumZeit) + dayOffset * DatumZeit.DAY_IN_MIN;
		short[] abfahrtszeiten = getAllAbfahrtszeitenByDay(DatumZeit.getDayType(currentDay));
		int currentDayPlusOffset = currentDay + offsetToFirstHaltestelle;

		for (int i = 0; i < abfahrtszeiten.length; i++) {

			// int abfahrt = currentDay + abfahrtszeiten[i] +
			// offsetToFirstHaltestelle;
			int abfahrt = currentDayPlusOffset + abfahrtszeiten[i];
			if (abfahrt >= datumZeit) {
				return abfahrt;
			}
		}
		return -1;
	}

	/**
	 * Gibt zurück, wann die nächste Bahn der Fahrt, VOR der gegeben Zeit (ODER
	 * zum gleichen Zeitpunkt), an der gegeben Haltestelle ankommt.
	 * 
	 * @param Haltestelle
	 *            von der es losgehen soll.
	 * @param Zeit
	 *            vor der die Bahn an der Haltestelle ankommen soll.
	 * @return Zeit wann die naechste Bahn der Fahrt an der Haltestelle faehrt.
	 */
	private int getAnkunftBeforeTime(Haltestelle haltestelle, int datumZeit) {

		// Bestimmt den offset den die Haltestelle von der letzten Haltestelle
		// entfernt ist
		int offsetToFirstHaltestelle = this.getTimeOffsetFromFirstHaltestelle(haltestelle);
		
		// Überprüfung, ob die Anfangshaltestelle gewählt wurde
		if (haltestelle == haltestellen[0]) {
			return -1;
		}
		
		int dayOffsetCurrent = 0;

		// Versucht ein Ergebnis zu finden.
		// Aber maximal bis eine Woche im voraus, also das Fahrten die einmal
		// die Woche sind noch funktionieren.
		while (dayOffsetCurrent > -8) {
			// Nur beim erstem mal machen (dayOffsetCurrent == 0) alle anderen
			// Tage werden auch so abgeklappert.
			int last = -1;
			if (dayOffsetCurrent == 0) {
				last = this.getAnkunftBeforeTime_Helper(datumZeit, offsetToFirstHaltestelle, dayOffsetCurrent + 1);
			}
			int current = this.getAnkunftBeforeTime_Helper(datumZeit, offsetToFirstHaltestelle, dayOffsetCurrent);

			// Sonderfall: Es kann vorkommen, dass eine Zeit mit dem Plan von
			// current kleiner ist als eine aus dem Plan von last.
			if (last != -1 && current != -1) {
				if (current < last) {
					return last;
				} else {
					return current;
				}
			}

			// Wenn Ergebnisse nicht paarweise vorhanden, einzeln ausgeben.
			if (last != -1) {
				return last;
			}
			if (current != -1) {
				return current;
			}

			// Nichts gefunden, also letzter Tag.
			dayOffsetCurrent--;
		}

		// Falls nichts gefunden wurde -1
		return -1;
	}

	/**
	 * Ist fast identisch mit der normaler Version, mit dem Unterschied das der
	 * Vergleich mit kleiner anstatt größer ist. Warum? ankunft=abfahrt bei uns,
	 * deshalb suchen wir beim rückwärtsgehen genau die gleiche Zeit, nur das
	 * wir sie nacher abziehen.
	 * 
	 * @param datumZeit
	 * @param offsetToFirstHaltestelle
	 * @param dayOffset
	 * @return
	 */
	private int getAnkunftBeforeTime_Helper(int datumZeit, int offsetToFirstHaltestelle, int dayOffset) {
		// Sonderfall Laufweg: Der fährt immer sofort
		if (this.isLaufweg())
			return datumZeit;

		int currentDay = DatumZeit.getDay(datumZeit) + dayOffset * DatumZeit.DAY_IN_MIN;
		short[] abfahrtszeiten = getAllAbfahrtszeitenByDay(DatumZeit.getDayType(currentDay));
		int currentDayPlusOffset = currentDay + offsetToFirstHaltestelle;

		// Bestimmte die beste Ankunftszeit
		// beste= <VorDatumZeit> & <soNaheWieMöglichAnDatumZeit>
		int best = -1;
		for (int i = 0; i < abfahrtszeiten.length; i++) {
			int abfahrt = currentDayPlusOffset + abfahrtszeiten[i];
			if ((abfahrt > best || best == -1) && abfahrt <= datumZeit) {
				best = abfahrt;
			}
		}

		return best;
	}

	/**
	 * getLastHaltestelle
	 * 
	 * @return Haltestelle - Gibt die letzte Haltestelle zurueck
	 */
	public Haltestelle getLastHaltestelle() {
		return this.haltestellen[this.haltestellen.length - 1];
	}

	/**
	 * getLastHaltestelle
	 * 
	 * @return Haltestelle - Gibt die erste Haltestelle zurueck
	 */
	public Haltestelle getFirstHaltestelle() {
		return this.haltestellen[0];
	}

	public boolean equals(Object obj) {
		// System.err.println(".equals() for Fahrt not implemented.");
		return super.equals(obj);
	}

	public Haltestelle[] getAlleHaltestellen() {
		return haltestellen;
	}
	
	public boolean isLaufweg() {
		return false;
	}

	public int getMemory() {
		int mem = Memory.getObjSize();

		mem += Memory.getArrayBaseSize();
		mem += Memory.getShortSize() * abfahrtszeitenWochentagSchule.length;

		mem += Memory.getArrayBaseSize();
		if (abfahrtszeitenWochentagSchule != abfahrtszeitenWochentagFerien) {
			mem += Memory.getShortSize() * abfahrtszeitenWochentagFerien.length;
		}

		mem += Memory.getArrayBaseSize();
		mem += Memory.getShortSize() * abfahrtszeitenSamstag.length;

		mem += Memory.getArrayBaseSize();
		mem += Memory.getShortSize() * abfahrtszeitenSonnFeiertag.length;

		mem += fahrzeiten.getMemory();

		mem += fahrzeitenKumuliert.getMemory();

		mem += Memory.getObjSize();

		mem += Memory.getArrayBaseSize() + Memory.getObjSize() * haltestellen.length;

		
		return mem;
	}

	public String getLinienInfo() {
		return getLinie().getScreenName();
	}
	
	public String toString() {
		return "Fahrt " + Data.getIdOfFahrt(this) + " " + this.getLinie();
	}
}
