// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package model;

import dev.Memory;

public class Haltestelle extends AbstractLocation {
	
	// NULL-Werte
	public static final String NULL_TARIFZONE = "";
	public static final String NULL_UNIQUEID = "null-Id";
	public static final Haltestelle NULL = new Haltestelle(VERKEHRSMITTEL_KEINE, "NULL-HALTESTELLE", null, NULL_UNIQUEID, NULL_TARIFZONE);
	
	// Error Code: Error Haltestellen die bei Fehlern uebergeben werden
	public static final Haltestelle ERR_HALTESTELLE_IST_ENDHALTESTELLE = new Haltestelle(VERKEHRSMITTEL_KEINE, "Die Haltestelle ist die End-Haltestelle", null, NULL_UNIQUEID, NULL_TARIFZONE);
	public static final Haltestelle ERR_KEINE_HALTESTELLE_DIESER_LINIE = new Haltestelle(VERKEHRSMITTEL_KEINE, "Haltestelle nicht auf der Fahrt", null, NULL_UNIQUEID, NULL_TARIFZONE);
	

	//Name der Haltestelle
	private final String name;

	//Name des Ortes
	private final Ort ort;

	//Referenzen auf die Fahrt Objete die durch diese Haltestelle fahren.
	private IFahrt[] fahrten = null;

	//ID der Haltestelle, damit man das Objekt schnell dem Index in der Liste zuordnen kann.
	private int id = -1;

	//ID die eindeutig unanbhängig vom verwendeted Datensatz für die Region ist
	private final String uniqueId;

	//TarifzoneZone
	private final String tarifZone;



	/**
	 * Konstruktor mit Name
	 * 
	 * @param s -
	 *            Maskierungs Variable des Anschluss-Typs
	 * @param n
	 * @param o
	 */
	public Haltestelle(byte verkehrsmittelMask, String n, Ort o, String uniqueId, String tarifZone) {
		super(verkehrsmittelMask);
		this.name = n;
		this.ort = o;
		if (o != null) {
			o.addHaltestelle(this);
		}
		this.uniqueId = uniqueId;
		this.tarifZone = tarifZone;
	}

	/**
	 * Gibt den Name kompletten Namen der Haltestelle zurueck (Ort + Haltestelle).
	 * 
	 * @return name als String
	 */
	public String getFullName() {
		if (ort == null) {
			return this.name;
		} else {
			return ort.getScreenName() + " " + this.name;
		}
	}

	/**
	 * Gibt den Name verkürzten Namen der Haltestelle zurueck (kurzOrt + Haltestelle).
	 * 
	 * @return name als String
	 */
	public String getFullShortName() {
		if (ort == null) {
			return this.name;
		} else {
			return ort.getShortName() + " " + this.name;
		}
	}

	/**
	 * Gibt den Ort in dem die Haltestelle liegt zurück
	 * 
	 * @return Ort
	 */
	public Ort getOrt() {
		return this.ort;
	}

	/**
	 * Setzt das Array mit den Fahrten an der Haltestelle Linie[] Array mit allen Fahrten der
	 * Haltestelle
	 */
	public void setFahrten(IFahrt[] f) {
		this.fahrten = f;
	}

	/**
	 * Gibt ein Array mit allen Fahrten zurueck
	 * 
	 * @return Array der Fahrten die an dieser Haltestelle fahren Achtung! Gibt null zurück, falls
	 *         keine Fahrten gesetzt wurden.
	 */
	public IFahrt[] getAlleFahrten() {
		return this.fahrten;
	}

	/**
	 * Gibt die Listen/Posisitons ID der Haltestelle zurück
	 * 
	 * @return int id der Haltestelle
	 */
	public int getId() {
		return this.id;
	}
	
	public String getUniqueId() {
		return this.uniqueId;
	}

	/**
	 * Setzt die Listen/Posisitons ID der Haltestelle zurück
	 * 
	 * @param id
	 *            int
	 */
	public void setId(int id) {
		this.id = id;
	}

	public int getSize() {
		int size = 0;
		size += 4; // id
		size += name.length() + 1; // name
		size += fahrten.length + 1; // fahrten
		return size;
	}

	/**
	 * HashCode eines Objekts vom Typ Haltestelle
	 * 
	 * @return HashCode = ID
	 */
	public int hashCode() {
		return this.id;
	}

	/**
	 * Zwei Objekte vom Typ Haltestelle sind gleich, wenn ihre IDs uebereinstimmen
	 * 
	 * @return Uebereinstimmung zweier Objekte vom Typ Haltestelle
	 */
	public boolean equals(Object obj) {
		if (obj instanceof Haltestelle) {
			return this.id == ((Haltestelle) obj).id;
		}
		return super.equals(obj);
	}
	
	/**
	 * Gibt den Name Namen (ohne Ort) der Haltestelle zurueck.
	 * 
	 * @return name als String
	 */
	public String getScreenName() {
		return this.name;
	}
	
	public int getMemory() {
		int mem = Memory.getShortSize();

		// AbstractLocation
		mem += super.getMemory();
				
		// name
		mem += Memory.getStringBaseSize() + name.length() * Memory.getStringCharakterSize();
		
		// ort
		mem += Memory.getObjSize();
		
		// fahrten
		mem += Memory.getArrayBaseSize() + fahrten.length * Memory.getObjSize();
		
		// id
		mem += Memory.getIntSize();

		// uniqueId
		mem += Memory.getStringBaseSize() + uniqueId.length() * Memory.getStringCharakterSize();
		
		// tarifZone
		mem += Memory.getStringBaseSize() + tarifZone.length() * Memory.getStringCharakterSize();
		
		return mem;
	}
		
	public String getTarifZone() {
		return this.tarifZone;
	}
	
	public String toString() {
		return this.getFullName() + " (Id=" + this.getId() + " UId=" + this.getUniqueId() + ")";
	}
}
