// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package model;

import java.util.Enumeration;
import java.util.Vector;

import algo.Algorithmus;
import algo.DataWrapperNormal;
import algo.IDataWrapper;
import control.Global;
import dev.Logger;

/**
 * Route speichert das Ergebnis des Algorithmus
 */
public class Route {
	// Routen Status
	public static final int STATUS_NULL_ROUTE = -1;
	public static final int STATUS_KEIN_FEHLER = 0;
	public static final int STATUS_KEINE_ROUTE_GEFUNDEN = 1;
	public static final int STATUS_GLEICHER_START_UND_ZIELORT = 2;
	
	// Routen Optimierungs Status
	public static final int OPTIMIERUNG_NOCHT_NICHT_OPTIMIERT = 0;
	public static final int OPTIMIERUNG_ABGEBROCHEN = 1;
	public static final int OPTIMIERUNG_ABGESCHLOSSEN = 2;
	
	// NULL-Werte
	public static final Route NULL = new Route(STATUS_NULL_ROUTE);
	
	// Routen Abschnitte der finalen Route in der richtigen Reihenfolge
	private final RoutenAbschnitt[] routenAbschnitte; 
	private RoutenAbschnitt[] routenAbschnitteKompakt = null;
	private int zeitDerAnforderung; // (NICHT die erste Abfahrtszeit!)
	private int anzUmsteigen = -1;
	private int error_code = Route.STATUS_KEIN_FEHLER;
	private IDataWrapper iDataWrapper = null;
	
	private short optimierungsTyp = Algorithmus.OPTIMIERUNG_KEINE;
	private int optimierungsStatus = Route.OPTIMIERUNG_NOCHT_NICHT_OPTIMIERT;

	public Route(RoutenAbschnitt[] routenAbschnitte, int zeitDerAnforderung, IDataWrapper wrapper) {
		this.routenAbschnitte = routenAbschnitte;
		this.zeitDerAnforderung = zeitDerAnforderung;
		this.iDataWrapper = wrapper;
	}

	public Route(int errorCode) {
		this.routenAbschnitte = null;
		this.error_code = errorCode;
		this.zeitDerAnforderung = DatumZeit.NULL;
	}

	/**
	 * Gruppiert alle RoutenAbschnitte auf der gleichen Linie zu einem RoutenAbschnitt.<br>
	 * Daten werden gechacht.
	 */
	public RoutenAbschnitt[] getKompakteRoutenAbschnitte() {
		// Leer bleibt leer.
		if (this.routenAbschnitte == null || this.routenAbschnitte.length == 0) {
			return new RoutenAbschnitt[0];
		}

		// Global can overwrite this for debugging pourposes.
		if (!Global.USE_KOMPAKTE_ROUTENABSCHNITTE) {
			return this.routenAbschnitte;
		}

		// Caching
		if (this.routenAbschnitteKompakt != null) {
			return this.routenAbschnitteKompakt;
		}

		// Let's rock
		Vector tmpRoutenAbschnitte = new Vector();
		for (int i = 0; i < this.routenAbschnitte.length; i++) {
			tmpRoutenAbschnitte.addElement(this.routenAbschnitte[i]);
		}

		// zwei aufeinanderfolgende Laufwege zusammenfassen
		for (int i = 1; i < tmpRoutenAbschnitte.size(); i++) {
			RoutenAbschnitt letzterRoutenAbschnitt = (RoutenAbschnitt) tmpRoutenAbschnitte.elementAt(i - 1);
			RoutenAbschnitt currentRoutenAbschnitt = (RoutenAbschnitt) tmpRoutenAbschnitte.elementAt(i);
			// falls beides Laufwege sind
			if (letzterRoutenAbschnitt.isLaufweg() && currentRoutenAbschnitt.isLaufweg()) {
				
				byte baseLaufZeit = (byte) (((Laufweg) letzterRoutenAbschnitt.getFahrt()).getBaseLaufzeit() + ((Laufweg) currentRoutenAbschnitt.getFahrt()).getBaseLaufzeit());

				Haltestelle von = letzterRoutenAbschnitt.getVon();
				Haltestelle nach = currentRoutenAbschnitt.getNach();
				
				Laufweg laufweg = new Laufweg(von, nach, baseLaufZeit);
				
				// den vorherigen Abschnitt löschen und zusammengefasst einfügen
				tmpRoutenAbschnitte.setElementAt(new RoutenAbschnitt(von, nach, letzterRoutenAbschnitt.getAbfahrtsZeit(), currentRoutenAbschnitt.getAnkunftsZeit(), laufweg), i-1);
				tmpRoutenAbschnitte.removeElementAt(i);
				i--;
			}
		}
		
		// Zusammenfassen von Fahrten auf gleicher Linie
		
		Vector kompakt = new Vector();
		RoutenAbschnitt letzterRoutenAbschnitt = (RoutenAbschnitt)tmpRoutenAbschnitte.elementAt(0);
		RoutenAbschnitt currentRoutenAbschnitt = null;
		int letzteAbfahrtsZeit = letzterRoutenAbschnitt.getAbfahrtsZeit();
		IFahrt letzteFahrt = letzterRoutenAbschnitt.getFahrt();
		Haltestelle letzteHaltestelle = letzterRoutenAbschnitt.getVon();
		
		for (int iRoutenAbschnitt = 1; iRoutenAbschnitt < tmpRoutenAbschnitte.size(); iRoutenAbschnitt++) {
			letzterRoutenAbschnitt = (RoutenAbschnitt) tmpRoutenAbschnitte.elementAt(iRoutenAbschnitt - 1);
			currentRoutenAbschnitt = (RoutenAbschnitt) tmpRoutenAbschnitte.elementAt(iRoutenAbschnitt);

			int warteZeit = currentRoutenAbschnitt.getAbfahrtsZeit() - letzterRoutenAbschnitt.getAnkunftsZeit();
			
			boolean split = false;
			
			// prüfe ob lange gewartet wird
			if (warteZeit >= 15) {
				split = true;
			} else {
				// prüfe ob umgestiegen wird 
				if (currentRoutenAbschnitt.getFahrt().getLinie() != letzteFahrt.getLinie()) {
					split = true;
				} else {
					// prüfe, ob es eine Fahrt mit der gleichen Linie vorher gibt
					Haltestelle haltestelleVon = currentRoutenAbschnitt.getVon();
					if (currentRoutenAbschnitt.getAbfahrtsZeit() != letzterRoutenAbschnitt.getAnkunftsZeit()) {
						IFahrt[] alleFahrten = haltestelleVon.getAlleFahrten();
						Linie letzteLinie = letzteFahrt.getLinie();
						int abfahrtszeit = currentRoutenAbschnitt.getAbfahrtsZeit(); 
						int letzteAnkunftszeit = letzterRoutenAbschnitt.getAnkunftsZeit();
						for (int i = 0; i < alleFahrten.length; i++) {
							IFahrt fahrt = alleFahrten[i];
							if (fahrt.getLinie() == letzteLinie) {
								int fahrtAbfahrt = fahrt.getAbfahrtAfterTime(haltestelleVon, letzteAnkunftszeit); 
								if (fahrtAbfahrt > 0 && fahrtAbfahrt < abfahrtszeit) {
									// falls das Fahrmittel eine kleine Schleife fährt es nicht anzeigen
									if (isGeschlosseneFahrt(fahrt.getLinie(), haltestelleVon, letzteAnkunftszeit, abfahrtszeit)) {
										continue;
									}
									split = true;
									break;
								}
							}
						}
					}
				}
			}
			
			if (split) {
				// Erstelle neuen kompakten RoutenAbschnitt
				kompakt.addElement(new RoutenAbschnitt(letzteHaltestelle, letzterRoutenAbschnitt.getNach(),
					letzteAbfahrtsZeit, letzterRoutenAbschnitt.getAnkunftsZeit(), letzterRoutenAbschnitt.getFahrt()));
				// Setzte Schleppointer
				letzteAbfahrtsZeit = currentRoutenAbschnitt.getAbfahrtsZeit();
				letzteFahrt = currentRoutenAbschnitt.getFahrt();
				letzteHaltestelle = currentRoutenAbschnitt.getVon();
				split = false;
			}
		}

		// Füge den letzten Eintrag ein, der bisher nicht eingefügt wurde, weil
		// das Array mit den RoutenAbschnitten geendet hat, ohne umsteigen.
		letzterRoutenAbschnitt = (RoutenAbschnitt) tmpRoutenAbschnitte.elementAt(tmpRoutenAbschnitte.size() - 1);
		kompakt.addElement(new RoutenAbschnitt(letzteHaltestelle, letzterRoutenAbschnitt.getNach(),
			letzteAbfahrtsZeit, letzterRoutenAbschnitt.getAnkunftsZeit(), letzterRoutenAbschnitt.getFahrt()));

		// In Array umwandeln
		RoutenAbschnitt[] kompaktAsArray = new RoutenAbschnitt[kompakt.size()];
		int pos = 0;
		for (Enumeration e = kompakt.elements(); e.hasMoreElements();) {
			kompaktAsArray[pos] = (RoutenAbschnitt) e.nextElement();
			pos++;
		}

		// Caching
		this.routenAbschnitteKompakt = kompaktAsArray;
		return kompaktAsArray;
	}

	/**
	 * prüft, ob mit den Fahrten der Linie die halltestelleLoop eine Schleife bildet
	 * funktioniert allerdings nur, falls die Ankunftszeit auch immer die Abfahrtszeit ist bei den Teilschritten
	 */
	private boolean isGeschlosseneFahrt(Linie linie, Haltestelle haltestelleLoop, int abfahrtszeit, int ankunftszeit) {
		// code sehr ähnlich zu getRoutenAbschnitteWithLoops
		IFahrt[] alleFahrten = haltestelleLoop.getAlleFahrten();
		Haltestelle currentHaltestelle = haltestelleLoop;
		int currentZeit = abfahrtszeit;
		boolean valid = true;
		int cnt = 0;
		while (currentZeit <= ankunftszeit && valid) {
			cnt++;
			if (cnt == 1000) {
				Logger.error(this, "Endlosschleife isGeschlosseneFahrt");
				break;
			}
			valid = false;
			for (int i = 0; i < alleFahrten.length; i++) {
				IFahrt fahrt = alleFahrten[i];
				if (fahrt.getLinie() == linie) {
					int currentAbfahrt = fahrt.getAbfahrtAfterTime(currentHaltestelle, currentZeit);
					if (currentZeit == currentAbfahrt) {
						valid = true;
						currentZeit = fahrt.getTimeAtNextHaltestelle(currentHaltestelle, currentZeit, linie);
						currentHaltestelle = fahrt.getNextHaltestelle(currentHaltestelle);
						alleFahrten = currentHaltestelle.getAlleFahrten();
						if (currentHaltestelle == haltestelleLoop) {
							if (currentZeit <= ankunftszeit) {
								return true;
							} else {
								return false;
							}
						}
						break;
					}
				}
			}
		}
		return false;
	}

	public RoutenAbschnitt[] getRoutenAbschnitte() {
		return this.routenAbschnitte;
	}

	public RoutenAbschnitt[] getRoutenAbschnitteWithLoops() {
		// code sehr ähnlich zu isGeschlosseneFahrt
		Vector tmpRoutenAbschnitte = new Vector();
		for (int i = 0; i < this.routenAbschnitte.length; i++) {
			tmpRoutenAbschnitte.addElement(this.routenAbschnitte[i]);
		}

		RoutenAbschnitt letzterRoutenAbschnitt = (RoutenAbschnitt)tmpRoutenAbschnitte.elementAt(0);
		RoutenAbschnitt currentRoutenAbschnitt = null;
		IFahrt letzteFahrt = letzterRoutenAbschnitt.getFahrt();
		
		for (int iRoutenAbschnitt = 1; iRoutenAbschnitt < tmpRoutenAbschnitte.size(); iRoutenAbschnitt++) {
			letzterRoutenAbschnitt = (RoutenAbschnitt) tmpRoutenAbschnitte.elementAt(iRoutenAbschnitt - 1);
			currentRoutenAbschnitt = (RoutenAbschnitt) tmpRoutenAbschnitte.elementAt(iRoutenAbschnitt);

			if (currentRoutenAbschnitt.getFahrt().getLinie() == letzteFahrt.getLinie()) {
				// prüfe, ob es eine Fahrt mit der gleichen Linie vorher gibt
				Haltestelle haltestelleVon = currentRoutenAbschnitt.getVon();
				if (currentRoutenAbschnitt.getAbfahrtsZeit() != letzterRoutenAbschnitt.getAnkunftsZeit()) {
					IFahrt[] alleFahrten = haltestelleVon.getAlleFahrten();
					Linie letzteLinie = letzteFahrt.getLinie();
					int abfahrtszeit = currentRoutenAbschnitt.getAbfahrtsZeit(); 
					int letzteAnkunftszeit = letzterRoutenAbschnitt.getAnkunftsZeit();
					for (int i = 0; i < alleFahrten.length; i++) {
						IFahrt fahrt = alleFahrten[i];
						Linie linie = fahrt.getLinie();
						if (linie == letzteLinie) {
							int fahrtAbfahrt = fahrt.getAbfahrtAfterTime(haltestelleVon, letzteAnkunftszeit); 
							if (fahrtAbfahrt > 0 && fahrtAbfahrt < abfahrtszeit) {
								// falls das Fahrmittel eine kleine Schleife fährt es nicht anzeigen
								if (isGeschlosseneFahrt(linie, haltestelleVon, letzteAnkunftszeit, abfahrtszeit)) {

									Haltestelle currentHaltestelle = haltestelleVon;
									int currentZeit = letzteAnkunftszeit;
									IFahrt[] alleLoopFahrten = haltestelleVon.getAlleFahrten();

									int cnt = 0;
									boolean valid = false;
									int inserted = 0;
									while (!valid) {
										cnt++;
										if (cnt == 1000) {
											Logger.error(this, "Endlosschleife getRoutenAbschnitteWithLoops");
											break;
										}
										for (int iLoopFahrt = 0; iLoopFahrt < alleLoopFahrten.length; iLoopFahrt++) {
											IFahrt loopFahrt = alleLoopFahrten[iLoopFahrt];
											if (loopFahrt.getLinie() == linie) {
												int currentAbfahrt = loopFahrt.getAbfahrtAfterTime(currentHaltestelle, currentZeit);
												if (currentZeit == currentAbfahrt) {
													int nextZeit = loopFahrt.getTimeAtNextHaltestelle(currentHaltestelle, currentZeit, fahrt.getLinie());
													Haltestelle nextHaltestelle = loopFahrt.getNextHaltestelle(currentHaltestelle);
	
													RoutenAbschnitt ra = new RoutenAbschnitt(currentHaltestelle, nextHaltestelle, currentAbfahrt, nextZeit, loopFahrt);
													tmpRoutenAbschnitte.insertElementAt(ra, iRoutenAbschnitt + (inserted++));
	
													currentZeit = nextZeit;
													currentHaltestelle = nextHaltestelle;
													alleLoopFahrten = currentHaltestelle.getAlleFahrten();
													if (currentHaltestelle == haltestelleVon) {
														valid = true;
													}
													break;
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
		// In Array umwandeln
		RoutenAbschnitt[] abschnitteAsArray = new RoutenAbschnitt[tmpRoutenAbschnitte.size()];
		int pos = 0;
		for (Enumeration e = tmpRoutenAbschnitte.elements(); e.hasMoreElements();) {
			abschnitteAsArray[pos] = (RoutenAbschnitt) e.nextElement();
			pos++;
		}
		return abschnitteAsArray;
	}

	/**
	 * true wenn kein Fehler aufgetreten ist
	 * 
	 * @return
	 */
	public boolean isValid() {
		return this.getErrorCode() == Route.STATUS_KEIN_FEHLER;
	}

	/**
	 * Gibt zurueck wie oft man auf der Route umsteigt.
	 * 
	 * @return int
	 */
	public int getAnzUmsteigen() {
		if (anzUmsteigen == -1) {
			anzUmsteigen = -1;
			RoutenAbschnitt[] abschnitte = this.getKompakteRoutenAbschnitte();
			for (int i = 0; i < abschnitte.length; i++) {
				if (!(abschnitte[i].isLaufweg())) {
					anzUmsteigen++;
				}
			}
			// Bei reinen Laufwegen, anzUmsteigen 0 und nicht -1
			if (anzUmsteigen == -1) {
				anzUmsteigen = 0;
			}
		}
		return anzUmsteigen;
	}

	/**
	 * getZeitDerAnforderung Achtung: Die Zeit die zur Routenberechnung übergeben wurde. NICHT die Zeit an der
	 * losgefahren wird!
	 * 
	 * @return Zeit
	 */
	public int getZeitDerAnforderung() {
		return this.zeitDerAnforderung;
	}

	/**
	 * Gibt die Von Haltestelle von der Route zurück
	 * 
	 * @return Haltestelle Von
	 */
	public Haltestelle getHaltestelleVon() {
		if (this.routenAbschnitte != null && this.routenAbschnitte.length != 0) {
			return this.routenAbschnitte[0].getVon();
		} else {
			return Haltestelle.NULL;
		}
	}

	/**
	 * Gibt die Nach Haltestelle von der Route zurück
	 * 
	 * @return Haltestelle Nach
	 */
	public Haltestelle getHaltestelleNach() {
		if (this.routenAbschnitte != null && this.routenAbschnitte.length != 0) {
			return this.routenAbschnitte[this.routenAbschnitte.length - 1].getNach();
		} else {
			return Haltestelle.NULL;
		}
	}

	/**
	 * Liefert das Haltestellenpaar der Start- und ZielHaltestelle
	 * 
	 * @return Start- und Zielhaltestelle
	 */
	public HaltestellenPaar getHaltestellenPaar() {
		return new HaltestellenPaar(getHaltestelleVon(), getHaltestelleNach());
	}

	/**
	 * Gibt die Zeit an der die Route losfährt zurück
	 * 
	 * @return Zeit int
	 */
	public int getAbfahrtszeit() {
		if (this.routenAbschnitte != null && this.routenAbschnitte.length != 0) {
			return this.routenAbschnitte[0].getAbfahrtsZeit();
		} else {
			return DatumZeit.NULL;
		}
	}

	/**
	 * Gibt die Zeit an der die Route an der letzten Haltestelle ankommt zurück
	 * 
	 * @return Zeit int
	 */
	public int getAnkunftszeit() {
		if (this.routenAbschnitte != null && this.routenAbschnitte.length != 0) {
			return this.routenAbschnitte[this.routenAbschnitte.length - 1].getAnkunftsZeit();
		} else {
			return DatumZeit.NULL;
		}
	}

	/**
	 * Gibt true zurueck wenn die Route nur aus Laufwegen besteht. False wenn nicht, oder wenn es keine
	 * routenAbschnitte gibt.
	 * 
	 * @return boolean
	 */
	public boolean nurLaufwege() {
		if (this.routenAbschnitte != null && this.routenAbschnitte.length != 0) {
			for (int i = 0; i < this.routenAbschnitte.length; i++) {
				if (!(this.routenAbschnitte[i].isLaufweg())) {
					return false;
				}
			}
			return true;
		} else {
			return false;
		}
	}

	public void setErrorCode(int c) {
		this.error_code = c;
	}

	public int getErrorCode() {
		return this.error_code;
	}

	public IDataWrapper getDataWrapper() {
		return this.iDataWrapper;
	}

	public boolean isOptimierungAbgebrochen() {
		return optimierungsStatus == OPTIMIERUNG_ABGEBROCHEN;
	}

	/**
	 * Setzt das die Route aufgrund des Zeitkriteriums im RoutenOptimierer abgebrochen wurde.
	 */
	public void setOptimierungAbgebrochen() {
		this.optimierungsStatus = OPTIMIERUNG_ABGEBROCHEN;
	}

	/**
	 * Gibt die Route als anzeigbaren String mit zusammengefassten Haltestellen aus.
	 */
	public String toString() {
		RoutenAbschnitt[] Abschnitte = this.getKompakteRoutenAbschnitte();
		String out = "";
		out += this.getHaltestelleVon().getId() + "-" + this.getHaltestelleVon().getFullName() + " -> "
			+ this.getHaltestelleNach().getId() + "-" + this.getHaltestelleNach().getFullName() + "\n";
		out += DatumZeit.toDateTimeString(this.getAbfahrtszeit()) + " Uhr - "
			+ DatumZeit.toDateTimeString(this.getAnkunftszeit()) + " Uhr\n";
		out += "Fahrzeit: " + (this.getAnkunftszeit() - this.getAbfahrtszeit()) + "\n";
		
		String stringOptimierungsTyp = Algorithmus.getOptimierungsString(optimierungsTyp);
		String stringOptimierungsStatus;
		switch (optimierungsStatus) {
		case OPTIMIERUNG_ABGEBROCHEN:
			stringOptimierungsStatus = "!!! ABGEBROCHEN !!!";
			break;
		case OPTIMIERUNG_NOCHT_NICHT_OPTIMIERT:
			stringOptimierungsStatus = "!!! NOCH NICHT OPTIMIERT !!!";
			break;
		case OPTIMIERUNG_ABGESCHLOSSEN:
			stringOptimierungsStatus = "abgeschlossen";
			break;
			default:
				Logger.error(this.getClass(), "unbekannter OptimierungsStatus");
			stringOptimierungsStatus = " FEHLER ";
		}
		out += "OptimierungsTyp: "+ stringOptimierungsTyp + "   OptimierungsStatus: " + stringOptimierungsStatus + "\n";
		
		out += "DataWrapper = "
			+ ((this.iDataWrapper != null) ? ((this.iDataWrapper instanceof DataWrapperNormal) ? "DataWrapperNormal"
				: "DataWrapperReverse")
				: ("null")) + "\n";
		for (int i = 0; i < Abschnitte.length; i++) {
			RoutenAbschnitt streckenAbschnitt = Abschnitte[i];
			out += streckenAbschnitt.getFahrt() + "\n";
			out += "^ " + DatumZeit.toDateTimeString(streckenAbschnitt.getAbfahrtsZeit()) + " "
				+ streckenAbschnitt.getVon().getFullName() + " ("+streckenAbschnitt.getVon().getUniqueId()+")\n";
			out += "^ " + DatumZeit.toDateTimeString(streckenAbschnitt.getAnkunftsZeit()) + " "
				+ streckenAbschnitt.getNach().getFullName() + " ("+streckenAbschnitt.getNach().getUniqueId()+")\n";
			out += "\n";
		}
		return out;
	}

	public boolean isReinerLaufweg() {
		if (error_code != STATUS_KEIN_FEHLER) {
			Logger.error(this.getClass(), "ausgeführt auf ungültige Route");
			return false;
		}

		for (int i = 0; i < routenAbschnitte.length; i++) {
			if (!routenAbschnitte[i].isLaufweg()) {
				return false;
			}
		}

		return true;
	}

	public void setDataWrapper(IDataWrapper iDataWrapper) {
		this.iDataWrapper = iDataWrapper;
	}

	public void setZeitDerAnforderung(int zeitDerAnforderung) {
		this.zeitDerAnforderung = zeitDerAnforderung;
	}

	/**
	 * 
	 * @param optimierungsTyp
	 *            Algorithmus.OPTIMIERUNG_KEINE <br>
	 *            Algorithmus.OPTIMIERUNG_NORMAL <br>
	 *            Algorithmus.OPTIMIERUNG_LANG <br>
	 *            Algorithmus.OPTIMIERUNG_ENDLOS
	 */
	public void setOptimierungsTyp(short optimierungsTyp) {
		this.optimierungsTyp = optimierungsTyp;
	}

	/**
	 * 
	 * @return
	 *            Algorithmus.OPTIMIERUNG_KEINE <br>
	 *            Algorithmus.OPTIMIERUNG_NORMAL <br>
	 *            Algorithmus.OPTIMIERUNG_LANG <br>
	 *            Algorithmus.OPTIMIERUNG_ENDLOS

	 */
	public short getOptimierungsTyp() {
		return optimierungsTyp;
	}

	/**
	 * 
	 * @return
	 * 		OPTIMIERUNG_NOCHT_NICHT_OPTIMIERT <br>
	 *		OPTIMIERUNG_ABGEBROCHEN <br>
	 *		OPTIMIERUNG_ABGESCHLOSSEN 
	 */
	public int getOptimierungsStatus() {
		return optimierungsStatus;
	}

	public void setOptimierungAbgeschlossen() {
		this.optimierungsStatus = OPTIMIERUNG_ABGESCHLOSSEN;
	}
}
