// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package utils;

import dev.Memory;

/**
 * Bietet eine performante Map, welche alle bekannten Map Funktionen, implementiert und die J2ME HashMap
 * ersetzen soll, weil diese immer Objekte als Eingabe verlangt, was bei reinen Int-Maps unnötiger Overhead
 * ist. Ausserdem erspart man sich das laestige casten der Funktions-Ausgaben.
 * <p>
 * Eine Map sollte verwendet werden wenn die Abstände zwischen den Integer Keys groß sind. Falls es im
 * Allgemeinen forlaufende Keys sind, welche bei 0 anfangen, ist die Verwengung von DynamicIntArray
 * empfehlenswert.
 */
public class DynamicIntMap {
	private int[] keys = null;
	private int[] values = null;
	private int nextFreePosition = 0;
	private int capacity = 10;
	private int capacityStepSize = 10;

	public static final int KEY_NOT_FOUND = -1;

	/**
	 * Bitte waehle eine gute Start Kapazitaet, welche im Normalfall nicht Ueberschritten wird. Falls diese
	 * ueberschritten wird, wird die Kapazitaet der Map jeweils um 10 erhoeht.
	 * 
	 * @param capacity
	 */
	public DynamicIntMap(int capacity) {
		this.capacity = capacity;
		this.clear();
	}

	/**
	 * Delete all entries and recreate internal structure
	 */
	public void clear() {
		keys = new int[capacity];
		values = new int[capacity];
		nextFreePosition = 0;
	}

	/**
	 * Fuegt einen neuen Eintrag in die Map hinzu.
	 * 
	 * @param key
	 * @param value
	 */
	public void add(int key, int value) {
		this.checkSize();
		keys[nextFreePosition] = key;
		values[nextFreePosition] = value;
		nextFreePosition++;
	}

	/**
	 * Versucht einen Eintrag mit einem neuen Value zu ersetzen. Falls der key nicht vorhanden ist, wird mit
	 * add(key, value) ein neuer hinzugefuegt.
	 * 
	 * @param key
	 * @param value
	 */
	public void replace(int key, int value) {
		int position = lookupPosition(key);
		if (position != DynamicIntMap.KEY_NOT_FOUND) {
			values[position] = value;
		} else {
			// Fall des Key noch nicht vorhanden ist, fuege die Daten normal hinzu.
			add(key, value);
		}
	}

	/**
	 * Gibt den value fuer einen Key zurueck Falls ein Key mehrmals hinzugefuegt wurde, wird der erste
	 * zurueckgegeben, falls das nicht gewuenscht ist sollte man die replace() Funktion verwenden.
	 * 
	 * @param Position die mit lookupPosition ermittelt wurde.
	 * @return Falls der Eintrag nicht gefunden wurde, gebe Integer.MIN_VALUE zurueck. Das impliziert das
	 *         dieser nie als Value verwendet werden darf.
	 */
	public int get(int position) {
		return values[position];
	}

	/**
	 * Bestimmt die Position eines Values im Array anhand des Keys.
	 * 
	 * @param key
	 * @return Die Positon des Values. DynamicIntMap.KEY_NOT_FOUND falls nicht gefunden.
	 */
	public int lookupPosition(int key) {
		// Man muss nur bis nextPosition suchen, weil danach ja nichts
		// mehr in der Map drin steht.
		for (int i = 0; i < nextFreePosition; i++) {
			if (keys[i] == key)
				return i;
		}
		return DynamicIntMap.KEY_NOT_FOUND;
	}

	/**
	 * Überprüft die Belegung der Map und vergrößert sich bei Bedarf entsprechend um die capacityStepSize.
	 */
	private void checkSize() {
		if (nextFreePosition == capacity) {
			// Array ist zu klein und muss vergroestert werden
			capacity = capacity + capacityStepSize;
			int[] newKeys = new int[capacity];
			int[] newValues = new int[capacity];

			// Kopiere die Daten in die neuen Arrays
			System.arraycopy(this.keys, 0, newKeys, 0, this.keys.length);
			System.arraycopy(this.values, 0, newValues, 0, this.values.length);

			// Ueberschreibe die alten Arrays
			keys = newKeys;
			values = newValues;
		}
	}

	/**
	 * Gibt an wieviele Einträge in der Map vorhanden sind.
	 */
	public int getSize() {
		return this.nextFreePosition;
	}

	/**
	 * Gibt die Keys der DynamicIntMp zurück.<br>
	 * ACHTUNG: Das Key Set darf nur bis der Position die in getSize() angegeben ist verwendet werden, danach
	 * steht nur noch Müll in dem Array!
	 * 
	 * @return
	 */
	public int[] getKeySet() {
		return this.keys;
	}

	public int getMemory() {
		return 2 * (Memory.getArrayBaseSize() + capacity * Memory.getIntSize()) + 3 * Memory.getIntSize() + Memory.getObjSize();
	}
	
	public String toString() {
		String out = "[";
		for(int i = 0; i<nextFreePosition; i++) {
			out += keys[i] + "=" + values[i] + ",";
		}
		return out + "]";
	}
}
