// This file is part of push&ride.

// Copyright 2006-2011 Tobias Binz, Andreas Moosbrugger, Michael Schlagmüller

// push&ride is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// The application is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with push&ride.  If not, see <http://www.gnu.org/licenses/>.

package utils;

/**
 * Da J2ME keine Collections unterstützt wie Java bietet diese Klasse ein String Array, welches sich bei
 * Bedarf automatisch vergrößert. (...aber nicht wieder verkleinert, da man keine Elemente löschen kann,
 * sondern nur wieder auf NULL_VALUE setzen.) *
 * <p>
 * Das DynamicStringArray kann an Stellen verwendet werden, wo die Keys nahe bei 0 starten und dann
 * einigermaßen fortlaufend sind, damit es keine Lücken in dem Array gibt. So wie bei einem normalen Array,
 * nur mit dem Vorteil, dass das dynamische Array sich bei Bedarf automatisch vergrößert.
 */
public class DynamicStringArray {

	// NULL_VALUE darf nicht null oder "" sein, da beispielsweise System Properties die null oder "" sind
	// jedesmal neu geholt werden und nicht aus dem RecordStore verwendet werden.
	public static String NULL_VALUE = "!N$";

	// Definiert um wie viele Felder das Array vergrößert wird wenn es voll ist.
	private int capacityStepSize;

	// Das Array mit Daten welches von dieser Klasse gekapselt wird.
	private String[] data = null;

	/**
	 * DynamicStringArray mit definierter Step size, also um wieviel das Array mindestens vergrößert werden
	 * soll wenn es voll ist.
	 */
	public DynamicStringArray(int capacityStepSize) {
		this.capacityStepSize = capacityStepSize;
		this.data = createNullValueArray(this.capacityStepSize);
	}

	/**
	 * Mit einem Array vordefiniertes DynamicStringArray.
	 */
	public DynamicStringArray(String[] initialArray, int capacityStepSize) {
		this.data = initialArray;
		this.capacityStepSize = capacityStepSize;
	}

	/**
	 * Liest einen Wert aus dem Array.
	 * 
	 * @param key
	 *            key > 0. Indizierung beginnt wie beim Array bei 0.
	 * @return DynamicStringArray.NULL_VALUE wenn Wert nicht gesetzt oder Wert nicht vorhanden, ansonsten den
	 *         gespeicherten String, der auch null sein kann.
	 */
	public String get(int key) {
		// Falls der key unzulässig oder nicht vorhanden ist NULL_VALUE zurückgeben.
		if (this.data.length <= key || key < 0) {
			return DynamicStringArray.NULL_VALUE;
		}
		// Ansonsten den Wert.
		return this.data[key];
	}

	/**
	 * Setzt den value an die Position key im Array. Falls key vorhanden ist wird er überschrieben. Falls key
	 * neu ist und/oder nicht mehr in das bisherige Array passt, wir das Array (unter berücksichtigung der
	 * capacityStepSize) entsprechend vergrößert.
	 * 
	 * @param key
	 *            > 0
	 * @param value
	 *            Beliebieger String
	 */
	public void set(int key, String value) {
		// Unzulässige Keys ignorieren
		if (key < 0) {
			return;
		}

		// Bei Bedarf vergößern
		if (this.data.length <= key) {
			// Der Key muss auf jeden Fall reinpassen
			// Außerdem wird die Größe mindestens um capacityStepSize vergrößert.
			int newSize = Math.max(this.data.length + this.capacityStepSize, key + 1);
			String[] newArray = createNullValueArray(newSize);
			for (int i = 0; i < this.data.length; i++) {
				newArray[i] = this.data[i];
			}
			this.data = newArray;
		}

		// Value hinzufügen
		this.data[key] = value;
	}

	/**
	 * Gibt das gesamte Daten Array zurück. ACHTUNG: Falls die capacityStepSize != 1 sind, kann es
	 * undefinierte Felder in diesem Array geben, die nicht == NULL_VALUE sind! Dieses Array nicht mit anderen
	 * Arrays aus dieser Funktion vergleichen (Pointervergleich), da das DynamicStringArray beim vergrößern
	 * ein neues Array anlegt, der Vergleich also ab spätestens dann immer false ist.
	 */
	public String[] getArray() {
		return this.data;
	}

	/**
	 * Legt ein neues Array mit der übergebenen Größe an und setzt alle Felder auf NULL_VALUE.
	 * 
	 * @param size
	 *            des neuen Arrays
	 * @return int[]
	 */
	private String[] createNullValueArray(int size) {
		String[] newArray = new String[size];
		for (int i = 0; i < newArray.length; i++) {
			newArray[i] = DynamicStringArray.NULL_VALUE;
		}
		return newArray;
	}
	
	public String toString() {
		String out = "[";
		for(int i = 0; i<data.length; i++) {
			out += data[i] + ",";
		}
		return out + "]";
	}
}
